dnl AC_TRY_CFLAGS (CFLAGS, [ACTION-IF-WORKS], [ACTION-IF-FAILS])
dnl check if $CC supports a given set of cflags
AC_DEFUN([AC_TRY_CFLAGS],
    [AC_MSG_CHECKING([if $CC supports $1 flags])
    SAVE_CFLAGS="$CFLAGS"
    CFLAGS="$1"
    AC_TRY_COMPILE([],[],[ac_cv_try_cflags_ok=yes],[ac_cv_try_cflags_ok=no])
    CFLAGS="$SAVE_CFLAGS"
    AC_MSG_RESULT([$ac_cv_try_cflags_ok])
    if test x"$ac_cv_try_cflags_ok" = x"yes" ; then
        ifelse([$2],[],[:],[$2])
    else
        ifelse([$3],[],[:],[$3])
    fi])

dnl AC_TRY_VALID_CFLAGS (CFLAGS)
dnl check if $CC supports a given set of cflags - only leave valid flags in set
AC_DEFUN([AC_TRY_VALID_CFLAGS],
    [ac_cv_try_valid_cflags=""
    for i in ${$1}; do
      AC_TRY_CFLAGS($i, ac_cv_try_valid_cflags="$ac_cv_try_valid_cflags $i")
    done
    $1=$ac_cv_try_valid_cflags])

dnl AC_TRY_FIRST_VALID_CFLAG (CFLAGS)
dnl check if $CC supports a given set of cflags - only leave the first valid
dnl flag in the set
AC_DEFUN([AC_TRY_FIRST_VALID_CFLAG],
    [ac_cv_try_valid_cflags=""
    for i in ${$1}; do
      if test x$ac_cv_try_valid_cflags = x ; then
        AC_TRY_CFLAGS($i, ac_cv_try_valid_cflags=$i)
      fi
    done
    $1=$ac_cv_try_valid_cflags])

dnl AC_TRY_FIRST_VALID_ARCH (CFLAGS)
dnl check if $CC supports a given set of arch flags - only leave the first valid
dnl flag in the set
AC_DEFUN([AC_TRY_FIRST_VALID_ARCH],
    [ac_cv_try_valid_cflags=""
    for i in ${$1}; do
      if test x$ac_cv_try_valid_cflags = x ; then
        AC_TRY_CFLAGS(-mcpu=$i, ac_cv_try_valid_cflags=$i)
      fi
    done
    $1=$ac_cv_try_valid_cflags])

dnl AC_TRY_CPU (FUNCTION, INSN)
dnl check if a) $CC supports insn b) if CPU supports insn
dnl Sets a define of HAVE_${FUNCTION} to 1 if compiler can compile insn
dnl or 2 if it can be compiled and executed.
dnl Global var ac_cv_opt_cflags_insn must be yes to test and set execution
dnl of insn.
AC_DEFUN([AC_TRY_CPU],
  [if test x$ac_cv_opt_cflags_insn = x ; then
    AC_MSG_ERROR(ac_try_cpu called before ac_opt_cflags!)
  fi
  ac_try_cpu_have_$1=0
  AC_ARG_ENABLE($1, [  --enable-$1        enable $1 code [default=yes]])
  if test x${enable_$1-yes} = xyes ; then
    AC_MSG_CHECKING(for $1-capable compiler)
    AC_TRY_COMPILE(, [__asm__ __volatile__ ("$2\n\t")], ac_try_cpu_have_$1=1)
    if test x${ac_try_cpu_have_$1} = x1 ; then
      AC_MSG_RESULT(yes)
      if test x$ac_cv_opt_cflags_insn = xyes ; then
        AC_MSG_CHECKING(for $1-capable CPU)
        AC_TRY_RUN([
	  int main(void)
	  {__asm__ __volatile__ ("$2\n\t"); return 0;}], ac_try_cpu_have_$1=2)
        if test x${ac_try_cpu_have_$1} = x2 ; then
          AC_MSG_RESULT(yes)
        else
          AC_MSG_RESULT(no)
        fi
      fi
    else
      AC_MSG_RESULT(no)
    fi
    AC_DEFINE_UNQUOTED(HAVE_$1, ${ac_try_cpu_have_$1}, $1)
  fi])

dnl AC_OPT_X86_EXT
dnl X86 processor feature detection. Using first instance as an example.
dnl Try compile a program with inline assembly instruction
dnl "psubusb %mm1, %mm2". If it compiles, set a config.h define for
dnl HAVE_X86_MMX=1. Next, if USE_CPU_INSN=yes, try to execute the
dnl instruction, if it executes, set HAVE_X86_MMX=2.
dnl Repeat with specific instructions for all features we wish to detect.
AC_DEFUN([AC_OPT_X86_EXT],
  [AC_TRY_CPU(X86_MMX, [pmulhw %mm1, %mm2])
  AC_TRY_CPU(X86_MMXEXT, [pmulhuw %mm1, %mm2])
  AC_TRY_CPU(X86_SSE, [mulps %xmm1, %xmm2])
  AC_TRY_CPU(X86_SSE2, [mulpd %xmm1, %xmm2])
  AC_TRY_CPU(X86_3DNOW, [pfmul %mm1, %mm2])
  AC_TRY_CPU(X86_3DNOWEXT, [pfnacc %mm1, %mm2])])

dnl AC_OPT_CFLAGS (DEFAULT)
dnl Setup CFLAGS and OPT_CFLAGS.
dnl 1) if cflags is passed to configure, use these.
dnl 2) add any CFLAGS passed in "default".
dnl 3) check configure option en/disable-cpu_insn and set
dnl    $ac_cv_opt_cflags_insn accordingly.
dnl 4) check configure option --with-cpu.
dnl    if it is "yes" (default) determine current cpu.
dnl    if it is "host" optimise for ac_canonical_host.
dnl    if it is "build" optimise for ac_canonical_build.
dnl    if it is "target" optimise for ac_canonical_target.
dnl    if it is "no" do not optimise for specific cpu.
dnl    otherwise use this value as the cpu type
dnl 5) determine optimum flags for cpu.
dnl 6) generate CFGLAGS and OPT_CFLAGS from this.
dnl
dnl 4 (contd.) when determining the current CPU type:
dnl   1. check /proc/cpuinfo for the field "model name" (Linux)
dnl   2. check sysctl -a for the field "hw.model" (BSD)
dnl   3. use $host_cpu
dnl
dnl Extra options:
dnl   --[en|dis]able-optimise
dnl        Enable / disable optimisations.  Defaults to enabled
dnl   --[en|dis]able-debug
dnl        Enable / disable debugging. Defaults to disabled.
dnl        Enabling implies --disable-optimise
AC_DEFUN([AC_OPT_CFLAGS],
 [if test x$ac_cv_opt_cflags_insn = x ; then
    AC_MSG_NOTICE(Beginning CFLAG optimisation)

    dnl --en/disable-cpu_insn argument
    AC_ARG_ENABLE(cpu_insn,
      [  --enable-cpu_insn          enable cpu specific instructions [default=yes]])
    if test x${enable_cpu_insn-yes} = xyes ; then
      ac_cv_opt_cflags_insn=yes
    else
      ac_cv_opt_cflags_insn=no
    fi

    dnl --cpu argument
    AC_ARG_WITH(cpu,
      [  --with-cpu=cpu type        choose CPU to optimise for (default=yes=current CPU)],
      ac_cv_opt_cflags_use_cpu=$withval,
      ac_cv_opt_cflags_use_cpu=yes)

    dnl handle default CFLAGS
    if test "x${ac_test_CFLAGS}" = xset ; then
      CFLAGS="$1 $CFLAGS"
    else
      CFLAGS="$1"
    fi

    dnl --en/disable-optimise argument
    AC_ARG_ENABLE(optimise,
      [  --enable-optimise          enable cflag optimisations [default=yes]])
    if test x${enable_optimise-yes} = xyes ; then
      AC_MSG_NOTICE(CFLAG optimisations enabled)
    else
      AC_MSG_NOTICE(CFLAG optimisations disabled)
      ac_cv_opt_cflags_use_cpu="nom"
    fi

    dnl --en/disable-debug argument
    AC_ARG_ENABLE(debug,
      [  --enable-debug          enable debug cflags [default=no]])
    if test x${enable_debug-no} = xyes ; then
      AC_MSG_NOTICE(Debug CFLAGS enabled)
      ac_cv_opt_cflags_use_cpu="debug"
    fi

    dnl handle cpu type detection if requested
    if test x$ac_cv_opt_cflags_use_cpu = xyes ; then
      AC_MSG_NOTICE(Attempting to determine host CPU type...)

      dnl Linux /proc/cpuinfo
      ac_cv_opt_cflags_tmp=`cat /proc/cpuinfo | grep "model name" | cut -f 2 -d : | cut -f 2- -d" " | cut -f 1 -d "
"`
      if test x"$ac_cv_opt_cflags_tmp" = x"" ; then
        AC_MSG_NOTICE(Linux style /proc/cpuinfo not found...)
      fi

      dnl If we haven't got it yet, try BSD style sysctl
      if test x"$ac_cv_opt_cflags_tmp" = x"" ; then
        ac_cv_opt_cflags_tmp=`sysctl -a | grep "hw.model" | cut -f 2 -d : | cut -f 2- -d" " | cut -f 1 -d "
"`
      fi
      if test x"$ac_cv_opt_cflags_tmp" = x"" ; then
        AC_MSG_NOTICE(BSD style sysctl not found...)
      fi

      dnl Any others?
      dnl Fall back to $host_cpu
      case x"$ac_cv_opt_cflags_tmp" in
        x"AMD Athlon(tm) XP 2000+") ac_cv_opt_cflags_use_cpu=k7;;
        x"AMD Athlon(tm) processor") ac_cv_opt_cflags_use_cpu=k7;;
        x"AMD Athlon(tm) Processor") ac_cv_opt_cflags_use_cpu=k7;;
        x"AMD-K6(tm) 3D processor") ac_cv_opt_cflags_use_cpu=k6;;
        x"Pentium II (Deschutes)") ac_cv_opt_cflags_use_cpu=i686;;
        x"Pentium II (Klamath)") ac_cv_opt_cflags_use_cpu=i686;;
        x"Pentium III (Coppermine)") ac_cv_opt_cflags_use_cpu=i686;;
        x"Pentium III (Katmai)") ac_cv_opt_cflags_use_cpu=i686;;
        x"Intel(R) Pentium(R) 4 CPU 1500MHz") ac_cv_opt_cflags_use_cpu=i686;;
        dnl !FIXME! Finish this list!
        *) AC_MSG_WARN(Could not determine CPU type please contact package maintainer.)
           if test x$host_cpu = x ; then
               AC_MSG_ERROR(ac_canonical_host must be called before ac_opt_cflags!)
           fi
           AC_MSG_NOTICE(Found ac_canonical_host: $host_cpu)
           ac_cv_opt_cflags_use_cpu=$host_cpu;;
      esac
    else
      dnl !FIXME! Disable cpu specific instructions entirely if we are not
      dnl building on the target CPU.
      AC_MSG_WARN(Targeting a CPU different from the host... Disabling CPU specific instructions)
      ac_cv_opt_cflags_insn=no
    fi

    if test x$ac_cv_opt_cflags_use_cpu = xbuild ; then
      if test x$build_cpu = x ; then
        AC_MSG_ERROR(ac_canonical_build must be called before ac_opt_cflags!)
      fi
      AC_MSG_NOTICE(Found ac_canonical_build: $build_cpu)
      ac_cv_opt_cflags_use_cpu=$build_cpu
    fi
    if test x$ac_cv_opt_cflags_use_cpu = xhost ; then
      if test x$host_cpu = x ; then
        AC_MSG_ERROR(ac_canonical_host must be called before ac_opt_cflags!)
      fi
      AC_MSG_NOTICE(Found ac_canonical_host: $host_cpu)
      ac_cv_opt_cflags_use_cpu=$host_cpu
    fi
    if test x$ac_cv_opt_cflags_use_cpu = xtarget ; then
      if test x$target_cpu = x ; then
        AC_MSG_ERROR(ac_canonical_target must be called before ac_opt_cflags!)
      fi
      AC_MSG_NOTICE(Found ac_canonical_target: $target_cpu)
      ac_cv_opt_cflags_use_cpu=$target_cpu
    fi

    AC_MSG_NOTICE(Optimising for CPU: $ac_cv_opt_cflags_use_cpu)

    if test x"$RPM_OPT_FLAGS" = "x" ; then
      AC_MSG_WARN(Using CPU specific optimisations!)
      dnl Choose optimisations for each CPU.
      dnl ac_cv_opt_cflags_tmp_flags is the list of prefererd -mcpu= values in order of preference
      dnl ac_cv_opt_cflags_opt_flag_nom is the list of preferred medium optimisation flags in
      dnl order of preference.
      dnl ac_cv_opt_cflags_insn_opt_flag_max is the list of preferred high optimisation flags in
      dnl order of preference.
      case x"$ac_cv_opt_cflags_use_cpu" in
        dnl x86 opts from http://www.tldp.org/LDP/solrhe/Securing-Optimizing-Linux-RH-Edition-v1.3/chap6sec64.html
        x"i686")
          ac_cv_opt_cflags_tmp_flags="i686 pentiumpro i586 pentium i486 i386"
          ac_cv_opt_cflags_opt_flag_nom="-O3 -O2 -O1 -O"
          ac_cv_opt_cflags_opt_flag_max="-O9 -O8 -O7 -O6 -O5 -O4 -O3 -O2 -O1 -O"
          CFLAGS="-funroll-loops -ffast-math -fomit-frame-pointer -fno-exceptions $CFLAGS"
          ;;
        x"i586")
          ac_cv_opt_cflags_tmp_flags="i586 pentium i486 i386"
          ac_cv_opt_cflags_opt_flag_nom="-O3 -O2 -O1 -O"
          ac_cv_opt_cflags_opt_flag_max="-O3 -O2 -O1 -O"
          CFLAGS="-ffast-math -funroll-loops -fomit-frame-pointer -fforce-mem -fforce-addr -fno-exceptions $CFLAGS"
          ;;
        x"i486")
          ac_cv_opt_cflags_tmp_flags="i486 i386"
          ac_cv_opt_cflags_opt_flag_nom="-O3 -O2 -O1 -O"
          ac_cv_opt_cflags_opt_flag_max="-O3 -O2 -O1 -O"
          CFLAGS="-funroll-all-loops -fomit-frame-pointer -fno-exceptions $CFLAGS"
          ;;
        dnl AMD optimisations 
        dnl !FIXME! NOT SURE!
        x"k7")
          ac_cv_opt_cflags_tmp_flags="athlon k7 i686 pentiumpro i586 pentium i486 i386"
          ac_cv_opt_cflags_opt_flag_nom="-O3 -O2 -O1 -O"
          ac_cv_opt_cflags_opt_flag_max="-O9 -O8 -O7 -O6 -O5 -O4 -O3 -O2 -O1 -O"
          CFLAGS="-fschedule-insns2 -funroll-loops -ffast-math -fomit-frame-pointer -fno-exceptions $CFLAGS"
          ;;
        x"k6")
          ac_cv_opt_cflags_tmp_flags="k6 i686 pentiumpro i586 pentium i486 i386"
          ac_cv_opt_cflags_opt_flag_nom="-O3 -O2 -O1 -O"
          ac_cv_opt_cflags_opt_flag_max="-O9 -O8 -O7 -O6 -O5 -O4 -O3 -O2 -O1 -O"
          CFLAGS="-fschedule-insns2 -funroll-loops -ffast-math -fomit-frame-pointer -fno-exceptions $CFLAGS"
          ;;
        dnl Optimisations turned off
        x"nom")
          ac_cv_opt_cflags_tmp_flags=""
          ac_cv_opt_cflags_opt_flag_nom="-O2 -O1 -O"
          ac_cv_opt_cflags_opt_flag_max="-O2 -O1 -O"
          CFLAGS="$CFLAGS"
          ;;
        dnl Debugging turned on
        x"debug")
          ac_cv_opt_cflags_tmp_flags=""
          ac_cv_opt_cflags_opt_flag_nom="-O"
          ac_cv_opt_cflags_opt_flag_max="-O"
          CFLAGS="-DDEBUG $CFLAGS -g"
          ;;
        dnl No specific optimisations
        x"no")
          ac_cv_opt_cflags_tmp_flags=""
          ac_cv_opt_cflags_opt_flag_nom="-O3 -O2 -O1 -O"
          ac_cv_opt_cflags_opt_flag_max="-O3 -O2 -O1 -O"
          CFLAGS="$CFLAGS"
          ;;
        *)
          AC_MSG_WARN(Could not determine CPU type please contact package maintainer.)
          ac_cv_opt_cflags_tmp_flags=""
          ac_cv_opt_cflags_opt_flag_nom="-O3 -O2 -O1 -O"
          ac_cv_opt_cflags_opt_flag_max="-O3 -O2 -O1 -O"
          CFLAGS="$CFLAGS"
          ;;
      esac

      dnl Choose best arch first (only set arch if we may use CPU specific insns)
      if test x"$ac_cv_opt_cflags_tmp_flags" != x ; then
        AC_TRY_FIRST_VALID_ARCH(ac_cv_opt_cflags_tmp_flags)
        if test x$ac_cv_opt_cflags_insn = xyes ; then
          CFLAGS="-march=$ac_cv_opt_cflags_tmp_flags $CFLAGS"
        fi
        CFLAGS="-mcpu=$ac_cv_opt_cflags_tmp_flags $CFLAGS"
      fi

      dnl Now choose the highest available O flag in our range
      AC_TRY_FIRST_VALID_CFLAG(ac_cv_opt_cflags_opt_flag_nom)
      AC_TRY_FIRST_VALID_CFLAG(ac_cv_opt_cflags_opt_flag_max)
    else
      AC_MSG_WARN(Using RPM specified optimisation flags and disabling CPU specific optimisations!)
      ac_cv_opt_cflags_insn=no
      CFLAGS="$CFLAGS $RPM_OPT_FLAGS"
      ac_cv_opt_cflags_opt_flag_nom=""
      ac_cv_opt_cflags_insn_opt_flag_max=""
    fi

    dnl Check CFLAGS
    AC_MSG_NOTICE(Checking CFLAGS against current compiler...)
    AC_TRY_VALID_CFLAGS(CFLAGS)

    dnl Set the optimisation levels
    OPT_CFLAGS="$ac_cv_opt_cflags_opt_flag_max $CFLAGS"
    CFLAGS="$ac_cv_opt_cflags_opt_flag_nom $CFLAGS"
  else
    AC_MSG_NOTICE(ac_opt_cflags called twice! Please contact package maintainer!)
  fi])
