#ifdef HAVE_CONVERT
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#define _CONVERT_INTERNAL_
#include "convert.h"
#include "nvrec.h"

void convert_end(convert_t * c)
{
	if(c) {
		if(c->vhook) {
			free(c->vhook);
		}
		if(c->obuf) {
			if(*c->obuf) {
				free(*c->obuf);
			}
			free(c->obuf);
		}
		free(c);
	}
}

convert_t * convert_new(unsigned int infmt, unsigned int outfmt, int width, int height, char * vhook)
{
	convert_t * c = NULL;
	int i;
	
	c = (convert_t *)malloc(sizeof(convert_t));
	if(!c) {
		EPRINT("Could not malloc new converter struct!\n");
		goto error;
	}
	memset(c, 0, sizeof(convert_t));
	c->width = width;
	c->height = height;
	c->infmt = infmt;
	switch(infmt) {
		case NV_RGB24:
			IPRINT("Input format is RGB24\n");
			c->ff_infmt = PIX_FMT_BGR24;
			break;
		case NV_YUV420:
			IPRINT("Input format is YUV420\n");
			c->ff_infmt = PIX_FMT_YUV420P;
			break;
		case NV_YUV422:
			IPRINT("Input format is YUV422\n");
			c->ff_infmt = PIX_FMT_YUV422;
			break;
		default:
			EPRINT("Input format is unknown!\n");
			goto error;
	}
	c->outfmt = outfmt;
	switch(outfmt) {
		case NV_RGB24:
			IPRINT("Output format is RGB24\n");
			c->ff_outfmt = PIX_FMT_RGB24;
			c->obuf = (unsigned char **)malloc(height * sizeof(unsigned char *));
			if(!c->obuf) {
				EPRINT("Could not malloc plane pointers!\n");
				goto error;
			}
			memset(c->obuf, 0, height * sizeof(unsigned char *));
			c->obuf[0] = (unsigned char *)malloc(width * height * 3);
			if(!c->obuf[0]) {
				EPRINT("Could not malloc frame\n");
				goto error;
			}
			memset(c->obuf[0], 0, width * height * 3);
			for(i = 1; i < height; i++) {
				c->obuf[i] = c->obuf[i - 1] + (width * 3);
			}
			c->opict.data[0] = c->obuf[0];
			c->opict.linesize[0] = width * 3;
			break;
		case NV_YUV420:
			IPRINT("Output format is YUV420\n");
			c->ff_outfmt = PIX_FMT_YUV420P;
			c->obuf = (unsigned char **)malloc(3 * sizeof(unsigned char *));
			if(!c->obuf) {
				EPRINT("Could not malloc plane pointers!\n");
				goto error;
			}
			memset(c->obuf, 0, 3 * sizeof(unsigned char *));
			c->obuf[0] = (unsigned char *)malloc((width * height * 3) / 2);
			if(!c->obuf[0]) {
				EPRINT("Could not malloc frame\n");
				goto error;
			}
			memset(c->obuf[0], 0, (width * height * 3) / 2);
			c->obuf[1] = c->obuf[0] + (width * height);
			c->obuf[2] = c->obuf[1] + (width * height / 4);
			c->opict.data[0] = c->obuf[0];
			c->opict.data[1] = c->obuf[1];
			c->opict.data[2] = c->obuf[2];
			c->opict.linesize[0] = width;
			c->opict.linesize[1] = width / 2;
			c->opict.linesize[2] = width / 2;
			break;
		case NV_YUV422:
			IPRINT("Output format is YUV422\n");
			c->ff_outfmt = PIX_FMT_YUV422;
			c->obuf = (unsigned char **)malloc(height * sizeof(unsigned char *));
			if(!c->obuf) {
				EPRINT("Could not malloc plane pointers!\n");
				goto error;
			}
			memset(c->obuf, 0, height * sizeof(unsigned char *));
			c->obuf[0] = (unsigned char *)malloc(width * height * 2);
			if(!c->obuf[0]) {
				EPRINT("Could not malloc frame\n");
				goto error;
			}
			memset(c->obuf[0], 0, width * height * 2);
			for(i = 1; i < height; i++) {
				c->obuf[i] = c->obuf[i - 1] + (width * 2);
			}
			c->opict.data[0] = c->obuf[0];
			c->opict.linesize[0] = width * 2;
			break;
		default:
			EPRINT("Input format is unknown!\n");
			goto error;
	}
	if(vhook) {
		int ac = 0;
		char *av[64];
		
		IPRINT("Registering vhook '%s'...\n", vhook);
		c->vhook = strdup(vhook);
		if(!c->vhook) {
			EPRINT("Could not malloc memory for vhook!\n");
			goto error;
		}
		av[0] = strtok(c->vhook, " ");
		while (ac < 62 && (av[++ac] = strtok(NULL, " ")));
		if((frame_hook_add(ac, av)) != 0) {
			EPRINT("Could not add vhook '%s'\n", vhook);
			goto error;
		}
	}
	return c;
error:
	convert_end(c);
	return NULL;
}

/*
 * FIXME: Not formally correct - does not verify that frame is packed!
 */

unsigned char ** convert_convert(convert_t * c, unsigned char ** buf)
{
	if(!c) {
		EPRINT("convert_convert called with NULL convert!\n");
		return NULL;
	}
	memset(&c->ipict, 0, sizeof(c->ipict));
	switch(c->infmt) {
		case NV_RGB24:
			c->ipict.data[0] = buf[0];
			c->ipict.linesize[0] = c->width * 3;
			break;
		case NV_YUV420:
			c->ipict.data[0] = buf[0];
			c->ipict.data[1] = buf[1];
			c->ipict.data[2] = buf[2];
			c->ipict.linesize[0] = c->width;
			c->ipict.linesize[1] = c->width / 2;
			c->ipict.linesize[2] = c->width / 2;
			break;
		case NV_YUV422:
			c->ipict.data[0] = buf[0];
			c->ipict.linesize[0] = c->width * 2;
			break;
		default:
			EPRINT("Input format is unknown!\n");
			return NULL;
	}
	if(img_convert(&c->opict, c->ff_outfmt, &c->ipict, c->ff_infmt, c->width, c->height) < 0) {
		EPRINT("Image conversion failed!\n");
		return NULL;
	}
	if(c->vhook) {
		frame_hook_process(&c->opict, c->ff_outfmt, c->width, c->height);
	}
	return c->obuf;
}


#else
#warning NO CONVERSION FUNCTIONS AVAILABLE
#endif
