#include <stdio.h>
#include <signal.h>
#include <setjmp.h>
#include "cpu_features.h"
#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

int cpu_features = -1;

volatile int cpu_features_state;
sigjmp_buf cpu_features_env;

RETSIGTYPE cpu_features_sigill_handler(int foo)
{
#ifdef DEBUG
	fprintf(stderr, "SIGILL!\n");
#endif
	siglongjmp(cpu_features_env, 1);
#if RETSIGTYPE != void
	return 0;
#endif
}

#ifdef DEBUG
#define FNTEST(insn, flag, name)	\
	__asm__ __volatile__ (insn);	\
	cpu_features |= flag;		\
	fprintf(stderr, "cpu_features_init: Detected " name "!\n");
#else
#define FNTEST(insn, flag, name)	\
	__asm__ __volatile__ (insn);	\
	cpu_features |= flag;
#endif

void cpu_features_init(void)
{
	struct sigaction newaction, oldaction;
	
	if(cpu_features >= 0)
		return;
#ifdef DEBUG
	fprintf(stderr, "cpu_features_init\n");
#endif

	cpu_features = 0;
	newaction.sa_handler = (void(*)(int))cpu_features_sigill_handler;
	sigfillset(&newaction.sa_mask);
	newaction.sa_flags = 0;
	if (sigaction(SIGILL, &newaction, &oldaction) != 0) {
		fprintf(stderr, "cpu_features_init: Could not capture SIGILL - all features set to disabled!\n");
		return;
	}

	cpu_features_state = 0;
	if(sigsetjmp(cpu_features_env, 1) != 0) {
		switch(cpu_features_state) {
			case 1: goto l1;
			case 2: goto l2;
			case 3: goto l3;
			case 4: goto l4;
			case 5: goto l5;
			case 6: goto l6;
			default:
				fprintf(stderr, "cpu_features_init: unhandled state!\n");
				return;
		}
	}

	cpu_features_state = 1;
#if HAVE_X86_MMX > 0
	FNTEST("pmulhw %mm1, %mm2\n\t", CPU_FEATURE_MMX, "MMX");
#endif
l1:
	cpu_features_state = 2;
#if HAVE_X86_MMXEXT > 0
	FNTEST("pmulhuw %mm1, %mm2\n\t", CPU_FEATURE_MMXEXT, "MMXEXT");
#endif
l2:
	cpu_features_state = 3;
#if HAVE_X86_SSE > 0
	FNTEST("mulps %xmm1, %xmm2\n\t", CPU_FEATURE_SSE, "SSE");
#endif
l3:
	cpu_features_state = 4;
#if HAVE_X86_SSE2 > 0
	FNTEST("mulpd %xmm1, %xmm2\n\t", CPU_FEATURE_SSE2, "SSE2");
#endif
l4:
	cpu_features_state = 5;
#if HAVE_X86_3DNOW > 0
	FNTEST("pfmul %mm1, %mm2\n\t", CPU_FEATURE_3DNOW, "3DNOW");
#endif
l5:
	cpu_features_state = 6;
#if HAVE_X86_3DNOWEXT > 0
	FNTEST("pfnacc %mm1, %mm2\n\t", CPU_FEATURE_3DNOWEXT, "3DNOWEXT");
#endif
l6:
	if(cpu_features)
		__asm__ __volatile__("emms\n\t");
	sigaction(SIGILL, &oldaction, NULL);
	return;
}
