/* 
   divx4file_core.c (C) Justin Schoeman 2001 (justin@suntiger.ee.up.ac.za)

    RAW support by Tilmann Bitterberg <transcode@tibit.org> (integrated by
    Justin Schoeman) 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdint.h>
#include <dlfcn.h>
#include "../avilib/avilib.h"
#include "mcs.h"
#ifdef HAVE_CONFIG_H
# include "../config.h"
#endif
#ifdef HAVE_ENCORE2
# include "encore2.h"
#endif
#ifdef HAVE_MP3LAME
# include <lame/lame.h>
#endif
#include "../nvrec.h"
#ifdef HAVE_CONVERT
#include "../convert.h"
#endif

#ifdef HAVE_ENCORE2
# ifdef ENCORE_MAJOR_VERSION
#  define DIVX_API ENCORE_MAJOR_VERSION
# else
#  define DIVX_API 4000
# endif
#endif

#define MP3BUFSIZE 65536

struct divx4file_core_t {
	avi_t * avi;
	mcs_t * mcs;
	int (*encore)(void *handle, int enc_opt, void *param1, void *param2);
	int divx_version;
	void * dh;
	void * dlh;
#ifdef HAVE_MP3LAME
	lame_global_flags * lgf;
#endif
	unsigned char * cbuf;
	unsigned char mbuf[MP3BUFSIZE];
#ifdef HAVE_ENCORE2
	ENC_FRAME enc_frame;
	ENC_RESULT enc_result;
#endif
	int achans;
	int aformat;
	int vformat;
	int vwidth;
	int vheight;
	uint64_t atime;
	uint64_t vtime;
	uint64_t tstep;
	int drop;
	short *lbuf;
	short *rbuf;
	int bufsize;
#ifdef HAVE_CONVERT
	convert_t * conv;
#endif
};

struct file_core_t divx4file_core;

void divx4file_core_audio(struct file_core_t * fct, unsigned char *buf, int n)
{
	struct divx4file_core_t * divx4c = (struct divx4file_core_t *)fct->priv;
	
	if(!divx4c)
		return;
	if(divx4c->aformat == 0x55) {
#ifdef HAVE_MP3LAME
		short *sbuf = (short *)buf;
		int i = 0;

		if(divx4c->achans > 1) {
			n >>= 2;
			if(divx4c->bufsize < (n * 2)) {
				if(divx4c->lbuf) free(divx4c->lbuf);
				if(divx4c->rbuf) free(divx4c->rbuf);
				divx4c->lbuf = malloc(n * 4);
				divx4c->rbuf = malloc(n * 4);
				divx4c->bufsize = n * 2;
			}
			for(i = 0; i < n; i++) {
				divx4c->lbuf[i] = sbuf[i << 1];
				divx4c->rbuf[i] = sbuf[(i << 1) + 1];
			}
		} else {
			n >>= 1;
/*
			if(divx4c->bufsize < (n * 2)) {
				if(divx4c->rbuf) free(divx4c->rbuf);
				divx4c->rbuf = malloc(n * 4);
				memset(divx4c->rbuf, 0, n * 4);
				divx4c->bufsize = n * 2;
			}
*/
			divx4c->lbuf = sbuf;
			divx4c->rbuf = sbuf;
		}
		i = lame_encode_buffer(divx4c->lgf, divx4c->lbuf, divx4c->rbuf, n, divx4c->mbuf, MP3BUFSIZE);
		if(i > 0) {
			if(divx4c->avi) {
				AVI_write_audio(divx4c->avi, divx4c->mbuf, i);
			} else {
				MCS_write_audio(divx4c->mcs, divx4c->mbuf, i, divx4c->atime);
			}
		}
#else
		EPRINT("MP3 encoder not available\n");
		divx4c->atime += divx4c->tstep;
		return;
#endif
	} else {
		if(divx4c->avi) {
			AVI_write_audio(divx4c->avi, buf, n);
		} else {
			MCS_write_audio(divx4c->mcs, buf, n, divx4c->atime);
		}
	}
	divx4c->atime += divx4c->tstep;
}

void divx4file_core_close(struct file_core_t * fct)
{
	struct divx4file_core_t * divx4c = (struct divx4file_core_t *)fct->priv;
	int i;
	
	if(!divx4c)
		return;
	if(divx4c->cbuf)
		free(divx4c->cbuf);
#ifdef HAVE_ENCORE2
	if(divx4c->encore) {
		if(divx4c->encore(divx4c->dh, ENC_OPT_RELEASE, NULL, NULL) != ENC_OK)
			EPRINT("Error freeing encore device!\n");
		dlclose(divx4c->dlh);
	}
#endif
#ifdef HAVE_MP3LAME
	if(divx4c->aformat == 0x55) {
		i = lame_encode_finish(divx4c->lgf, divx4c->mbuf, MP3BUFSIZE);
		if(i > 0) {
			if(divx4c->avi)
				AVI_write_audio(divx4c->avi, divx4c->mbuf, i);
			else
				MCS_write_audio(divx4c->mcs, divx4c->mbuf, i, divx4c->atime);
		}
	}
#endif
	if(divx4c->avi)
		AVI_close(divx4c->avi);
	if(divx4c->mcs)
		MCS_close(divx4c->mcs);
#ifdef HAVE_CONVERT
	if(divx4c->conv)
		convert_end(divx4c->conv);
#endif
	free(divx4c);
	free(fct);
}

void divx4file_core_nullvideo(struct file_core_t * fct)
{
	struct divx4file_core_t * divx4c = (struct divx4file_core_t *)fct->priv;
	
	if(!divx4c)
		return;
	if(divx4c->encore) {
		if(divx4c->avi)
#if 1
			AVI_write_frame(divx4c->avi, divx4c->cbuf, 0, AVI_FLAG_NONE);
#else
			AVI_dup_frame(divx4c->avi);
#endif
		else
			MCS_write_frame(divx4c->mcs, divx4c->cbuf, 0, divx4c->vtime);
	} else {
		/* FIXME! */
	}
	divx4c->vtime += divx4c->tstep;
}

void divx4file_core_video(struct file_core_t * fct, unsigned char **buf)
{
	struct divx4file_core_t * divx4c = (struct divx4file_core_t *)fct->priv;
	
	if(!divx4c)
		return;
#ifdef HAVE_CONVERT
	if(divx4c->conv) {
		buf = convert_convert(divx4c->conv, buf);
		if(!buf) {
			EPRINT("Error converting image!\n");
			return;
		}
	}
#endif
	if(divx4c->encore) {
#ifdef HAVE_ENCORE2
		memset(&divx4c->enc_frame, 0, sizeof(ENC_FRAME));
		divx4c->enc_frame.image = buf[0];
		divx4c->enc_frame.bitstream = divx4c->cbuf;
		divx4c->enc_frame.colorspace = divx4c->vformat;
		if(divx4c->encore(divx4c->dh, ENC_OPT_ENCODE, &divx4c->enc_frame, &divx4c->enc_result) != ENC_OK) {
			EPRINT("Error encoding frame!\n");
			if(divx4c->avi)
				AVI_write_frame(divx4c->avi, divx4c->cbuf, 0, AVI_FLAG_NONE);
			else
				MCS_write_frame(divx4c->mcs, divx4c->cbuf, 0, divx4c->vtime);
		} else {
			if(divx4c->avi)
				AVI_write_frame(divx4c->avi, 
						divx4c->cbuf, 
						divx4c->enc_frame.length, 
						divx4c->enc_result.is_key_frame ? 
							AVI_FLAG_KEYFRAME : AVI_FLAG_NONE);
			else
				MCS_write_frame(divx4c->mcs, 
						divx4c->cbuf, 
						divx4c->enc_frame.length, 
						divx4c->vtime);
		}
#else
		EPRINT("No divx encoder available\n");
#endif
	} else {
		write(1, buf[0], divx4c->vwidth * divx4c->vheight);
		write(1, buf[2], divx4c->vwidth / 2 * divx4c->vheight / 2);
		write(1, buf[1], divx4c->vwidth / 2 * divx4c->vheight / 2);
		AVI_write_frame(divx4c->avi, 
			NULL, 
			0,
			AVI_FLAG_NONE);
	}
	divx4c->vtime += divx4c->tstep;
}

void divx4file_core_sync(struct file_core_t * fct)
{
	return;
}

#ifdef HAVE_CONVERT
static int divx4file_core_formats[] = {NV_YUV420, NV_YUV422, NV_RGB24, NV_END_FORMAT};
//static int divx4file_core_formats[] = {NV_YUV422, NV_RGB24, NV_END_FORMAT};

unsigned int *divx4file_core_format(char **args)
{
	return divx4file_core_formats;
}
#else
static int divx4file_core_formats[] = {NV_YUV420, NV_YUV422, NV_RGB24, NV_END_FORMAT};
static int divx4file_core_formats_raw[] = {NV_YUV420, NV_END_FORMAT};

unsigned int *divx4file_core_format(char **args)
{
	NV_ARGS;
	NV_SCAN_ARGS {
		if(strcmp(NV_THIS_ARG, "-o") == 0) {
			if(!NV_NEXT_ARG)
				break;
			if(strncmp(NV_THIS_ARG, "raw://", 6) == 0)
				return divx4file_core_formats_raw;
		}
	}
	return divx4file_core_formats;
}
#endif

void divx4file_core_usage()
{
	MPRINT("DIVX4 file core options:\n");
#ifdef HAVE_MP3LAME
	MPRINT("\t-ab <audio mp3 bitrate (kbps)>               [?]\n");
	MPRINT("\t-aq <audio mp3 quality factor: 0..100>       [?]\n");
	MPRINT("\t-wav use raw PCM instead of mp3              [mp3]\n");
#endif
#ifdef HAVE_ENCORE2
	MPRINT("\t-xvid <use xvid codec>                       [DivX;)]\n");
	MPRINT("\t-xvid_divx <use xvid codec with DivX fourcc> [DivX;)]\n");
	MPRINT("\t-vb <divx bitrate (kbps)>                    [?]\n");
	MPRINT("\t-vq <divx quality factor: 0..100>            [?]\n");
	MPRINT("\t-vg <maximum key interval>                   [?]\n");
	MPRINT("\t-dqmin <DivX minimum quantiser>              [?]\n");
	MPRINT("\t-dqmax <DivX maximum quantizer>              [?]\n");
	MPRINT("\t-drcp <DivX rate control period>             [?2000]\n");
	MPRINT("\t-drcrp <DivX rate control reaction period>   [?10]\n");
	MPRINT("\t-drcrr <DivX rate control reaction ratio>    [?20]\n");
	MPRINT("\t-deinterlace <deinterlace> *BROKEN*          [off]\n");
#if (DIVX_API >= 5000)
	MPRINT("\t-dbidir enable DivX B frames                 [off]\n");
	MPRINT("\t-ddint enable DivX deinterlacer              [off]\n");
	MPRINT("\t-dtfilt <DivX temporal filter level: 0.0-1.0>[off]\n");
	MPRINT("\t-dsfilt <DivX spatial filter level: 0.0-1.0> [off]\n");
	MPRINT("\t-dspass <DivX spatial filter passes: 1-3>    [off]\n");
#endif
#endif
#ifdef HAVE_CONVERT
	MPRINT("\t-vhook \"module name and params\"              [NONE]\n");
#endif
	MPRINT("\t-o <output filename>                         [NONE]\n");
	MPRINT("\t * filename starting with udp:// indicates multicast.\n");
	MPRINT("\t * filename starting with tcp:// indicates unicast.\n");
	MPRINT("\t * filename starting with raw:// indicates raw video\n");
	MPRINT("\t   will be written to stdout and audio to filename.\n");
}

struct file_core_t * divx4file_core_init(char ** args, unsigned int format, ntime_t vrate)
{
	struct file_core_t * fct;
	struct divx4file_core_t * divx4c;
	double f;
 	int arate = -1;
 	int abits = -1;
 	int abr = -1;
 	int aq = -1;
 	char enclib[50];
 	char fcc[5];
 	char *fname = NULL;
#ifdef HAVE_ENCORE2
 	ENC_PARAM enc_param;
#endif
	char *vhook = NULL;
 	NV_ARGS;
 	
 	fct =  malloc(sizeof(struct file_core_t));
 	if(!fct) {
 		PERROR("Could no malloc new cap core");
 		return NULL;
 	}
 	divx4c = malloc(sizeof(struct divx4file_core_t));
 	if(!divx4c) {
 		PERROR("Could no malloc new divx4 core");
 		free(fct);
 		return NULL;
 	}
 	memcpy(fct, &divx4file_core, sizeof(struct file_core_t));
 	fct->priv = (void *)divx4c;
 	memset(divx4c, 0, sizeof(struct divx4file_core_t));
#ifdef HAVE_MP3LAME
 	divx4c->aformat = 0x55;
#else
 	divx4c->aformat = WAVE_FORMAT_PCM;
#endif

 	strncpy(enclib, "libdivxencore.so", 50);
 	strncpy(fcc, "DIVX", 5);
 	
#ifdef HAVE_ENCORE2
 	switch(format) {
 		case NV_RGB24:	divx4c->vformat = ENC_CSP_RGB24; break;
 		case NV_YUV420:	divx4c->vformat = ENC_CSP_I420; break;
 		case NV_YUV422:	divx4c->vformat = ENC_CSP_YUY2; break;
 		default:
 			EPRINT("Invalid colour format!\n");
 			goto error0;
 	}
	memset(&enc_param, 0, sizeof(ENC_PARAM));
#endif
 	
	NV_SCAN_ARGS {
		if(strcmp(NV_THIS_ARG, "-r") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -r.\n");
				goto usage;
			}
			arate = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-b") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -b.\n");
				goto usage;
			}
			abits = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-s") == 0) {
			divx4c->achans = 2;
		} else if(strcmp(NV_THIS_ARG, "-m") == 0) {
			divx4c->achans = 1;
		} else if(strcmp(NV_THIS_ARG, "-w") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -w.\n");
				goto usage;
			}
			divx4c->vwidth = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-h") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -h.\n");
				goto usage;
			}
			divx4c->vheight = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-vhook") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -vhook.\n");
				goto usage;
			}
			vhook = NV_THIS_ARG;
#ifdef HAVE_MP3LAME
		} else if(strcmp(NV_THIS_ARG, "-ab") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -ab.\n");
				goto usage;
			}
			abr = atoi(NV_THIS_ARG);
			if(abr <= 0) {
				EPRINT("Invalid argument to -ab (>0).\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-aq") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -aq.\n");
				goto usage;
			}
			aq = atoi(NV_THIS_ARG);
			aq /= 10;
			aq = (aq > 9) ? 9 : aq;
			aq = 9 - aq;
			if((aq < 0) || (aq > 9)) {
				EPRINT("Invalid argument to -aq (0..100).\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-wav") == 0) {
			divx4c->aformat = WAVE_FORMAT_PCM;
#endif
#ifdef HAVE_ENCORE2
		} else if(strcmp(NV_THIS_ARG, "-vb") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -vb.\n");
				goto usage;
			}
			enc_param.bitrate = atoi(NV_THIS_ARG);
			if(enc_param.bitrate <= 0) {
				EPRINT("Invalid argument to -vb (>0).\n");
				goto usage;
			}
			enc_param.bitrate *= 1000;
		} else if(strcmp(NV_THIS_ARG, "-vq") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -vq.\n");
				goto usage;
			}
			enc_param.quality = atoi(NV_THIS_ARG);
			enc_param.quality += 20;
			enc_param.quality /= 20;
			enc_param.quality =
				(enc_param.quality > 5) ? 5 : enc_param.quality;
			if((enc_param.quality < 1) || (enc_param.quality > 5)) {
				EPRINT("Invalid argument to -vq (0..100).\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-dqmin") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -dqmin.\n");
				goto usage;
			}
			enc_param.min_quantizer = atoi(NV_THIS_ARG);
			if(enc_param.min_quantizer <= 0) {
				EPRINT("Invalid argument to -dqmin (>0).\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-dqmax") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -dqmax.\n");
				goto usage;
			}
			enc_param.max_quantizer = atoi(NV_THIS_ARG);
			if(enc_param.max_quantizer <= 0) {
				EPRINT("Invalid argument to -dqmax (>0).\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-vg") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -vg.\n");
				goto usage;
			}
			enc_param.max_key_interval = atoi(NV_THIS_ARG);
			if(enc_param.max_key_interval <= 0) {
				EPRINT("Invalid argument to -vg (>0).\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-drcp") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -drcp.\n");
				goto usage;
			}
			enc_param.rc_period = atoi(NV_THIS_ARG);
			if(enc_param.rc_period <= 0) {
				EPRINT("Invalid argument to -drcp (>0).\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-drcrp") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -drcrp.\n");
				goto usage;
			}
			enc_param.rc_reaction_period = atoi(NV_THIS_ARG);
			if(enc_param.rc_reaction_period <= 0) {
				EPRINT("Invalid argument to -drcrp (>0).\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-drcrr") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -drcrr.\n");
				goto usage;
			}
			enc_param.rc_reaction_ratio = atoi(NV_THIS_ARG);
			if(enc_param.rc_reaction_ratio <= 0) {
			
				EPRINT("Invalid argument to -drcrr (>0).\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-deinterlace") == 0) {
			enc_param.deinterlace = 1;
#endif

		} else if(strcmp(NV_THIS_ARG, "-o") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -o.\n");
				goto usage;
			}
			fname = NV_THIS_ARG;
		} else if(strcmp(NV_THIS_ARG, "-xvid") == 0) {
		 	strncpy(enclib, "libxvidcore.so", 50);
		 	strncpy(fcc, "XVID", 5);
		} else if(strcmp(NV_THIS_ARG, "-xvid_divx") == 0) {
		 	strncpy(enclib, "libxvidcore.so", 50);
		 	strncpy(fcc, "DIVX", 5);
		} else if(strcmp(NV_THIS_ARG, "-vr") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -vr.\n");
				goto usage;
			}
			vrate = (unsigned long long)(1e9 / atof(NV_THIS_ARG));
		}
#ifdef HAVE_ENCORE2
#if (DIVX_API >= 5000)
		else if(divx4c->divx_version > 5000) {
			if(strcmp(NV_THIS_ARG, "-dbidir") == 0) {
				enc_param.extensions.use_bidirect = 1;
			} else if(strcmp(NV_THIS_ARG, "-ddint") == 0) {
				enc_param.extensions.interlace_mode = 1;
			} else if(strcmp(NV_THIS_ARG, "-dtfilt") == 0) {
				enc_param.extensions.temporal_enable = 1;
				if(!NV_NEXT_ARG) {
					EPRINT("Missing argument to -dtfilt.\n");
					goto usage;
				}
				enc_param.extensions.temporal_level = atof(NV_THIS_ARG);
			} else if(strcmp(NV_THIS_ARG, "-dsfilt") == 0) {
				if(!NV_NEXT_ARG) {
					EPRINT("Missing argument to -dsfilt.\n");
					goto usage;
				}
				enc_param.extensions.spatial_level = atof(NV_THIS_ARG);
			} else if(strcmp(NV_THIS_ARG, "-dspass") == 0) {
				if(!NV_NEXT_ARG) {
					EPRINT("Missing argument to -dspass.\n");
					goto usage;
				}
				enc_param.extensions.spatial_passes = atoi(NV_THIS_ARG);
			}
		}
#endif
#endif
	}
	if(!fname) {
		EPRINT("Missing -o <filename> argument.\n");
		goto usage;
	}
	IPRINT("File: %s (%dx%d)\n", fname, divx4c->vwidth, divx4c->vheight);
	f = 1e9 / (double)vrate;
	IPRINT("Frame rate is %f\n", f);
	
	if(strncmp(fname, "raw://", 6) != 0) {
#ifdef HAVE_ENCORE2
		divx4c->dlh = dlopen(enclib, RTLD_NOW);
		if(!divx4c->dlh) {
			EPRINT("Error loading %s (perhaps you could try -xvid option?)\n", dlerror());
			goto error0;
		}
	
		divx4c->encore = dlsym(divx4c->dlh, "encore");
		if(!divx4c->encore) {
			EPRINT("Find encore %s\n", dlerror());
			goto error1;
		}
	
#ifdef ENC_OPT_VERSION
		if(strcmp(enclib, "libxvidcore.so") != 0) {
			divx4c->divx_version = divx4c->encore(NULL, ENC_OPT_VERSION, NULL, NULL);
			if(divx4c->divx_version != ENCORE_VERSION) {
				EPRINT("Encore binary does not match header this was compiled for!\n");
				goto error1;
			}
			divx4c->divx_version = DIVX_API;
			if(divx4c->divx_version == 5000) {
				EPRINT("This divx4 codec does not contain an encoder, please download the latest version.\n");
				goto error1;
			}
			if(divx4c->divx_version > 5000) {
				strncpy(fcc, "DX50", 5);
			}
		} else {
			divx4c->divx_version = 4000;
		}
#else
		divx4c->divx_version = divx4c->encore(NULL, 4, NULL, NULL);
		if(divx4c->divx_version > 20020000) {   
			EPRINT("You seem to have a newer divx codec installed - please recompile nvrec\n");
			goto error1;
		}
		divx4c->divx_version = 4000;
#endif
		if( (strncmp(fname + 3, "://", 3) == 0) ||
			(strcmp(fname, "-") == 0) ) {
			divx4c->mcs = MCS_open(fname);
			divx4c->avi = NULL;
			if(!divx4c->mcs) {
				PERROR("Could not open multicast");
				goto error1;
			}
		} else {
			divx4c->avi = AVI_open_output_file(fname);
			divx4c->mcs = NULL;
			if(!divx4c->avi) {
				PERROR("Could not open file");
				goto error1;
			}
		}
		enc_param.x_dim = divx4c->vwidth;
		enc_param.y_dim = divx4c->vheight;
		enc_param.framerate = f;
		if(divx4c->encore(NULL, ENC_OPT_INIT, &enc_param, NULL) != ENC_OK) {
			EPRINT("Error initing encoder\n");
			goto error1;
		}
		divx4c->dh = enc_param.handle;
		if(!divx4c->dh) {
			EPRINT("Error initialising divx4 encore.\n");
			goto error2;
		}
	
		IPRINT("Got DivX bitrate: %d\n", enc_param.bitrate);
	
		divx4c->cbuf = malloc(divx4c->vwidth * divx4c->vheight * 6);
		if(!divx4c->cbuf) {
			PERROR("Could not allocate compression buffer");
			goto error3;
		}
	
		if(divx4c->avi)
			AVI_set_video(divx4c->avi, divx4c->vwidth, divx4c->vheight, f, fcc);
		else
			MCS_set_video(divx4c->mcs, divx4c->vwidth, divx4c->vheight, f, fcc);
#else
		EPRINT("No divx encoder available\n");
		goto error1;
#endif
	} else {
		if(format != NV_YUV420) {
#ifdef HAVE_CONVERT
			IPRINT("No native format found - initialise conversion\n");
			divx4c->conv = convert_new(format, NV_YUV420, divx4c->vwidth, divx4c->vheight, vhook);
			if(!divx4c->conv) {
				EPRINT("Could not init format conversion core!\n");
				goto error1;
			}
#else
			EPRINT("No native format found, and conversion not available - giving up.\n");
			goto error1;
#endif
		}
		divx4c->avi = AVI_open_output_file(fname+6);
		AVI_set_video(divx4c->avi, 0, 0, f, fcc);
		divx4c->mcs = NULL;
		if(!divx4c->avi) {
			PERROR("Could not open file");
			goto error1;
		}
	}
#ifdef HAVE_CONVERT
	if((!divx4c->conv) && (vhook)) {
		IPRINT("Instantiating conversion function for vhooks...\n");
 		divx4c->conv = convert_new(format, format, divx4c->vwidth, divx4c->vheight, vhook);
 		if(!divx4c->conv) {
 			EPRINT("Could not init format conversion core!\n");
 			goto error1;
 		}
	}
#endif
	if(divx4c->aformat == 0x55) {
#ifdef HAVE_MP3LAME
		divx4c->lgf = lame_init();
		if(!divx4c->lgf) {
			EPRINT("Could not init lame lib!\n");
			goto error3;
		}
		if(abr > 0)
			lame_set_brate(divx4c->lgf, abr);
		if(aq > 0)
			lame_set_quality(divx4c->lgf, aq);
		if(divx4c->achans == 1)
			lame_set_mode(divx4c->lgf, MONO);
		else
			lame_set_mode(divx4c->lgf, STEREO);
		lame_set_num_channels(divx4c->lgf, divx4c->achans);
		lame_set_in_samplerate(divx4c->lgf, arate);
		if(lame_init_params(divx4c->lgf) < 0) {
			EPRINT("Could not init lame parameters!\n");
			goto error3;
		}
		if(divx4c->avi)
			AVI_set_audio(divx4c->avi, divx4c->achans, arate, abits, 
				divx4c->aformat, 
				(lame_get_brate(divx4c->lgf) * 1000) / 8, 
				lame_get_framesize(divx4c->lgf));
		else
			MCS_set_audio(divx4c->mcs, divx4c->achans, arate, abits, f, 
				divx4c->aformat);
#else
		EPRINT("No mp3 encoder available\n");
		goto error3;
#endif
	} else if(divx4c->aformat == WAVE_FORMAT_PCM) {
		if(divx4c->avi)
			AVI_set_audio(divx4c->avi, divx4c->achans, arate, abits, divx4c->aformat, 0, 0);
		else
			MCS_set_audio(divx4c->mcs, divx4c->achans, arate, abits, f, divx4c->aformat);
	} else {
		EPRINT("Oops - bad audio format!\n");
		goto error3;
	}
	divx4c->tstep = vrate;
	return (void *)fct;

error3:	
#ifdef HAVE_ENCORE2
	if(divx4c->encore)
		divx4c->encore(divx4c->dh, ENC_OPT_RELEASE, NULL, NULL);
#endif
error2:
	if(divx4c->avi)
		AVI_close(divx4c->avi);
	if(divx4c->mcs)
		MCS_close(divx4c->mcs);
error1:
#ifdef HAVE_CONVERT
	if(divx4c->conv)
		convert_end(divx4c->conv);
#endif
	if(divx4c->dlh)
		dlclose(divx4c->dlh);
error0:
	free(fct);
	free(divx4c);
	return NULL;

usage:
	EPRINT("Error parsing options.  Please run without any command line options for\n");
	EPRINT("detailed usage information.\n");
	free(fct);
	free(divx4c);
	return NULL;
}

void divx4file_core_overrides(char **args)
{
	nvcore_add_arg(args, "-b");
	nvcore_add_arg(args, "16");
	return;
}

struct file_core_t divx4file_core = {
	file_core_init: divx4file_core_init,
	file_core_close: divx4file_core_close,
	file_core_nullvideo: divx4file_core_nullvideo,
	file_core_fastvideo: divx4file_core_video,
	file_core_video: divx4file_core_video,
	file_core_audio: divx4file_core_audio,
	file_core_sync: divx4file_core_sync,
	file_core_format: divx4file_core_format,
	file_core_overrides: divx4file_core_overrides,
	file_core_usage: divx4file_core_usage,
	priv: NULL
};
