#include <sys/socket.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <stdint.h>
#include <pthread.h>
#include <netdb.h>

struct mcs_header {
	uint64_t sync __attribute__ ((packed)); /* sync word */
	uint64_t time __attribute__ ((packed)); /* timestamp */
	uint32_t size __attribute__ ((packed)); /* size of data to follow */
	uint32_t type __attribute__ ((packed)); /* type of frame */
};

#define MCS_CHUNK_SIZE 1024
#define MCS_HEADER_SIZE (sizeof(struct mcs_header))

/* sync word */
#define MCS_SYNC 0xffffffffffaa5500ULL

/* frame types */
#define MCS_FRAME_SYSTEM 0
#define MCS_FRAME_VIDEO 1
#define MCS_FRAME_AUDIO 2

struct mcs_system {
	uint16_t video_width __attribute__ ((packed));
	uint16_t video_height __attribute__ ((packed));
	uint32_t video_size __attribute__ ((packed));
	uint32_t video_format __attribute__ ((packed));
	uint64_t video_time __attribute__ ((packed));
	
	uint32_t audio_rate __attribute__ ((packed));
	uint8_t audio_bits __attribute__ ((packed));
	uint8_t audio_channels __attribute__ ((packed));
	uint32_t audio_size __attribute__ ((packed));
	uint32_t audio_uncomp_size __attribute__ ((packed));
	uint16_t audio_format __attribute__ ((packed));
	uint64_t audio_time __attribute__ ((packed));
};

struct mcs_tcp {
	int sock;
	struct mcs_tcp *next;
};           

typedef struct {
	int fd;
	struct sockaddr_in addr;
	struct mcs_header header;
	struct mcs_system system;
	char buffer[MCS_CHUNK_SIZE];
	int buffer_pos;
	int pkt_count;
	struct mcs_tcp *tcp;
	pthread_mutex_t tcp_lock;
	pthread_t mcs_accept;
} mcs_t;

#define MCS_HEADER_CNT 50

mcs_t *MCS_open(char * fname);
void MCS_set_video(mcs_t *mcs, int width, int height, double fps, char *compressor);
void MCS_set_audio(mcs_t *mcs, int channels, long rate, int bits, double fps, int format);
int  MCS_write_frame(mcs_t *mcs, char *data, long bytes, uint64_t time);
int  MCS_write_audio(mcs_t *mcs, char *data, long bytes, uint64_t time);
void MCS_close(mcs_t *mcs);
