/* 
   divxfile_core.c (C) Justin Schoeman 2001 (justin@suntiger.ee.up.ac.za)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
*/

/*
	Mercilessly ripped from x2divx (Ulrich Hecht uli@emulinks.de)
	by Justin Schoeman (c) 2001
*/

#ifdef HAVE_CONVERT
#warning conversion code being disabled until linking can be fixed...
#undef HAVE_CONVERT
#endif

#include <videoencoder.h>
#include <avifile.h>
#include <aviplay.h>
#include <version.h>
#include "avifile_version.h"
#ifdef AVIFILE_NEW_API
#include <fourcc.h>
#include <creators.h>
#include <stdio.h>
#include <stdlib.h>
#endif
#include "nvrec.h"
#ifdef HAVE_CONVERT
#include "convert.h"
#endif

#ifdef AVIFILE_NEW_API
using namespace Creators;
#endif

struct divxfile_core_t {
	IAviWriteFile * avifile;
	unsigned char * framebuffer;
 	CImage * im;
	IAviVideoWriteStream * vstream;
	IAviAudioWriteStream * astream;
	int height;
	int width;
	int ashift;
	int clip;
#ifdef HAVE_CONVERT
	convert_t * conv;
#endif
};

// struct file_core_t divxfile_core;


void divxfile_core_usage()
{
	MPRINT("DivX ;-) core options:\n");
	MPRINT("\t-vb <divx bitrate>        [1800]\n");
	MPRINT("\t-vq <divx quality>        [80]\n");
	MPRINT("\t-dk <divx keyframe rate>  [unchanged]\n");
	MPRINT("\t-ab <mp3 bitrate>         [depends on audio format]\n");
	MPRINT("\t-c <no of lines to clip>  [0]\n");
#ifdef HAVE_CONVERT
	MPRINT("\t-vhook \"module name and params\"  [NONE]\n");
#endif
	MPRINT("\t-o <output filename>      [NONE]\n");
}

struct file_core_t * divxfile_core_init(char ** args, unsigned int format, ntime_t vrate)
{
	struct file_core_t * fct;
	struct divxfile_core_t * divxc;
	double framerate;
 	int arate = -1;
 	int abits = -1;
 	int achans = -1;
 	char *fname = NULL;
 	int divxbitrate = 1800;
 	int mp3bitrate = -1;
 	int divxquality = 8000;
 	int divxkeyrate = -1;
 	int fccHandler;
 	int codec = fccDIV4;
 	BITMAPINFOHEADER bh;
 	WAVEFORMATEX fmt;
 	char *vhook = NULL;
 	NV_ARGS;
 	
 	fct = (struct file_core_t *)malloc(sizeof(struct file_core_t));
 	if(!fct) {
 		PERROR("Could no malloc new file core");
 		return NULL;
 	}
 	divxc = (struct divxfile_core_t *)malloc(sizeof(struct divxfile_core_t));
 	if(!divxc) {
 		PERROR("Could no malloc new divx core");
 		free(fct);
 		return NULL;
 	} else {
		/* initialise values */
		divxc->clip=0;
	}
 	memcpy(fct, &divxfile_core, sizeof(struct file_core_t));
 	fct->priv = (void *)divxc;
 	
 	if(GetAvifileVersion()!=AVIFILE_VERSION) {
 		EPRINT("Compiled for wrong avifile version!\n");
 		goto error0;
 	}
 	
	NV_SCAN_ARGS{
		if(strcmp(NV_THIS_ARG, "-r") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -r.\n");
				goto usage;
			}
			arate = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-b") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -b.\n");
				goto usage;
			}
			abits = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-s") == 0) {
			achans = 2;
		} else if(strcmp(NV_THIS_ARG, "-m") == 0) {
			achans = 1;
		} else if(strcmp(NV_THIS_ARG, "-w") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -w.\n");
				goto usage;
			}
			divxc->width = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-h") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -h.\n");
				goto usage;
			}
			divxc->height = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-vb") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -vb.\n");
				goto usage;
			}
			divxbitrate = atoi(NV_THIS_ARG);
			if(divxbitrate < 0) {
				EPRINT("Invalid argument to -vb (>=0).\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-vq") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -vq.\n");
				goto usage;
			}
			divxquality = atoi(NV_THIS_ARG);
			if((divxquality < 0) ||
			   (divxquality > 100)) {
				EPRINT("Invalid argument to -vq (0..100).\n");
				goto usage;
			}
			divxquality *= 100;
		} else if(strcmp(NV_THIS_ARG, "-dk") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -dk.\n");
				goto usage;
			}
			divxkeyrate = atoi(NV_THIS_ARG);
			if(divxkeyrate < 0) {
				EPRINT("Invalid argument to -dk (>=0).\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-ab") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -ab.\n");
				goto usage;
			}
			mp3bitrate = atoi(NV_THIS_ARG);
			if(mp3bitrate < 0) {
				EPRINT("Invalid argument to -ab (>=0).\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-c") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -c.\n");
				goto usage;
			}
			divxc->clip = atoi(NV_THIS_ARG);
			if(divxc->clip < 0) {
				EPRINT("Invalid argument to -c (>=0).\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-o") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -o.\n");
				goto usage;
			}
			fname = NV_THIS_ARG;
		} else if(strcmp(NV_THIS_ARG, "-vr") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -vr.\n");
				goto usage;
			}
			vrate = (unsigned long long)(1e9 / atof(NV_THIS_ARG));
		} else if(strcmp(NV_THIS_ARG, "-vhook") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -vhook.\n");
				goto usage;
			}
			vhook = NV_THIS_ARG;
		}
	}
	if(!fname) {
		EPRINT("Missing -o <filename> argument.\n");
		goto usage;
	}
	IPRINT("File: %s\n", fname);
	
 	if(format != NV_RGB24) {
#ifdef HAVE_CONVERT
		IPRINT("No native format found - initialise format conversion\n");
		divxc->conv = convert_new(format, NV_RGB24, divxc->width, divxc->height, vhook);
		if(!divxc->conv) {
			EPRINT("Could not init format conversion core!\n");
			goto error0;
		}
#else
		EPRINT("No native format found, and conversion not available - giving up.\n");
		goto error0;
#endif 	
 	}

#ifdef HAVE_CONVERT
	if((!divxc->conv) && (vhook)) {
		IPRINT("Instantiating conversion function for vhooks...\n");
 		divxc->conv = convert_new(format, format, divxc->width, divxc->height, vhook);
 		if(!divxc->conv) {
 			EPRINT("Could not init format conversion core!\n");
 			goto error0;
 		}
	}
#endif

	divxc->ashift = (abits == 16) ? 1 : 0;
	
	unlink(fname);
	divxc->avifile = CreateIAviWriteFile(fname);
#ifdef AVIFILE_NEW_API
	SetCodecAttr(*CodecInfo::match(codec, CodecInfo::Video,0),"BitRate", divxbitrate);
#else
	IVideoEncoder::SetExtendedAttr(codec, "BitRate", divxbitrate);
#endif
	fccHandler = codec;
	framerate = 1e9 / (double)vrate;
	
	divxc->framebuffer = (unsigned char *)malloc(divxc->width * divxc->height * 4);
	if(!divxc->framebuffer) {
		EPRINT("Could not allocate memory for framebuffer\n");
		goto error0;
	} 
	
	memset(&bh, 0, sizeof(BITMAPINFOHEADER));
	bh.biSize = sizeof(BITMAPINFOHEADER);
	bh.biWidth = divxc->width;
	bh.biHeight = divxc->height - (divxc->clip * 2);
	bh.biPlanes = 1;
	bh.biBitCount = 24;
	bh.biCompression = 0;
	bh.biSizeImage = divxc->width * divxc->height * 3;
	divxc->im=new CImage((BitmapInfo*)&bh, divxc->framebuffer, false);
	divxc->vstream = divxc->avifile->AddVideoStream(fccHandler, &bh, (int)(1000000.0/framerate));
	divxc->vstream->SetQuality(divxquality);
	if(divxkeyrate >= 0)
		divxc->vstream->SetKeyFrame(divxkeyrate);
	divxc->vstream->Start();
	
	memset(&fmt,0,sizeof(WAVEFORMATEX));
	fmt.wFormatTag = 1;
	fmt.nChannels = achans;
	fmt.nSamplesPerSec = arate;
	fmt.nAvgBytesPerSec = achans * arate; // * (abits/8);
	fmt.nBlockAlign = ((achans > 1) ? 2 : 1);
	fmt.wBitsPerSample = abits;
	fmt.cbSize = sizeof(WAVEFORMATEX);
	if(mp3bitrate == -1) switch(arate) {
		case 48000: mp3bitrate = (achans > 1 ? 160 : 80); break;
		case 44100: mp3bitrate = (achans > 1 ? 112 : 56); break;
		case 22050: mp3bitrate = (achans > 1 ? 64 : 32); break;
		default: EPRINT("unknown sample rate %d Hz, please specify MP3 bitrate\n", arate);
			goto error0;
	}
	divxc->astream = divxc->avifile->AddAudioStream(0x55, &fmt, (mp3bitrate*1000)/8);
	divxc->astream->Start();
	return fct;

error0:
#ifdef HAVE_CONVERT
	if(divxc->conv) {
		convert_end(divxc->conv);
	}
#endif
	free(fct);
	free(divxc);
	return NULL;
usage:
	EPRINT("Error parsing options.  Please run without any command line options for\n");
	EPRINT("detailed usage information.\n");
	free(fct);
	free(divxc);
	return NULL;
}

void divxfile_core_close(struct file_core_t * fct)
{
	struct divxfile_core_t * divxc = (struct divxfile_core_t *)fct->priv;
	
	if(!divxc)
		return;
	divxc->vstream->Stop();
//	divxc->astream->Stop();
	delete divxc->avifile;
#ifdef HAVE_CONVERT
	if(divxc->conv) {
		convert_end(divxc->conv);
	}
#endif
	free(divxc);
	free(fct);
}

void divxfile_core_nullvideo(struct file_core_t * fct)
{
	struct divxfile_core_t * divxc = (struct divxfile_core_t *)fct->priv;
	if(!divxc)
		return;
	divxc->vstream->AddFrame(0);
}

void divxfile_core_video(struct file_core_t * fct, unsigned char **buf)
{
	struct divxfile_core_t * divxc = (struct divxfile_core_t *)fct->priv;
	int i;
	unsigned char * b;
	
	if(!divxc)
		return;
#ifdef HAVE_CONVERT
	if(divxc->conv) {
		buf = convert_convert(divxc->conv, buf);
		if(!buf) {
			EPRINT("Error converting image!\n");
			return;
		}
	}
#endif
	b = divxc->framebuffer;
	for(i = ((divxc->height - divxc->clip) - 1); i >= divxc->clip; i--) {
		memcpy(b, buf[i], divxc->width * 3);
		b += divxc->width * 3;
	}
	divxc->vstream->AddFrame(divxc->im);
}

void divxfile_core_audio(struct file_core_t * fct, unsigned char *buf, int n)
{
	struct divxfile_core_t * divxc = (struct divxfile_core_t *)fct->priv;
	if(!divxc)
		return;
	divxc->astream->AddData(buf, n >> divxc->ashift);
}

void divxfile_core_sync(struct file_core_t * fct)
{
	/* FIXME - find out how to do this! */
	return;
}

static unsigned int divxfile_core_formats[] = {NV_RGB24, NV_YUV422, NV_YUV420, NV_END_FORMAT};
//static unsigned int divxfile_core_formats[] = {NV_YUV422, NV_YUV420, NV_END_FORMAT};

inline unsigned int *divxfile_core_format(char **args)
{
	return divxfile_core_formats;
}

struct file_core_t divxfile_core = {
	file_core_init: divxfile_core_init,
	file_core_close: divxfile_core_close,
	file_core_nullvideo: divxfile_core_nullvideo,
	file_core_fastvideo: divxfile_core_video,
	file_core_video: divxfile_core_video,
	file_core_audio: divxfile_core_audio,
	file_core_sync: divxfile_core_sync,
	file_core_format: divxfile_core_format,
	file_core_overrides: NULL,
	file_core_usage: divxfile_core_usage,
	priv: NULL
};
