/* 
   ffmpegfile_core.c (C) Justin Schoeman 2001 (justin@suntiger.ee.up.ac.za)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
*/
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include "avformat.h"
#include "nvrec.h"

struct ffmpegfile_core_t {
	AVFormatContext * oc;
	AVPicture pic;
	int video_stream;
	int audio_stream;
	unsigned char video_buffer[128*1024];
	unsigned char * audio_buffer;
	unsigned char * audio_out;
	int abuf_used;
	int frame_bytes;
};

void ffmpegfile_core_usage()
{
	MPRINT("FFMPEG file (mpeg-1) core usage:\n");
	MPRINT("\t-o <output filename>                     [NONE]\n");
	MPRINT("\t-ff <file format \"help\" for list>        [from ext]\n");
	MPRINT("\t-vc <video codec \"help\" for list>        [from ext]\n");
	MPRINT("\t-vb <kbits per second for video>         [1800]\n");
	MPRINT("\t-vg <video gop size - 0 for I frame only>[25]\n");
	MPRINT("\t-vq <fixed q scale 0..100>               [variable]\n");
	MPRINT("\t     * remember - use 0 for VBR!\n");
	MPRINT("\t-ac <audio codec \"help\" for list>        [from ext]\n");
	MPRINT("\t-ab <kbits per second for audio>         [80]\n");
	MPRINT("\t-o <output filename - may not contain :> [NONE]\n");
}

struct file_core_t ffmpegfile_core;

void ffmpegfile_core_novideo(struct file_core_t * fct, unsigned char **buf)
{
}

void ffmpegfile_core_nonullvideo(struct file_core_t * fct)
{
}

void ffmpegfile_core_noaudio(struct file_core_t * fct, unsigned char *buf, int n)
{
}

struct file_core_t * ffmpegfile_core_init(char ** args, unsigned int format, ntime_t vrate)
{
	struct file_core_t * fct;
	struct ffmpegfile_core_t * ffmpegc;
	AVFormatContext * oc;
	AVStream *st = NULL;
	AVCodec *codec;
	int use_video, use_audio;
	int nb_streams = 0;
	int i;
 	int arate = -1;
 	int abits = -1;
 	int achans = 1;
 	int width = -1;
 	int height = -1;
 	char *fname = NULL;
 	char filename[256];
 	char *ff = NULL;
 	char *vc = NULL;
 	char *ac = NULL;
 	int abr = 80000;
 	int vbr = 1800000;
 	int vbrt = 1800000;
 	int vg = 25;
 	int vq = -1;
 	NV_ARGS;
 	
 	if(format != NV_YUV420) {
 		EPRINT("Invalid format!\n");
 		return NULL;
 	}
 	
 	register_all();
 	
 	fct = (struct file_core_t *)malloc(sizeof(struct file_core_t));
 	if(!fct) {
 		PERROR("Could no malloc new file core");
 		return NULL;
 	}
 	ffmpegc = (struct ffmpegfile_core_t *)malloc(sizeof(struct ffmpegfile_core_t));
 	if(!ffmpegc) {
 		PERROR("Could no malloc new ffmpeg core");
 		free(fct);
 		return NULL;
 	}
 	memset(ffmpegc, 0, sizeof(struct ffmpegfile_core_t));
 	memcpy(fct, &ffmpegfile_core, sizeof(struct file_core_t));
 	fct->priv = (void *)ffmpegc;
 	
	NV_SCAN_ARGS{
		if(strcmp(NV_THIS_ARG, "-r") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -r.\n");
				goto usage;
			}
			arate = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-b") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -b.\n");
				goto usage;
			}
			abits = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-s") == 0) {
			achans = 2;
		} else if(strcmp(NV_THIS_ARG, "-m") == 0) {
			achans = 1;
		} else if(strcmp(NV_THIS_ARG, "-w") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -w.\n");
				goto usage;
			}
			width = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-h") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -h.\n");
				goto usage;
			}
			height = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-vb") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -vb.\n");
				goto usage;
			}
			vbr = atoi(NV_THIS_ARG);
			vbr *= 1000;
		} else if(strcmp(NV_THIS_ARG, "-vbt") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -vbt.\n");
				goto usage;
			}
			vbrt = atoi(NV_THIS_ARG);
			vbrt *= 1000;
		} else if(strcmp(NV_THIS_ARG, "-ab") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -ab.\n");
				goto usage;
			}
			abr = atoi(NV_THIS_ARG);
			abr *= 1000;
		} else if(strcmp(NV_THIS_ARG, "-vg") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -vg.\n");
				goto usage;
			}
			vg = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-vq") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -vq.\n");
				goto usage;
			}
			vq = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-ff") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -ff.\n");
				goto usage;
			}
			if(strcmp(NV_THIS_ARG, "help") == 0) {
				AVFormat *f;
				MPRINT("Listing FFMPEG file formats:\n");
				for(f = first_format; f != NULL; f = f->next) {
					if (f->write_header)
						MPRINT("  %s\n", f->name);
				}
				goto usage;
			}
			ff = NV_THIS_ARG;
		} else if(strcmp(NV_THIS_ARG, "-vc") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -vc.\n");
				goto usage;
			}
			if(strcmp(NV_THIS_ARG, "help") == 0) {
				AVCodec *p;
				MPRINT("Listing FFMPEG video codecs:\n");
				for(p = first_avcodec; p != NULL; p = p->next) {
					if ((p->encode) && (p->type == CODEC_TYPE_VIDEO))
						MPRINT("  %s\n", p->name);
				}
				goto usage;
			}
			vc = NV_THIS_ARG;
		} else if(strcmp(NV_THIS_ARG, "-ac") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -ac.\n");
				goto usage;
			}
			if(strcmp(NV_THIS_ARG, "help") == 0) {
				AVCodec *p;
				MPRINT("Listing FFMPEG audio codecs:\n");
				for(p = first_avcodec; p != NULL; p = p->next) {
					if ((p->encode) && (p->type == CODEC_TYPE_AUDIO))
						MPRINT(" %s\n", p->name);
				}
				goto usage;
			}
			ac = NV_THIS_ARG;
		} else if(strcmp(NV_THIS_ARG, "-o") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -o.\n");
				goto usage;
			}
			fname = NV_THIS_ARG;
		} else if(strcmp(NV_THIS_ARG, "-vr") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -vr.\n");
				goto usage;
			}
			vrate = (unsigned long long)(1e9 / atof(NV_THIS_ARG));
		}
	}
	if (!fname) {
		EPRINT("Missing -o <filename> argument.\n");
		goto usage;
	}
	if (width&15) {
		EPRINT("Width must be a multiple of 16.\n");
		goto usage;
	}
	if (height&15) {
		EPRINT("Height must be a multiple of 16.\n");
		goto usage;
	}

	if(strcmp(fname, "-") == 0) {
		snprintf(filename, 256, "pipe:");
	}
	snprintf(filename, 256, "%s", fname);
	IPRINT("File: %s\n", filename);

 	ffmpegc->oc = av_mallocz(sizeof(AVFormatContext));
 	oc = ffmpegc->oc;
 	if(ff)
		oc->format = guess_format(ff, NULL, NULL);
	else
		oc->format = guess_format(NULL, filename, NULL);
	if(!oc->format) {
		EPRINT("Could not match format for %s:%s\n", ff, fname);
		goto error0;
	}
	use_video = oc->format->video_codec != CODEC_ID_NONE;
	use_audio = oc->format->audio_codec != CODEC_ID_NONE;
	IPRINT("Got format : %s (%d,%d)\n", oc->format->name, use_video, use_audio);

	if(use_video){
		AVCodecContext *video_enc;
		
		if(vc) {
			codec = avcodec_find_decoder_by_name(vc);
			oc->format->video_codec = codec->id;
		}
		
		st = av_mallocz(sizeof(AVStream));
		if(!st){
			EPRINT("Error mallocing AVstream\n");
			goto error0;
		}
		video_enc = &st->codec;
		video_enc->codec_id = oc->format->video_codec;
		IPRINT("Video codec id %d\n", video_enc->codec_id);
		video_enc->codec_type = CODEC_TYPE_VIDEO;
		video_enc->flags = 0;
		video_enc->bit_rate = vbr;
		video_enc->frame_rate = (int)((1000000000.0/(double)vrate)*(double)FRAME_RATE_BASE);
		video_enc->width = width;
		video_enc->height = height;
		video_enc->gop_size = vg;
		if(vq > 0) {
			video_enc->quality = (vq * 16) / 100;
			if(video_enc->quality == 0)
				video_enc->quality = 1;
			if(video_enc->quality > 16)
				video_enc->quality = 16;
			video_enc->flags |= CODEC_FLAG_QSCALE;
		}
		ffmpegc->video_stream = nb_streams;
		oc->streams[nb_streams++] = st;
		ffmpegc->pic.linesize[0] = width;
		ffmpegc->pic.linesize[1] = width / 2;
		ffmpegc->pic.linesize[2] = width / 2;
	} else {
		fct->file_core_video = ffmpegfile_core_novideo;
		fct->file_core_fastvideo = ffmpegfile_core_novideo;
		fct->file_core_nullvideo = ffmpegfile_core_nonullvideo;
	}

	if(use_audio){
		AVCodecContext *audio_enc;

		if(ac) {
			codec = avcodec_find_decoder_by_name(ac);
			oc->format->audio_codec = codec->id;
		}

		st = av_mallocz(sizeof(AVStream));
		if(!st){
			EPRINT("Error mallocing AVstream\n");
			goto error0;
		}
		audio_enc = &st->codec;
		audio_enc->codec_id = oc->format->audio_codec;
		IPRINT("Audio codec id %d\n", audio_enc->codec_id);
		audio_enc->codec_type = CODEC_TYPE_AUDIO;
		audio_enc->bit_rate = abr;
		audio_enc->sample_rate = arate;
		audio_enc->channels = achans;
		ffmpegc->audio_stream = nb_streams;
		oc->streams[nb_streams++] = st;
	} else {
		fct->file_core_audio = ffmpegfile_core_noaudio;
	}

	oc->nb_streams = nb_streams;
	
	if(!nb_streams) {
		EPRINT("No audio or video selected\n");
		goto error0;
	}

	dump_format(oc, 1, filename, 1);
	
	if (url_fopen(&oc->pb, filename, URL_WRONLY) < 0) {
		EPRINT("Could not open '%s'\n", filename);
		goto error0;
	}

	for(i=0; i<nb_streams; i++){
		st = oc->streams[i];
		codec = avcodec_find_encoder(st->codec.codec_id);
		if(!codec){
			EPRINT("Could not find codec\n");
			goto error0;
		}
		if(avcodec_open(&st->codec, codec) < 0){
			EPRINT("Cannot open avencoder %s\n", codec->name);
			goto error0;
		}
	}

	if(use_audio){
		ffmpegc->frame_bytes = 2 * achans * oc->streams[ffmpegc->audio_stream]->codec.frame_size;
		ffmpegc->audio_buffer = malloc(ffmpegc->frame_bytes);
		if(!ffmpegc->audio_buffer) {
			EPRINT("Error mallocing audio_buffer\n");
			goto error0;
		}
		ffmpegc->audio_out = malloc(ffmpegc->frame_bytes);
		if(!ffmpegc->audio_out) {
			EPRINT("Error mallocing audio_out\n");
			goto error0;
		}
	}

	oc->format->write_header(oc);
	
	return fct;

error0:
	free(fct);
	free(ffmpegc);
	return NULL;
usage:
	EPRINT("Error parsing options.  Please run without any command line options for\n");
	EPRINT("detailed usage information.\n");
	free(fct);
	free(ffmpegc);
	return NULL;
}

void ffmpegfile_core_close(struct file_core_t * fct)
{
	struct ffmpegfile_core_t * ffmpegc = (struct ffmpegfile_core_t *)fct->priv;
	int i;
	
	if(!ffmpegc)
		return;
	
	for(i = 0; i < ffmpegc->oc->nb_streams; i++)
		avcodec_close(&ffmpegc->oc->streams[i]->codec);
	ffmpegc->oc->format->write_trailer(ffmpegc->oc);

	free(ffmpegc);
	free(fct);
}

void ffmpegfile_core_nullvideo(struct file_core_t * fct)
{
	struct ffmpegfile_core_t * ffmpegc = (struct ffmpegfile_core_t *)fct->priv;
	static AVPacket pkt;
	
	if(!ffmpegc)
		return;

	pkt.stream_index = ffmpegc->video_stream;
	pkt.data = 0;
	pkt.size = 0;
	av_write_packet(ffmpegc->oc, &pkt);
}

void ffmpegfile_core_video(struct file_core_t * fct, unsigned char **buf)
{
	struct ffmpegfile_core_t * ffmpegc = (struct ffmpegfile_core_t *)fct->priv;
	static AVPacket pkt;
	int ret;
	
	if(!ffmpegc)
		return;

	ffmpegc->pic.data[0] = buf[0];
	ffmpegc->pic.data[1] = buf[1];
	ffmpegc->pic.data[2] = buf[2];
	ret = avcodec_encode_video(&ffmpegc->oc->streams[ffmpegc->video_stream]->codec, 
					ffmpegc->video_buffer, 
					sizeof(ffmpegc->video_buffer), 
					&ffmpegc->pic);
	pkt.stream_index = ffmpegc->video_stream;
	pkt.data = ffmpegc->video_buffer;
	pkt.size = ret;
	av_write_packet(ffmpegc->oc, &pkt);
}

void ffmpegfile_core_audio(struct file_core_t * fct, unsigned char *buf, int n)
{
	struct ffmpegfile_core_t * ffmpegc = (struct ffmpegfile_core_t *)fct->priv;
	static AVPacket pkt;
	int c;

	if(!ffmpegc)
		return;

	while((ffmpegc->abuf_used + n) >= ffmpegc->frame_bytes) {
		c = ffmpegc->frame_bytes - ffmpegc->abuf_used;
		memcpy(ffmpegc->audio_buffer + ffmpegc->abuf_used, 
			buf,
			c);
		n -= c;
		buf += c;
		c = avcodec_encode_audio(&ffmpegc->oc->streams[ffmpegc->audio_stream]->codec,
				ffmpegc->audio_out,
				ffmpegc->frame_bytes,
				(short *)ffmpegc->audio_buffer);
		pkt.stream_index = ffmpegc->audio_stream;
		pkt.data = ffmpegc->audio_out;
		pkt.size = c;
		av_write_packet(ffmpegc->oc, &pkt);
		ffmpegc->abuf_used = 0;
	}
	if(n) {
		memcpy(ffmpegc->audio_buffer + ffmpegc->abuf_used, buf, n);
		ffmpegc->abuf_used += n;
	}
}

void ffmpegfile_core_sync(struct file_core_t * fct)
{
	return;
}

static unsigned int ffmpegfile_core_formats[] = {NV_YUV420, NV_END_FORMAT};

unsigned int *ffmpegfile_core_format(char **args)
{
	return ffmpegfile_core_formats;
}

void ffmpegfile_core_overrides(char **args)
{
	nvcore_add_arg(args, "-b");
	nvcore_add_arg(args, "16");
	return;
}

struct file_core_t ffmpegfile_core = {
	file_core_init: ffmpegfile_core_init,
	file_core_close: ffmpegfile_core_close,
	file_core_nullvideo: ffmpegfile_core_nullvideo,
	file_core_fastvideo: ffmpegfile_core_video,
	file_core_video: ffmpegfile_core_video,
	file_core_audio: ffmpegfile_core_audio,
	file_core_sync: ffmpegfile_core_sync,
	file_core_format: ffmpegfile_core_format,
	file_core_overrides: ffmpegfile_core_overrides,
	file_core_usage: ffmpegfile_core_usage,
	priv: NULL
};
