/* 
   ffmpegfile_core.c (C) Justin Schoeman 2001 (justin@suntiger.ee.up.ac.za)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
*/
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include "avformat.h"
#include "nvrec.h"
#ifdef HAVE_CONVERT
#include "convert.h"
#endif

struct ffmpegfile_core_t {
	AVFormatContext * oc;
	AVFrame *pic;
	int video_stream;
	int audio_stream;
	AVCodecContext *video_cc;
	AVCodecContext *audio_cc;
	unsigned char video_buffer[128*1024];
	unsigned char * audio_buffer;
	unsigned char * audio_out;
	int abuf_used;
	int frame_bytes;
#ifdef HAVE_CONVERT
	convert_t * conv;
#endif
};

void ffmpegfile_core_usage()
{
	MPRINT("FFMPEG file (mpeg-1) core usage:\n");
	MPRINT("\t-o <output filename>                     [NONE]\n");
	MPRINT("\t-ff <file format \"help\" for list>        [from ext]\n");
	MPRINT("\t-vc <video codec \"help\" for list>        [from ext]\n");
	MPRINT("\t-vb <kbits per second for video>         [1800]\n");
	MPRINT("\t-vg <video gop size - 0 for I frame only>[25]\n");
	MPRINT("\t-ac <audio codec \"help\" for list>        [from ext]\n");
	MPRINT("\t-ab <kbits per second for audio>         [80]\n");
	MPRINT("\t-vp <parameter> [value] set video parameter\n");
	MPRINT("\t   quality <1..31>                       [1]\n");
	MPRINT("\t-ap <parameter> [value] set audio parameter\n");
	MPRINT("\t   quality <1..31>                       [1]\n");
#ifdef HAVE_CONVERT
	MPRINT("\t-vhook \"module name and params\"          [NONE]\n");
#endif
	MPRINT("\t-o <output filename - may not contain :> [NONE]\n");
}

struct file_core_t ffmpegfile_core;

void ffmpegfile_core_novideo(struct file_core_t * fct, unsigned char **buf)
{
}

void ffmpegfile_core_nonullvideo(struct file_core_t * fct)
{
}

void ffmpegfile_core_noaudio(struct file_core_t * fct, unsigned char *buf, int n)
{
}

int ffmpeg_setparm(char **args, int vid, AVCodecContext *cc)
{
 	NV_ARGS;
	
	if(vid) {
		cc->bit_rate = 1800 * 1000;
		cc->bit_rate_tolerance = 1024 * 8;
	} else {
		cc->bit_rate = 80 * 1000;
		cc->bit_rate_tolerance = 1024 * 8;
	}
	cc->flags = 0;
	cc->sub_id = 0;
	cc->me_method = 0;
	cc->gop_size = 250;
	
	cc->qcompress = 0.5;
	cc->qblur = 0.5;
	cc->qmin = 3;
	cc->qmax = 15;
	cc->max_qdiff = 3;
	cc->max_b_frames = 0;
	cc->b_quant_factor = 2;
	cc->rc_strategy = 2;		/* Obsolete - remove ? */
	cc->b_frame_strategy = 0;	/* Obsolete - remove ? */
	
	cc->rtp_mode = 0;
	cc->rtp_payload_size = 4096;
	cc->rtp_callback = NULL;
	
	cc->workaround_bugs = 1;
	cc->strict_std_compliance = 0;
	cc->error_resilience = 1;
	if(vid) {
		cc->luma_elim_threshold = 0;
		cc->chroma_elim_threshold = 0;
	}
	cc->b_quant_offset = 0;
	
	NV_SCAN_ARGS{
#define SETINT(x, min, max) if(!NV_NEXT_ARG) return -1; cc->x = atoi(NV_THIS_ARG); cc->x=(cc->x>min)?cc->x:min; cc->x=(cc->x>max)?max:cc->x
#define VIDARG(x) if((strcmp(NV_THIS_ARG, #x) == 0) && (vid))
#define AUDARG(x) if((strcmp(NV_THIS_ARG, #x) == 0) && (!vid))
#define ARG(x) if(strcmp(NV_THIS_ARG, #x) == 0)
		VIDARG(-vb) {
			SETINT(bit_rate, 4, 24000);
			cc->bit_rate *= 1000;
		} else AUDARG(-ab) {
			SETINT(bit_rate, 16, 256);
			cc->bit_rate *= 1000;
		} else ARG(-vg) {
			SETINT(gop_size, 0, 1000000);
		} else VIDARG(-vp) {
			if(!NV_NEXT_ARG) return -1;
			ARG(quality) {
/* FIXME			SETINT(quality, 1, 31); */
			}
		} else AUDARG(-ap) {
			if(!NV_NEXT_ARG) return -1;
			ARG(quality) {
/* FIXME			SETINT(quality, 1, 31); */
			}
		}
	}
	return 0;
}

struct file_core_t * ffmpegfile_core_init(char ** args, unsigned int format, ntime_t vrate) {
	struct file_core_t * fct;
	struct ffmpegfile_core_t * ffmpegc;
	int use_video = 0, use_audio = 0;
	int nb_streams = 0;
	int i;
 	int arate = -1;
 	int abits = -1;
 	int achans = 1;
 	int width = -1;
 	int height = -1;
 	char *fname = NULL;
 	char filename[256];
 	char *ff = NULL;
 	char *vc = NULL;
 	char *ac = NULL;
 	char *vhook = NULL;
 	NV_ARGS;
 	
 	av_register_all();
 	
 	fct = (struct file_core_t *)malloc(sizeof(struct file_core_t));
 	if(!fct) {
 		PERROR("Could no malloc new file core");
 		return NULL;
 	}
 	ffmpegc = (struct ffmpegfile_core_t *)malloc(sizeof(struct ffmpegfile_core_t));
 	if(!ffmpegc) {
 		PERROR("Could no malloc new ffmpeg core");
 		free(fct);
 		return NULL;
 	}
 	memset(ffmpegc, 0, sizeof(struct ffmpegfile_core_t));
 	memcpy(fct, &ffmpegfile_core, sizeof(struct file_core_t));
 	fct->priv = (void *)ffmpegc;
 	
	NV_SCAN_ARGS{
		if(strcmp(NV_THIS_ARG, "-r") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -r.\n");
				goto usage;
			}
			arate = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-b") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -b.\n");
				goto usage;
			}
			abits = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-s") == 0) {
			achans = 2;
		} else if(strcmp(NV_THIS_ARG, "-m") == 0) {
			achans = 1;
		} else if(strcmp(NV_THIS_ARG, "-w") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -w.\n");
				goto usage;
			}
			width = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-h") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -h.\n");
				goto usage;
			}
			height = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-ff") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -ff.\n");
				goto usage;
			}
			if(strcmp(NV_THIS_ARG, "help") == 0) {
				AVOutputFormat *f;
				MPRINT("Listing FFMPEG file formats:\n");
				for(f = first_oformat; f; f = f->next) {
					MPRINT("  %s\n", f->name);
				}
				goto usage;
			}
			ff = NV_THIS_ARG;
		} else if(strcmp(NV_THIS_ARG, "-vc") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -vc.\n");
				goto usage;
			}
			if(strcmp(NV_THIS_ARG, "help") == 0) {
				AVCodec *p;
				MPRINT("Listing FFMPEG video codecs:\n");
				for(p = first_avcodec; p; p = p->next) {
					if ((p->encode) && (p->type == CODEC_TYPE_VIDEO))
						MPRINT("  %s\n", p->name);
				}
				goto usage;
			}
			vc = NV_THIS_ARG;
		} else if(strcmp(NV_THIS_ARG, "-ac") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -ac.\n");
				goto usage;
			}
			if(strcmp(NV_THIS_ARG, "help") == 0) {
				AVCodec *p;
				MPRINT("Listing FFMPEG audio codecs:\n");
				for(p = first_avcodec; p; p = p->next) {
					if ((p->encode) && (p->type == CODEC_TYPE_AUDIO))
						MPRINT("  %s\n", p->name);
				}
				goto usage;
			}
			ac = NV_THIS_ARG;
		} else if(strcmp(NV_THIS_ARG, "-o") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -o.\n");
				goto usage;
			}
			fname = NV_THIS_ARG;
		} else if(strcmp(NV_THIS_ARG, "-vr") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -vr.\n");
				goto usage;
			}
			vrate = (unsigned long long)(1e9 / atof(NV_THIS_ARG));
		} else if(strcmp(NV_THIS_ARG, "-vhook") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -vhook.\n");
				goto usage;
			}
			vhook = NV_THIS_ARG;
		}
	}
	if(!fname) {
		EPRINT("Missing -o <filename> argument.\n");
		goto usage;
	}
	if((width&15) || (height&15)) {
		EPRINT("Width and height must be multiples of 16.\n");
		goto usage;
	}

 	if(format != NV_YUV420) {
#ifdef HAVE_CONVERT
 		IPRINT("No native format found - initialise format conversion\n");
 		ffmpegc->conv = convert_new(format, NV_YUV420, width, height, vhook);
 		if(!ffmpegc->conv) {
 			EPRINT("Could not init format conversion core!\n");
 			goto usage;
 		}
#else
		EPRINT("No native format found, and conversion not available - giving up.\n");
		goto usage;
#endif
 	}

#ifdef HAVE_CONVERT
	if((!ffmpegc->conv) && (vhook)) {
		IPRINT("Instantiating conversion function for vhooks...\n");
 		ffmpegc->conv = convert_new(format, format, width, height, vhook);
 		if(!ffmpegc->conv) {
 			EPRINT("Could not init format conversion core!\n");
 			goto usage;
 		}
	}
#endif
 	
	if(strcmp(fname, "-") == 0) {
		snprintf(filename, 256, "pipe:");
	}
	snprintf(filename, 256, "%s", fname);
	IPRINT("File: %s\n", filename);

 	ffmpegc->oc = av_mallocz(sizeof(AVFormatContext));
 	if(ff)
		ffmpegc->oc->oformat = guess_format(ff, NULL, NULL);
	else
		ffmpegc->oc->oformat = guess_format(NULL, filename, NULL);
	if(!ffmpegc->oc->oformat) {
		EPRINT("Could not match format for %s:%s\n", ff, fname);
		goto error0;
	}

	if((ff) && (strcmp(ff, "ffm") == 0) && (strncmp(fname, "http:", 5) == 0)) {
		int err;
		AVFormatContext *ic;
		
		IPRINT("Starting http streaming...\n");
		err = av_open_input_file(&ic, fname, NULL, FFM_PACKET_SIZE, NULL);
		if(err < 0) {
			EPRINT("Could not open input stream '%s'\n", fname);
			goto error0;
		}
		IPRINT("Requested %d streams\n", ic->nb_streams);
		if(ic->nb_streams > 2) {
			EPRINT("ffmpegrec can only stream 1 video and 1 audio stream (max 2 streams)\n");
			goto error0;
		}
		for(nb_streams = 0; nb_streams < ic->nb_streams; nb_streams++) {
			AVStream *st;
			AVCodecContext *c;
			c = &ic->streams[nb_streams]->codec;
			if(c->codec_type == CODEC_TYPE_VIDEO) {
				IPRINT("Stream %d is video\n", nb_streams);
				IPRINT(" Video frame rate %d (%d)\n", c->frame_rate, (int)((1000000000.0/(double)vrate)*(double)FRAME_RATE_BASE));
 				if(c->frame_rate != (int)((1000000000.0/(double)vrate)*(double)FRAME_RATE_BASE)) {
 					EPRINT("ffmpegrec and ffserver frame rate differ - try '-vr %.2f'\n", (float)c->frame_rate/(float)FRAME_RATE_BASE);
 					goto error0;
 				}
				IPRINT(" Video width %d (%d)\n", c->width, width);
 				if(c->width != width) {
 					EPRINT("ffmpegrec and ffserver video width differ - try '-w %d'\n", c->width);
 					goto error0;
 				}
				IPRINT(" Video height %d (%d)\n", c->height, height);
 				if(c->height != height) {
 					EPRINT("ffmpegrec and ffserver video height differ - try '-h %d'\n", c->height);
 					goto error0;
 				}
				if(ffmpegc->video_cc) {
					EPRINT("Video stream already configured!\n");
					goto error0;
				} else {
					st = av_mallocz(sizeof(AVFormatContext));
					memcpy(st, ic->streams[nb_streams], sizeof(AVStream));
					ffmpegc->oc->streams[nb_streams] = st;
					ffmpegc->video_cc = &st->codec;
					ffmpegc->video_stream = nb_streams;
					use_video = 1;
				}
			} else if(c->codec_type == CODEC_TYPE_AUDIO) {
				IPRINT("Stream %d is audio\n", nb_streams);
				IPRINT(" Audio sample rate %d (%d)\n", c->sample_rate, arate);
 				if(c->sample_rate != arate) {
 					EPRINT("ffmpegrec and ffserver audio sample rate differ - try '-r %d'\n", c->sample_rate);
 					goto error0;
 				}
				IPRINT(" Audio channels %d (%d)\n", c->channels, achans);
 				if(c->channels != achans) {
 					EPRINT("ffmpegrec and ffserver audio channels differ - try '%s'\n", (c->channels==1)?"-m":"-s");
 					goto error0;
 				}
				if(ffmpegc->audio_cc) {
					EPRINT("Audio stream already configured!\n");
					goto error0;
				} else {
					st = av_mallocz(sizeof(AVFormatContext));
					memcpy(st, ic->streams[nb_streams], sizeof(AVStream));
					ffmpegc->oc->streams[nb_streams] = st;
					ffmpegc->audio_cc = &st->codec;
					ffmpegc->audio_stream = nb_streams;
					use_audio = 1;
				}
			} else {
				EPRINT("Stream %d is unknown!\n", nb_streams);
				goto error0;
			}
		}
		av_close_input_file(ic);
	} else {
	use_video = ffmpegc->oc->oformat->video_codec != CODEC_ID_NONE;
	use_audio = ffmpegc->oc->oformat->audio_codec != CODEC_ID_NONE;

	IPRINT("Got format : %s (%d,%d)\n", ffmpegc->oc->oformat->name, use_video, use_audio);

	if(use_video){
		AVCodec *codec;
		AVStream *st = av_mallocz(sizeof(AVStream));
		
		if(!st){
			EPRINT("Error mallocing AVstream\n");
			goto error0;
		}
		avcodec_get_context_defaults(&st->codec);
		
		if(vc) {
			codec = avcodec_find_encoder_by_name(vc);
			ffmpegc->oc->oformat->video_codec = codec->id;
		}

		ffmpegc->video_cc = &st->codec;
		ffmpegc->video_cc->codec_id = ffmpegc->oc->oformat->video_codec;
		IPRINT("Video codec id %d\n", ffmpegc->video_cc->codec_id);
		ffmpegc->video_cc->codec_type = CODEC_TYPE_VIDEO;
		ffmpegc->video_cc->frame_rate = (int)((1000000000.0/(double)vrate)*(double)FRAME_RATE_BASE);
		ffmpegc->video_cc->width = width;
		ffmpegc->video_cc->height = height;
		if(ffmpeg_setparm(args, 1, ffmpegc->video_cc) < 0) {
			EPRINT("Error setting ffmpeg video parameters.\n");
			goto usage;
		}
		
		ffmpegc->video_stream = nb_streams;
		ffmpegc->oc->streams[nb_streams++] = st;
	} else {
		fct->file_core_video = ffmpegfile_core_novideo;
		fct->file_core_fastvideo = ffmpegfile_core_novideo;
		fct->file_core_nullvideo = ffmpegfile_core_nonullvideo;
	}

	if(use_audio){
		AVCodec *codec;
		AVStream *st = av_mallocz(sizeof(AVStream));
		
		if(!st){
			EPRINT("Error mallocing AVstream\n");
			goto error0;
		}
		avcodec_get_context_defaults(&st->codec);
		
		if(ac) {
			codec = avcodec_find_encoder_by_name(ac);
			ffmpegc->oc->oformat->audio_codec = codec->id;
		}

		ffmpegc->audio_cc = &st->codec;
		ffmpegc->audio_cc->codec_id = ffmpegc->oc->oformat->audio_codec;
		IPRINT("Audio codec id %d\n", ffmpegc->audio_cc->codec_id);
		ffmpegc->audio_cc->codec_type = CODEC_TYPE_AUDIO;
		ffmpegc->audio_cc->sample_rate = arate;
		ffmpegc->audio_cc->channels = achans;
		ffmpegc->audio_cc->sample_fmt = SAMPLE_FMT_S16;
		if(ffmpeg_setparm(args, 0, ffmpegc->audio_cc) < 0) {
			EPRINT("Error setting ffmpeg audio parameters.\n");
			goto usage;
		}

		ffmpegc->audio_stream = nb_streams;
		ffmpegc->oc->streams[nb_streams++] = st;
	} else {
		fct->file_core_audio = ffmpegfile_core_noaudio;
	}
	}
	if(!nb_streams) {
		EPRINT("No audio or video selected\n");
		goto error0;
	}

	ffmpegc->oc->nb_streams = nb_streams;

	if(use_video) {
		ffmpegc->pic = avcodec_alloc_frame();
		if(!ffmpegc->pic) {
			EPRINT("Out of memory allocating picture buffers\n");
			goto usage;
		}
		ffmpegc->pic->linesize[0] = width;
		ffmpegc->pic->linesize[1] = width / 2;
		ffmpegc->pic->linesize[2] = width / 2;
	}

	if(nvcore_verbose>NV_VERBOSE_QUIET)	
		dump_format(ffmpegc->oc, 1, filename, 1);
	
	if (url_fopen(&ffmpegc->oc->pb, filename, URL_WRONLY) < 0) {
		EPRINT("Could not open '%s'\n", filename);
		goto error0;
	}

	for(i=0; i<nb_streams; i++){
		AVCodec *codec;
		AVStream *st;

		st = ffmpegc->oc->streams[i];
		codec = avcodec_find_encoder(st->codec.codec_id);
		if(!codec){
			EPRINT("Could not find codec\n");
			goto error0;
		}
		if(avcodec_open(&st->codec, codec) < 0){
			EPRINT("Cannot open avencoder %s\n", codec->name);
			goto error0;
		}
	}

	if(use_audio){
		ffmpegc->frame_bytes = 2 * achans * ffmpegc->oc->streams[ffmpegc->audio_stream]->codec.frame_size;
		ffmpegc->audio_buffer = malloc(ffmpegc->frame_bytes);
		if(!ffmpegc->audio_buffer) {
			EPRINT("Error mallocing audio_buffer\n");
			goto error0;
		}
		ffmpegc->audio_out = malloc(ffmpegc->frame_bytes);
		if(!ffmpegc->audio_out) {
			EPRINT("Error mallocing audio_out\n");
			goto error0;
		}
	}

	av_write_header(ffmpegc->oc);
	
	return fct;

error0:
	free(fct);
	free(ffmpegc);
	return NULL;
usage:
	EPRINT("Error parsing options.  Please run without any command line options for\n");
	EPRINT("detailed usage information.\n");
	free(fct);
	free(ffmpegc);
	return NULL;
}

void ffmpegfile_core_close(struct file_core_t * fct)
{
	struct ffmpegfile_core_t * ffmpegc = (struct ffmpegfile_core_t *)fct->priv;
	int i;
	
	if(!ffmpegc)
		return;
	
	for(i = 0; i < ffmpegc->oc->nb_streams; i++)
		avcodec_close(&ffmpegc->oc->streams[i]->codec);
	av_write_trailer(ffmpegc->oc);
	
	if(ffmpegc->pic)
		free(ffmpegc->pic);
	if(ffmpegc->audio_out)
		free(ffmpegc->audio_out);
	if(ffmpegc->audio_buffer)
		free(ffmpegc->audio_buffer);
#ifdef HAVE_CONVERT
	if(ffmpegc->conv) {
		convert_end(ffmpegc->conv);
	}
#endif
	free(ffmpegc);
	free(fct);
}

void ffmpegfile_core_nullvideo(struct file_core_t * fct)
{
	struct ffmpegfile_core_t * ffmpegc = (struct ffmpegfile_core_t *)fct->priv;
	
	if(!ffmpegc)
		return;

	av_write_frame(ffmpegc->oc, ffmpegc->video_stream, NULL, 0);
}

void ffmpegfile_core_video(struct file_core_t * fct, unsigned char **buf)
{
	struct ffmpegfile_core_t * ffmpegc = (struct ffmpegfile_core_t *)fct->priv;
	int ret;
	
	if(!ffmpegc)
		return;

#ifdef HAVE_CONVERT
	if(ffmpegc->conv) {
		buf = convert_convert(ffmpegc->conv, buf);
		if(!buf) {
			EPRINT("Error converting image!\n");
			return;
		}
	}
#endif
	ffmpegc->pic->data[0] = buf[0];
	ffmpegc->pic->data[1] = buf[1];
	ffmpegc->pic->data[2] = buf[2];
	ret = avcodec_encode_video(ffmpegc->video_cc, 
		ffmpegc->video_buffer, 
		sizeof(ffmpegc->video_buffer), 
		ffmpegc->pic);
	av_write_frame(ffmpegc->oc, ffmpegc->video_stream, ffmpegc->video_buffer, ret);
}

void ffmpegfile_core_audio(struct file_core_t * fct, unsigned char *buf, int n)
{
	struct ffmpegfile_core_t * ffmpegc = (struct ffmpegfile_core_t *)fct->priv;
	int c;

	if(!ffmpegc)
		return;

	while((ffmpegc->abuf_used + n) >= ffmpegc->frame_bytes) {
		c = ffmpegc->frame_bytes - ffmpegc->abuf_used;
		memcpy(ffmpegc->audio_buffer + ffmpegc->abuf_used, 
			buf,
			c);
		n -= c;
		buf += c;
		c = avcodec_encode_audio(ffmpegc->audio_cc,
			ffmpegc->audio_out,
			ffmpegc->frame_bytes,
			(short *)ffmpegc->audio_buffer);
		av_write_frame(ffmpegc->oc, ffmpegc->audio_stream, ffmpegc->audio_out, c);
		ffmpegc->abuf_used = 0;
	}
	if(n) {
		memcpy(ffmpegc->audio_buffer + ffmpegc->abuf_used, buf, n);
		ffmpegc->abuf_used += n;
	}
}

void ffmpegfile_core_sync(struct file_core_t * fct)
{
	return;
}

static unsigned int ffmpegfile_core_formats[] = {NV_YUV420, NV_YUV422, NV_RGB24, NV_END_FORMAT};
//static unsigned int ffmpegfile_core_formats[] = {NV_RGB24, NV_END_FORMAT};
//static unsigned int ffmpegfile_core_formats[] = {NV_YUV422, NV_END_FORMAT};

unsigned int *ffmpegfile_core_format(char **args)
{
	return ffmpegfile_core_formats;
}

void ffmpegfile_core_overrides(char **args)
{
	nvcore_add_arg(args, "-b");
	nvcore_add_arg(args, "16");
	return;
}

struct file_core_t ffmpegfile_core = {
	file_core_init: ffmpegfile_core_init,
	file_core_close: ffmpegfile_core_close,
	file_core_nullvideo: ffmpegfile_core_nullvideo,
	file_core_fastvideo: ffmpegfile_core_video,
	file_core_video: ffmpegfile_core_video,
	file_core_audio: ffmpegfile_core_audio,
	file_core_sync: ffmpegfile_core_sync,
	file_core_format: ffmpegfile_core_format,
	file_core_overrides: ffmpegfile_core_overrides,
	file_core_usage: ffmpegfile_core_usage,
	priv: NULL
};
