/* 
   nvrec.c (C) Justin Schoeman 2001 (justin@suntiger.ee.up.ac.za)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
*/
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>

#include "nvrec.h"

#ifdef NUV
#define FILE_CORE nuvfile_core
#define CONFIG_STRING "nuppelrec"
#endif
#ifdef DIVX
#define FILE_CORE divxfile_core
#define CONFIG_STRING "divxrec"
#endif
#ifdef DIVX4
#define FILE_CORE divx4file_core
#define CONFIG_STRING "divx4rec"
#endif
#ifdef QT
#define FILE_CORE qtfile_core
#define CONFIG_STRING "qtrec"
#endif
#ifdef RTE
#define FILE_CORE rtefile_core
#define CONFIG_STRING "rterec"
#endif
#ifdef FFMPEG
#define FILE_CORE ffmpegfile_core
#define CONFIG_STRING "ffmpegrec"
#endif

int main(int argc, char **argv){
	struct cap_core_t * cct;
	struct aud_core_t * audt;
	struct file_core_t * fct;
	char **args;
	unsigned int *formats;
	
	NV_MODULE;
	NV_USAGE(argc, argv, oss_core, v4lgen_core, FILE_CORE);

	if(!(args = nvcore_args(argc, argv, &v4lgen_core, &oss_core, &FILE_CORE, CONFIG_STRING))) {
		PERROR("nvrec: parsing arguments");
		return -1;
	}
	formats = FILE_CORE.file_core_format(args);
	if(!formats) {
		EPRINT("Could not get a valid capture format for this codec!\n");
		return -1;
	}
	if(!(cct = v4lgen_core.cap_core_init(args, &formats))) {
		PERROR("nvrec: initialising capture device");
		return -1;
	}
	if(!(audt = oss_core.aud_core_init(args, cct->cap_core_ftime(cct)))) {
		PERROR("nvrec: initialising audio device");
		return -1;	
	}
	if(!(fct = FILE_CORE.file_core_init(args, *formats, cct->cap_core_ftime(cct)))) {
		PERROR("nvrec: opening output file");
		return -1;	
	}
	
	nvcore_main(args, cct, audt, fct, 0, *formats);

	DXPRINT("closing file core\n");
	fct->file_core_close(fct);
	DXPRINT("closing capture core\n");
	cct->cap_core_close(cct);
	DXPRINT("closing audio core\n");
	audt->aud_core_close(audt);
	DXPRINT("closing nvrec core\n");
	nvcore_close(args);
	DXPRINT("restoring signal handlers\n");
	nvcore_sigrestore();
	DXPRINT("leave\n");
	return 0;
}
