#include <inttypes.h>
#include <string.h>
#include <sys/time.h>
#include <unistd.h>
#include <errno.h>
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#define NV_RGB24 0
#define NV_YUV420 1
#define NV_YUV422 2
#define NV_END_FORMAT 65536

#define NV_ARGS char ** arg
#define NV_SCAN_ARGS for(arg = args + 1; *arg; arg++)
#define NV_NEXT_ARG (*(++arg))
#define NV_THIS_ARG (*arg)
#define NV_USAGE(argc, argv, a, v, f) \
	if((argc<2)||(!strcmp(argv[argc-1],"-h"))||(!strcmp(argv[argc-1],"--help"))) \
	{\
		fprintf(stderr, "Usage: %s [options]\n", argv[0]); \
		nvcore_usage(); \
		a.aud_core_usage(); \
		v.cap_core_usage(); \
		f.file_core_usage(); \
		return -1; \
	}
#define NV_MODULE fprintf(stderr, "Starting " __FILE__ ", version " VERSION " (" PACKAGE "-" VERSION ").\n")
#define NV_MAX_ARGS 256
#define NV_CONF_FILE "/etc/nvrec.conf"
#define NV_HOME_CONF "/.nvrec.conf"

typedef int64_t ntime_t;	// times in nanosec;

static inline ntime_t getntime()
{
	struct timeval tv;
	ntime_t t;
	gettimeofday(&tv, NULL);
	t = ((ntime_t)tv.tv_sec * 1000000000LL) +
		((ntime_t)tv.tv_usec * 1000);
	return t;
}

struct cap_core_t {
	struct cap_core_t * (* cap_core_init)(char ** args, unsigned int **pfmt);
	void (* cap_core_close)(struct cap_core_t * cct);
	unsigned char ** (* cap_core_next)(struct cap_core_t * cct, ntime_t * ftime);
	double (* cap_core_status)(struct cap_core_t * cct);
	ntime_t (* cap_core_ftime)(struct cap_core_t * cct);
	void (* cap_core_overrides)(char **args);
	void (* cap_core_usage)();
	void * priv;
};

struct aud_core_t {
	struct aud_core_t * (* aud_core_init)(char ** args, ntime_t ftime);
	void (* aud_core_close)(struct aud_core_t * audt);
	ntime_t (* aud_core_tailtime)(struct aud_core_t * audt);
	void (* aud_core_droptail)(struct aud_core_t * audt);
	int (* aud_core_read)(struct aud_core_t * audt, ntime_t t, int drate, unsigned char * buf, int n);
	unsigned long long (* aud_core_bps)(struct aud_core_t * audt);
	void (* aud_core_overrides)(char **args);
	void (* aud_core_usage)();
	void * priv;
};

struct file_core_t {
	struct file_core_t * (* file_core_init)(char ** args, unsigned int format, ntime_t vrate);
	void (* file_core_close)(struct file_core_t * fct);
	void (* file_core_nullvideo)(struct file_core_t * fct);
	void (* file_core_fastvideo)(struct file_core_t * fct, unsigned char **buf);
	void (* file_core_video)(struct file_core_t * fct, unsigned char **buf);
	void (* file_core_audio)(struct file_core_t * fct, unsigned char *buf, int n);
	void (* file_core_sync)(struct file_core_t * fct);
	unsigned int *(* file_core_format)(char **args);
	void (* file_core_overrides)(char **args);
	void (* file_core_usage)();
	void * priv;
};

extern struct cap_core_t v4l2_core;
extern struct cap_core_t v4l1_core;
extern struct cap_core_t v4lgen_core;
extern struct aud_core_t oss_core;
extern struct file_core_t qtfile_core;
extern struct file_core_t divxfile_core;
extern struct file_core_t nuvfile_core;
extern struct file_core_t rtefile_core;
extern struct file_core_t divx4file_core;
extern struct file_core_t ffmpegfile_core;

extern int nvcore_main(char **args, struct cap_core_t * cct, struct aud_core_t * audt, struct file_core_t * fct, ntime_t tofs, unsigned int vfmt);
extern void nvcore_add_arg(char **args, char *newarg);
extern char **nvcore_args(int argc, char **argv, struct cap_core_t * cct, struct aud_core_t * audt, struct file_core_t * fct, char* prog);
	
extern void nvcore_close(char ** args);
extern void nvcore_sigrestore();
extern void nvcore_usage();

#define NV_VERBOSE_QUIET 0x0
#define NV_VERBOSE_INFO  0x1
#define NV_VERBOSE_DEBUG 0x2
#define NV_VERBOSE_DEBUG_EXIT 0x100

extern int nvcore_verbose;

#define FORMAT_PRINT(format, args...) fprintf(stderr, "%s(%d): " format, __FUNCTION__, __LINE__ ,## args)
#define DPRINT(format, args...) if(nvcore_verbose>NV_VERBOSE_INFO) FORMAT_PRINT(format ,## args)
#define DXPRINT(format, args...) if(nvcore_verbose==NV_VERBOSE_DEBUG_EXIT) FORMAT_PRINT(format ,## args)
#define IPRINT(format, args...) if(nvcore_verbose>NV_VERBOSE_QUIET) FORMAT_PRINT(format ,## args)
#define EPRINT(format, args...) FORMAT_PRINT(format ,## args)
#define MPRINT(format, args...) fprintf(stderr, format ,## args)
#define PERROR(string) FORMAT_PRINT(string ": %s\n", strerror(errno))
