/* 
   qtplay.c (C) Justin Schoeman 2000 (justin@suntiger.ee.up.ac.za)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
*/

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <errno.h>
#include <unistd.h>
#include <SDL/SDL.h>
#include <signal.h>
#include <sys/time.h>
#include <quicktime/quicktime.h>
#include <quicktime/colormodels.h>
#include <linux/soundcard.h>
#include <sys/ioctl.h>
#include <fcntl.h>

/* quicktime */
quicktime_t *qtfile;

SDL_Surface *video_screen;

void dsig(int i){};
struct itimerval video_timer;
struct sigaction video_sig={dsig, 0, SA_RESTART|SA_NOMASK, NULL};

int audio_dsp;

void audio_init(int speed, int bits, int channels)
{
	int f;

	audio_dsp=open("/dev/dsp", O_WRONLY);
	if(audio_dsp<=0)
	{
		perror("dsp open");
		exit(-1);
	}
	f=0x00800008U;
	if (ioctl(audio_dsp, SNDCTL_DSP_SETFRAGMENT, &f)<0)
	{
		perror("setfrag");
		exit(-1);
	}
	f=bits;
	if (ioctl(audio_dsp, SNDCTL_DSP_SAMPLESIZE, &f)<0)
	{
		perror("samplesize");
		exit(-1);
	}
	f=channels-1;
	if (ioctl(audio_dsp, SNDCTL_DSP_STEREO, &f)<0)
	{
		perror("stereo");
		exit(-1);
	}
	f=speed;
	ioctl(audio_dsp, SNDCTL_DSP_SPEED, &f);
	fprintf(stderr, "Got audio block speed: %d\n", f);
	f = 256;
	if (ioctl(audio_dsp, SNDCTL_DSP_GETBLKSIZE, &f)<0)
	{
		perror("getblk");
		exit(-1);
	}
	fprintf(stderr, "Got audio block size: %d\n", f);
}


int main(int argc, char **argv)
{
	unsigned int cframe, mframe;
	SDL_Event event;
	int run=1;
	int w, h;
	double frate;
	struct audio_buf_info audio_info;
	int audio_samples;
	int audio_bytes;
	int audio_size = 0;
	unsigned short *audio_buffer;
	unsigned short *audio_buffer2;
	int bcm;
	unsigned char **rows;
	int i;
	unsigned char *tbuf;
	unsigned char **trows;
	int x,y;
	int arate;
	int render888=0;
	int audio_bits, audio_channels;
	unsigned long long audio_pos = 0;

	if (SDL_Init(SDL_INIT_AUDIO|SDL_INIT_VIDEO) < 0)
	{
		fprintf(stderr, "Couldn't initialize SDL: %s\n",SDL_GetError());
	        exit(1);
	}
	atexit(SDL_Quit);
	
	qtfile = quicktime_open(argv[1], 1, 0);
	if(!qtfile)
	{
		fprintf(stderr, "Could not open %s!\n", argv[1]);
		exit(-1);
	}
	h = quicktime_video_height(qtfile, 0);
	w = quicktime_video_width(qtfile, 0);
	quicktime_set_preload(qtfile, w*h*30);
	frate = quicktime_frame_rate(qtfile, 0);
	frate = 1/frate;
	frate *= 1000000.0;
	mframe = quicktime_video_length(qtfile, 0);
	if(!quicktime_supported_video(qtfile, 0))
	{
		fprintf(stderr, "Video format %s not supported.\n", quicktime_video_compressor(qtfile, 0));
		exit(0);
	}
	if(!quicktime_supported_audio(qtfile, 0))
	{
		fprintf(stderr, "Video format %s not supported.\n", quicktime_audio_compressor(qtfile, 0));
		exit(0);
	}
	fprintf(stderr, "audio: %4s video: %4s\n", quicktime_audio_compressor(qtfile, 0), quicktime_video_compressor(qtfile, 0));
	if ((video_screen=SDL_SetVideoMode(w, h, 0, SDL_HWSURFACE)) < 0)
	{
		fprintf(stderr, "Couldn't set video mode: %s\n",SDL_GetError());
		exit(1);
	}

	rows = (unsigned char **)malloc(h*sizeof(unsigned char *));
	for(i=0; i<h; i++)
	{
		rows[i] = (unsigned char *)(video_screen->pixels+i*video_screen->pitch);
	}
	switch (video_screen->format->BytesPerPixel)
	{
		case 1: bcm = BC_RGB8;
			break;
		case 2: bcm = BC_RGB565;
			break;
		case 3: bcm = BC_BGR888;
			break;
		case 4: bcm = BC_BGR8888;
			break;
		default: fprintf(stderr, "Unsupported display depth\n");
			exit(-1);
			break;
	}
	if(!quicktime_reads_cmodel(qtfile, bcm, 0))
	{
		fprintf(stderr, "Unsupported color model - falling back to 888 conversion.\n");
		render888=1;
		tbuf = malloc(w*h*3);
		trows = (unsigned char **)malloc(h*sizeof(unsigned char *));
		for(i=0; i<h; i++)
			trows[i] = (tbuf+w*3*i);
	}
	arate = quicktime_sample_rate(qtfile, 0);
	audio_bits = quicktime_audio_bits(qtfile, 0);
	audio_channels = quicktime_track_channels(qtfile, 0);
	fprintf(stderr, "audio channels: %d\n", audio_channels);
	audio_init(arate, 16, audio_channels);
	audio_samples = (int)((float)arate/quicktime_frame_rate(qtfile, 0));
	audio_bytes = audio_samples * 2 * audio_channels;
	fprintf(stderr, "audio spf: %d\n", audio_samples);
	audio_buffer = malloc(audio_bytes);
	if(!audio_buffer)
	{
		perror("malloc audio buffer");
		exit(-1);
	}
	audio_buffer2 = malloc(audio_bytes);
	if(!audio_buffer2)
	{
		perror("malloc audio buffer2");
		exit(-1);
	}
	video_timer.it_interval.tv_sec=0;
	video_timer.it_interval.tv_usec=frate;
	video_timer.it_value.tv_sec=0;
	video_timer.it_value.tv_usec=frate;
	if(setitimer(ITIMER_REAL, &video_timer, NULL)!=0)
	{
		perror("setitimer");
	        exit(-1);
	}
        if(sigaction(SIGALRM, &video_sig, NULL)!=0)
	{
		fprintf(stderr, "Could not capture alarm signal\n");
	        exit(-1);
	}
                  
        cframe = 0;
	while((run)&&(cframe<mframe))
	{
		if(SDL_PollEvent(&event))
		{
			switch (event.type)
			{
				case SDL_QUIT:
				case SDL_KEYDOWN:
					run=0;
			}
		}
		sleep(1);
		if(audio_channels > 1) {
			quicktime_set_audio_position(qtfile, audio_pos, 0);
			quicktime_decode_audio(qtfile, audio_buffer2, NULL, audio_samples, 0);
			quicktime_set_audio_position(qtfile, audio_pos, 0);
			quicktime_decode_audio(qtfile, audio_buffer2 + audio_samples, NULL, audio_samples, 1);
			for(i = 0; i < audio_samples; i++){
				audio_buffer[i * 2] = audio_buffer2[i];
				audio_buffer[i * 2 + 1] = audio_buffer2[audio_samples + i];
			}
		} else {
			quicktime_set_audio_position(qtfile, audio_pos, 0);
			quicktime_decode_audio(qtfile, audio_buffer, NULL, audio_samples, 0);
		}
		audio_pos += audio_samples;
		if (ioctl(audio_dsp, SNDCTL_DSP_GETOSPACE, &audio_info)<0)
		{
			perror("getospace");
			exit(-1);
		}
		if(cframe==0)
		{
			audio_size = audio_info.bytes;
		}
		if((audio_size-audio_info.bytes)<audio_bytes)
			write(audio_dsp, audio_buffer, audio_bytes);
		if((audio_size-audio_info.bytes)<(audio_bytes*2))
			write(audio_dsp, audio_buffer, audio_bytes);
		
		if ( SDL_LockSurface(video_screen) < 0 )
	 	{
 			fprintf(stderr, "Couldn't lock the display surface: %s\n",
 	        		SDL_GetError());
	 	        exit(2);
 		} 
 		if(!render888)
 		{
			quicktime_decode_scaled(qtfile, 0, 0, w, h, w, h, bcm, rows, 0);
		} else
		{
			quicktime_decode_scaled(qtfile, 0, 0, w, h, w, h, BC_RGB888, trows, 0);
			for(y=0; y<h; y++)
			{
				static int tmp;
				i=0;
				for(x=0; x<(w*3); x+=3)
				{
					switch(bcm)
					{
						case BC_BGR8888:
							rows[y][i++] = trows[y][x+2];
							rows[y][i++] = trows[y][x+1];
							rows[y][i++] = trows[y][x+0];
							rows[y][i++] = 0;
							break;
						case BC_BGR888:
							rows[y][i++] = trows[y][x+2];
							rows[y][i++] = trows[y][x+1];
							rows[y][i++] = trows[y][x+0];
							break;
						case BC_RGB565:
							tmp = (int)(trows[y][x+2]>>3);
							tmp|=(int)(((int)trows[y][x+1] >> 2) << 5);
							tmp|=(int)(((int)trows[y][x+0] >> 3) << 11);
							rows[y][i++] = tmp&0xff;
							rows[y][i++] = tmp>>8;
							break;
						case BC_RGB8:
							tmp = trows[y][x+2];
							tmp += trows[y][x+1];
							tmp += trows[y][x+0];
							rows[y][i++] = tmp/3;
							break;
					}
				}
			}
		}
		SDL_UnlockSurface(video_screen);
		SDL_UpdateRect(video_screen, 0, 0, 0, 0);
		cframe++;
	}
	SDL_Quit();
	return 0;
}
