/* tick.h - Compute successive integer multiples of a rational
 * number without long-term rounding error.
 * (c)2002 by Lennert Buytenhek <buytenh@gnu.org>
 * File licensed under the GPL, see http://www.fsf.org/ for more info.
 * Dedicated to Marija Kulikova.
 */

typedef struct {
	int value;
	int inrate;
	int outrate;
	int div;
	int mod;
} ticker_t;

inline int nvrec_gcd(long long a, long long b)
{
	long long c;
 
	while (1) {
		c = a % b;
		if (c == 0)
			return b;
		a = b;
		b = c;
	}
}

inline void nvrec_ticker_init(ticker_t *tick, long long inrate, long long outrate)
{
	int g;

	g = nvrec_gcd(inrate, outrate);
	inrate /= g;
	outrate /= g;

	tick->value = -outrate/2;

	tick->inrate = inrate;
	tick->outrate = outrate;
	tick->div = tick->outrate / tick->inrate;
	tick->mod = tick->outrate % tick->inrate;
}

inline int nvrec_ticker_tick(ticker_t *tick)
{
	int n = tick->div;

	tick->value += tick->mod;
	if (tick->value > 0) {
		tick->value -= tick->inrate;
		n++;
	}

	return n;
}
