/* 
   qtdivx.cpp (C) Justin Schoeman 2001 (justin@suntiger.ee.up.ac.za)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
*/

// qt2divx (c) 2001 Justin Schoeman
// (very heavily) based on:
// mpeg2divx
// 2000/10/23 Ulrich Hecht <uli@emulinks.de>
// last changed 2000/12/15
// licensed under the GNU Public License (GPL) v2

#define APPNAME "qt2divx"
#define APPVERSION "0.8"

#include <videoencoder.h>
#include <avifile.h>
#include <aviplay.h>
#include <version.h>
#include "../avifile_version.h"
#ifdef AVIFILE_NEW_API
#include <fourcc.h>
#include <creators.h>
#include <iostream.h>
#include <string.h>
#include <math.h>
#else
#include <aviutil.h>
#endif
#include <quicktime.h>
#include <sys/time.h>

#ifdef AVIFILE_NEW_API
using namespace Creators;
#endif

quicktime_t *qtedopen(char *fname, long * count, unsigned long long * apos)
{
	quicktime_t *f = NULL;
	int i, n, l;
	int c = 0;
	char * spos = NULL;
	char * cpos = NULL;
	for(i = 0; fname[i]; i++) {
		if(fname[i] == ':') {
			switch (c) {
				case 0: spos = fname + i + 1; break;
				case 1: cpos = fname + i + 1; break;
				default:
					fprintf(stderr, "Error parsing filename!\n");
					*count = 0;
					return NULL;
			}
			c++;
			fname[i] = 0;
		}
	}
	i = 0;
	n = 0x7fffffff;
	switch (c) {
		case 2: n = atoi(cpos);
		case 1: i = atoi(spos);
		case 0: fprintf(stderr, "Opening file: %s\n", fname);
			f = quicktime_open(fname, 1, 0);
	}
	if((i<0) || (n<0)) {
		fprintf(stderr, "Error parsing filename (%d %d)!\n", n, i);
		*count = 0;
		return NULL;
	}
	if(!f) {
		fprintf(stderr, "Error opening file (%s)!\n", fname);
		*count = 0;
		return NULL;
	}
	l = quicktime_video_length(f, 0);
	if(((i + n) > l)||(n == 0x7fffffff))
		n = l - i;
	if(i > 0) {
		double as = (double)i;
		as /= (double)quicktime_frame_rate(f, 0);
		as *= (double)quicktime_sample_rate(f, 0);
		fprintf(stderr, "Seeking to frame %d (%f)\n", i, as);
		*apos = (unsigned long long)as;

		int w = quicktime_video_width(f, 0);
		int h = quicktime_video_height(f, 0);
		unsigned char buf[w * h * 3];
		unsigned char *r[h];
		int j;
		for(j=0; j<h; j++)
			r[j] = &(buf[j*w*3]);
		while(i-- > 0)
			quicktime_decode_video(f, r, 0);
	}
	fprintf(stderr, "Encoding %d frames:\n", n);
	*count = n;
	return f;
}

int main(int argc, char** argv)
{
	if(GetAvifileVersion()!=AVIFILE_VERSION)
	{
	        cout<<"This binary was compiled for Avifile ver. "<<AVIFILE_VERSION<<", but the library is ver. "<<GetAvifileVersion()<<". Aborting."<<endl;
	        return 0;
	}

	quicktime_t* file;

	if(argc < 3)
	{
		printf("\nUsage: %s [OPTION]... [input MOV]... [output AVI]\n",argv[0]);
		printf("\nOptions:\n");
		printf("  -b\tDivX ;-) bitrate (default 1800)\n");
		printf("  -q\tDivX quality (default 80)\n");
		printf("  -k\tDivX ;-) keyframe rate (default ?)\n");
		printf("  -f\tuse fast-motion codec (default low-motion)\n");
		printf("  -a\tMP3 bitrate (kBit, default auto)\n");
                printf("  -c\tnumber of lines to clip from top/bottom (default none)\n");
		printf("  -o\toffset of clipping window (default 0 = center)\n");
                printf("  -bf\tradius for blur filter (default off)\n");
		printf("  -w\toutput width (default input width)\n");
		printf("  -h\toutput height (default input height)\n");
		printf("  -as\taudio stream number (default 0)\n");
                printf("  -vs\tvideo stream number (default 0)\n");
                printf("\n\n  Filename format: <name>:[start frame[:frame count]]\n");
		printf("\n" APPNAME " version " APPVERSION "\n");
		exit(0);
	}

	int argbase=0;
	int divxbitrate=1800;
	int divxquality=8000;
	int divxkeyrate=-1;
	int mp3bitrate=-1;
	int codec=fccDIV3;
	int cliplines=0;
	int clipoffset=0;
	int blur=0;
	int owidth=0;
	int oheight=0;
	int austream=0;
	int vistream=0;
	long frames=0; 
	unsigned long long audio_pos = 0;

	static char x1, x2;
	static unsigned int NewY1, NewX1, NewX2, Size;
	static double ScaleX, ScaleY, t1, t2, t4, f, ft;
	double PI      = 3.14159265359f;

	bool moreopts=true;

	while(moreopts)
	{
		moreopts=false;
		
		if(strcmp(argv[argbase+1],"-b")==0)
		{
			divxbitrate=atoi(argv[argbase+2]);
			argbase+=2;
			moreopts=true;
		}

		if(strcmp(argv[argbase+1],"-q")==0)
		{
			divxquality=atoi(argv[argbase+2]);
			divxquality *= 100;
			argbase+=2;
			moreopts=true;
		}

		if(strcmp(argv[argbase+1],"-k")==0)
		{
			divxkeyrate=atoi(argv[argbase+2]);
			argbase+=2;
			moreopts=true;
		}

		if(strcmp(argv[argbase+1],"-f")==0)
		{
			codec=fccDIV4;
			argbase++;
			moreopts=true;
		}

		if(strcmp(argv[argbase+1],"-a")==0)
		{
			mp3bitrate=atoi(argv[argbase+2]);
			argbase+=2;
			moreopts=true;
		}
		
		if(strcmp(argv[argbase+1],"-c")==0)
		{
			cliplines=atoi(argv[argbase+2]);
			argbase+=2;
			moreopts=true;
		}

		if(strcmp(argv[argbase+1],"-o")==0)
		{
			clipoffset=atoi(argv[argbase+2]);
			argbase+=2;
			moreopts=true;
		}

		if(strcmp(argv[argbase+1],"-bf")==0)
		{
			blur=atoi(argv[argbase+2]);
			argbase+=2;
			moreopts=true;
		}

		if(strcmp(argv[argbase+1],"-w")==0)
		{
			owidth=atoi(argv[argbase+2]);
			argbase+=2;
			moreopts=true;
		}

		if(strcmp(argv[argbase+1],"-h")==0)
		{
			oheight=atoi(argv[argbase+2]);
			argbase+=2;
			moreopts=true;
		}

		if(strcmp(argv[argbase+1],"-as")==0)
		{
			austream=atoi(argv[argbase+2]);
			argbase+=2;
			moreopts=true;
		}
		if(strcmp(argv[argbase+1],"-vs")==0)
		{
			vistream=atoi(argv[argbase+2]);
			argbase+=2;
			moreopts=true;
		}
 	}

	int numfiles=argc-argbase-2;
	char* inputfiles[numfiles];
	for(int i=argbase+1; i<argc-1; i++)
	{
		inputfiles[i-argbase-1]=argv[i];
		printf("ifiles %d assigned %s\n",i-argbase-1,argv[i]);
	}

	char* outputfile=argv[argc-1];

	file = qtedopen(inputfiles[0], &frames, &audio_pos);
	if(!quicktime_supported_video(file, 0))
	{
		printf("Input file %s has unsupported video codec.\n", inputfiles[0]);
		exit(0);
	}
	if(!quicktime_supported_audio(file, 0))
	{
		printf("Input file %s has unsupported audio codec.\n", inputfiles[0]);
		exit(0);
	}

	int astrm=quicktime_audio_tracks(file);
	austream=((austream>(astrm - 1)) ? astrm - 1 : austream);
	printf("set audio stream %d of %d\n", austream, astrm);

	int vstrm=quicktime_video_tracks(file);
	vistream=(vistream>vstrm - 1 ? vstrm - 1 : vistream);
	printf("set video stream %d of %d\n", vistream, vstrm);

	int achan=quicktime_track_channels(file, austream);
	int arate=quicktime_sample_rate(file, austream);
	printf("channels %d, rate %d\n",achan,arate);

	int width=quicktime_video_width(file, vistream);
	int height=quicktime_video_height(file, vistream);
	float framerate=quicktime_frame_rate(file, vistream);
	printf("width %d, height %d, framerate %f, frames %ld\n",width,height,framerate,frames);

	unsigned char* framebuffer = new unsigned char[width*height*3];
	unsigned char* backbuffer = new unsigned char[width*height*3];
	unsigned char** rowptr = new (unsigned char*)[height];
	unsigned char *rowbuf = new unsigned char[width*3];
	rowptr[0] = framebuffer;
	int i;
	for (i=1; i<height; i++)
		rowptr[i]=rowptr[i-1]+(width*3);

	height-=cliplines*2;
	//printf("framebuffer: %d",(unsigned int)framebuffer);
	
	short* audiobufferint = NULL;
	short* audiobuffertemp = NULL;
	int sampsperframe = 0;
	if(astrm>0)
	{
		sampsperframe = (int)(arate/framerate);
		audiobufferint = new short[sampsperframe*achan];
		if(achan > 1)
			audiobuffertemp = new short[sampsperframe*achan];
	}
	
	int fccHandler=codec;

	IAviWriteFile* avifile;

	avifile=CreateIAviWriteFile(outputfile);
	
#ifdef AVIFILE_NEW_API
	SetCodecAttr(*CodecInfo::match(codec, CodecInfo::Video,0),"BitRate", divxbitrate);
#else
	IVideoEncoder::SetExtendedAttr(codec, "BitRate", divxbitrate);
#endif

	BITMAPINFOHEADER bh;
	memset(&bh,0,sizeof(BITMAPINFOHEADER));
	bh.biSize=sizeof(BITMAPINFOHEADER);
	bh.biWidth=width;
	bh.biHeight=height;
	bh.biPlanes=1;
	bh.biBitCount=24;
	bh.biCompression=0;
	bh.biSizeImage=width*height*3;

	unsigned char* scaledbuffer = NULL;
	CImage* im;
	if(!(oheight || owidth))
		im=new CImage((BitmapInfo*)&bh, framebuffer+(cliplines-clipoffset)*width*3, false);
	else
	{
		if(oheight)
			bh.biHeight=oheight;
		if(owidth)
			bh.biWidth=owidth;
		bh.biSizeImage=bh.biHeight*bh.biWidth*3;
                            scaledbuffer= new unsigned char[bh.biHeight*bh.biWidth*3];
		im=new CImage((BitmapInfo*)&bh, scaledbuffer, false);
	}

	IAviVideoWriteStream* stream=avifile->AddVideoStream(fccHandler, &bh, (int)(1000000.0/framerate));
	stream->SetQuality(divxquality);
	if(divxkeyrate >= 0)
		stream->SetKeyFrame(divxkeyrate);
	stream->Start();
	IAviAudioWriteStream* astream = NULL;
	if(astrm>0)
	{
		WAVEFORMATEX fmt;
		memset(&fmt,0,sizeof(WAVEFORMATEX));
		fmt.wFormatTag=1;
		fmt.nChannels=(achan>2 ? 2 : achan);
		fmt.nSamplesPerSec=arate;
		fmt.nAvgBytesPerSec=(achan>1 ? arate : arate*2);
		fmt.nBlockAlign=(achan>1 ? 2 : 1);
		fmt.wBitsPerSample=16;
		fmt.cbSize=sizeof(WAVEFORMATEX);

		if(mp3bitrate == -1) switch(arate) {
			case 48000: mp3bitrate=(achan>1?160:80*achan); break;
			case 44100: mp3bitrate=(achan>1?112:56*achan); break;
			case 22050: mp3bitrate=(achan>1?64:32*achan); break;
			default: printf("unknown sample rate %d Hz, please specify MP3 bitrate\n", arate); exit(1);
		}
		astream=avifile->AddAudioStream(0x55, &fmt, (mp3bitrate*1000)/8);
		astream->Start();
	}
	int samples_read=0;

	int totalframes=0;

	struct timeval tv;
	struct timezone tz={0,0};
	gettimeofday(&tv,&tz);
	long startsec=tv.tv_sec;
	long startusec=tv.tv_usec;
	double fps=0;
	for (int j=0; j< numfiles; j++)
	{
		if(j)
		{
			quicktime_close(file);
			printf("opening MPG %s\n",inputfiles[j]);
			file = qtedopen(inputfiles[j], &frames, &audio_pos);
			if(!quicktime_supported_video(file, 0))
			{
				printf("Input file %s has unsupported video codec.\n", inputfiles[0]);
				continue;
			}
			if(!quicktime_supported_audio(file, 0))
			{
				printf("Input file %s has unsupported audio codec.\n", inputfiles[0]);
				continue;
			}
		}
		for (int i=0; i<frames ; i++, totalframes++)
		{
			//if(totalframes%10 == 0) printf("Encoding frame %d...\n",totalframes);
			gettimeofday(&tv,&tz);
			fps=totalframes/((tv.tv_sec+tv.tv_usec/1000000.0)-(startsec+startusec/1000000.0));
			printf("\rEncoding frame %d of %ld, %.1lf frames per sec    ",totalframes,frames,fps); fflush(stdout);

			int h = height + cliplines*2;
	 		if(quicktime_decode_video(file, rowptr, vistream))
			{
				printf("error reading frame %d\n",i);
				memcpy(framebuffer, backbuffer, width*h*3);
			} else {
				memcpy(backbuffer, framebuffer, width*h*3);
			}

			int x, y;
			char c;
			for(y=0; y<h/2; y++)
			{
				memcpy(rowbuf, rowptr[y], width*3);
				memcpy(rowptr[y], rowptr[(h-1)-y], width*3);
				memcpy(rowptr[(h-1)-y], rowbuf, width*3);
			}
			for(y=0; y<h; y++)
			{
				for(x=0; x<width; x++)
				{
					c = rowptr[y][x*3+0];
					rowptr[y][x*3+0] = rowptr[y][x*3+2];
					rowptr[y][x*3+2] = c;
				}
			}

			if(astrm>0)
			{
				int i;
				if(achan > 1) {
					quicktime_set_audio_position(file, audio_pos, 0);
					quicktime_decode_audio(file, audiobuffertemp, NULL, sampsperframe, 0);
					quicktime_set_audio_position(file, audio_pos, 0);
					quicktime_decode_audio(file, audiobuffertemp + sampsperframe, NULL, sampsperframe, 1);
					for(i = 0; i < sampsperframe; i++) {
						audiobufferint[i * 2] = audiobuffertemp[i];
						audiobufferint[i * 2 + 1] = audiobuffertemp[sampsperframe + i];
					}
				} else {
					quicktime_set_audio_position(file, audio_pos, 0);
					quicktime_decode_audio(file, audiobufferint, NULL, sampsperframe, 0);
				}
				audio_pos += sampsperframe;
 				astream->AddData(audiobufferint,sampsperframe*achan);
 			}


			//printf("framebuffer = %d %d %d\n",framebuffer[0],framebuffer[1],framebuffer[2]);
 			//printf("Writing frame %d...\n",i);
			if(oheight || owidth)             //(y+cliplines-clipoffset)
			{
				ScaleX = (float) bh.biWidth / width;
				ScaleY = (float) bh.biHeight / height;

				for (y =0; y < bh.biHeight; y++) {
					NewY1 = (unsigned int)(y / ScaleY + cliplines-clipoffset) * /*Image->Bps*/ width * 3;
					for (x = 0; x < bh.biWidth; x++) {
						t1 = x / (double)bh.biWidth;
						t4 = t1 * bh.biWidth;
						t2 = t4 - (unsigned int)(t4);
						ft = t2 * PI;
						f = (1.0 - cos(ft)) * .5;
						NewX1 = ((unsigned int)(t4 / ScaleX)) * 3;
						NewX2 = ((unsigned int)(t4 / ScaleX) + 1) * 3;

						Size = y * bh.biWidth * 3/*Scaled->Bps*/ + x * 3;
						for (c = 0; c < 3; c++) {
							x1 = framebuffer[NewY1 + NewX1 + c];
							x2 = framebuffer[NewY1 + NewX2 + c];
                                                        			//printf ("X1: %d, X2: %d, Size: %d\n", NewY1 + NewX1 + c, NewY1 + NewX2 + c, width * height *3);
							scaledbuffer[Size + c] = (unsigned char)((1.0 - f) * x1 + f * x2);
                                                        			//printf("Scale: %d, Size: %d\n", Size + c, bh.biWidth*bh.biHeight *3 );
						}
					}
				}
			}
			if(blur)
			{
				im->ToYUV();
				im->Blur(blur);
			}
 			stream->AddFrame(im);

			samples_read+=sampsperframe;
			//printf("samples: %d, samples_read: %d\n",samples,samples_read);
		}
	}
	stream->Stop();
	delete avifile;
	quicktime_close(file);
}
