/* Ripped from minimad.c
 * Copyright (C) 2000-2001 Robert Leslie
 *  By Justin Schoeman (C) 2001
 */

#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <stdlib.h>

#include "mcplay.h"
#include "aconv_mad.h"

#define MAXBUF 65536

static enum mad_flow aconv_mad_input(void *data, struct mad_stream *stream)
{
	struct aconv_mad_priv_t *a = (struct aconv_mad_priv_t *)data;
	struct rbuf_data_t *rd = NULL;
	int i = 0;

	if(stream->next_frame) {
		/* The number of bytes written last time */
		i = a->lastwrite - ((unsigned long)stream->next_frame - 
			(unsigned long)(a->buf));
		memmove(a->buf, stream->next_frame, i);
	}
	while((rd = rbuf_pop_tail(a->rcaud)) == NULL)
			usleep(100000);
	a->atime_c = rd->time;
	DPRINT2("GOT: %d %d\n", rd->full, i);
	a->lastwrite = rd->full + i;
	if(a->lastwrite > MAXBUF) {
		EPRINT("MAD DECODE, BUFFER TOO SMALL!\n");
		return MAD_FLOW_STOP;
	}
	memcpy(a->buf + i, rd->data, rd->full);
	rbuf_push_tail(a->rcaud);
	mad_stream_buffer(stream, a->buf, rd->full + i);
	return MAD_FLOW_CONTINUE;
}

static inline signed int aconv_mad_scale(mad_fixed_t sample)
{
	sample += (1L << (MAD_F_FRACBITS - 16));
	if (sample >= MAD_F_ONE)
		sample = MAD_F_ONE - 1;
	else if (sample < -MAD_F_ONE)
		sample = -MAD_F_ONE;
	return sample >> (MAD_F_FRACBITS + 1 - 16);
}

static enum mad_flow aconv_mad_output(void *data,
				struct mad_header const *header,
				struct mad_pcm *pcm)
{
	struct aconv_mad_priv_t *a = (struct aconv_mad_priv_t *)data;
	unsigned int nchannels, nsamples;
	mad_fixed_t const *left_ch, *right_ch;
	signed int sample;
	short *optr;

	DPRINT2("write: %d\n", pcm->length);

	nchannels = pcm->channels;
	nsamples  = pcm->length;
	left_ch   = pcm->samples[0];
	right_ch  = pcm->samples[1];

	while (nsamples--) {
		while(!a->rd) {
			a->rd = rbuf_pop_head(a->raud);
			if(!a->rd) {
				usleep(100000);
			} else {
				a->rd->time = a->atime_c;
				a->rdused = 0;
			}
		}
		optr = (short *)((unsigned long)(a->rd->data) + a->rdused);
		sample = aconv_mad_scale(*left_ch++);
		*(optr++) = sample;
		a->rdused += 2;
		if (nchannels == 2) {
			sample = aconv_mad_scale(*right_ch++);
			*(optr++) = sample;
			a->rdused += 2;
		}
		if(a->rdused >= a->targ_size) {
			rbuf_push_head(a->raud, a->targ_size);
			a->rd = NULL;
		}
	}
	return MAD_FLOW_CONTINUE;
}

static enum mad_flow aconv_mad_error(void *data,
				struct mad_stream *stream,
				struct mad_frame *frame)
{
	EPRINT("decoding error 0x%04x\n", stream->error);
	return MAD_FLOW_CONTINUE;
}

int aconv_mad_thread(void * foo)
{
	struct aconv_mad_priv_t *a = (struct aconv_mad_priv_t *)foo;
	mad_decoder_run(&(a->decoder), MAD_DECODER_MODE_SYNC);
	return 0;
}

int aconv_mad_start(struct aconv_mad_priv_t *a, struct rbuf_t *rcaud, struct rbuf_t *raud, int targ_size)
{
	memset(a, 0, sizeof(struct aconv_mad_priv_t));
	a->targ_size = targ_size;
	a->rcaud = rcaud;
	a->raud = raud;
	a->buf = malloc(targ_size + 4);
	if(!a->buf) {
		perror("Could not malloc buffer");
		return -1;
	}
	mad_decoder_init(&(a->decoder),
			a,
			aconv_mad_input, 
			0 /* header */, 
			0 /* filter */, 
			aconv_mad_output,
			aconv_mad_error, 
			0 /* message */);
	a->aconv = SDL_CreateThread(aconv_mad_thread, a); 
	return 0;
}

void aconv_mad_stop(struct aconv_mad_priv_t *a) {
	if(a->aconv)
		SDL_KillThread(a->aconv);
	mad_decoder_finish(&(a->decoder));
	free(a->buf);
}
