#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <linux/soundcard.h>
#include <SDL_thread.h>

#include "mcplay.h"
#include "rbuf.h"
#include "audio_oss.h"

#include "x_audio_next.h"

int oss_audio_thread(void *foo)
{
	struct x_audio_priv_t *p = (struct x_audio_priv_t *)foo;
	void *targ;
	int len;
	
	DPRINT("Audio thread starting!\n");
	while(*(p->video_time) == MAXU64);
	DPRINT("Audio thread is go!\n");
	while(p->run) {
		x_audio_next_n(p, p->bsize, &targ, &len);
		write(p->fd, targ, len);
	}
	p->run = 1;
	return 0;
}

int x_audio_init(struct x_audio_priv_t *p, struct rbuf_t *rbuf, unsigned long long vstep, volatile unsigned long long * video_time, int speed, int bits, int channels)
{
	int i;
	
	DPRINT("Initialising OSS audio %d,%d,%d:\n", speed, bits, channels);
	
	p->fd = open("/dev/dsp", O_WRONLY);
	if(p->fd < 0) {
		perror("Could not open dsp");
		return -1;
	}
	i = 0x00800008U;
	if(ioctl(p->fd, SNDCTL_DSP_SETFRAGMENT, &i) < 0) {
		perror("Setting fragment size");
	}
	i = (bits == 8) ? AFMT_U8 : AFMT_S16_LE;
	if(ioctl(p->fd, SNDCTL_DSP_SAMPLESIZE, &i) < 0) {
		perror("Setting samplesize");
		return -1;
	}
	i = channels - 1;
	if(ioctl(p->fd, SNDCTL_DSP_STEREO, &i) < 0) {
		perror("Setting mono/stereo");
		return -1;
	}
	i = speed;
	if(ioctl(p->fd, SNDCTL_DSP_SPEED, &i) < 0) {
		perror("Setting speed");
		return -1;
	}
	if(ioctl(p->fd, SNDCTL_DSP_GETBLKSIZE, &i) < 0) {
		perror("Getting block size");
		i = 256;
	}
	DPRINT("Got oss block size: %i\n", i);
	p->bsize = i;
	
	i = speed;
	if(bits == 16) i *= 2;
	if(channels == 2) i *= 2;
	p->sps = i;			/* samples per second */
	p->window = 100000000ULL;	/* window is 1/10 second */

	p->rbuf = rbuf;
	p->rd = NULL;
	p->used = 0;
	p->silence = 0;
	p->discard = 0;
	p->video_time = video_time;
	
	if(bits == 16) {
		memset(&p->sblock[0], 0, SILENCE_BLOCK);
	} else {
		memset(&p->sblock[0], 128, SILENCE_BLOCK);
	}
	
	p->run = 1;
	if(SDL_CreateThread(oss_audio_thread, p) == NULL) {
		EPRINT("Could not create timer thread!\n");
		return -1;
	}
	
	return 0;
}

void x_audio_stop(struct x_audio_priv_t *p)
{
	p->run = 0;
	DPRINT("Stopping audio\n");
	while(!p->run);
	DPRINT("Stopped audio\n");
}
