#include <string.h>
#include <unistd.h>

#include "mcplay.h"
#include "rbuf.h"
#include "audio_sdl.h"

#define ASAMPLES 1024

#include "x_audio_next.h"

void sdl_audio_callback(void *userdata, Uint8 *stream, int len)
{
	struct x_audio_priv_t *p = (struct x_audio_priv_t *)userdata;
	int tlen;
	void *targ;
	int n = 0;  
	
	DPRINT2("AUDIO CALLBACK!\n");
	if(*(p->video_time) == MAXU64) {
		memset(stream, p->spec.silence, len);
		return;
	}
	while (n < len) {
		x_audio_next_n(p, len - n, &targ, &tlen);
		memcpy(stream + n, targ, tlen);
		n += tlen;
	}
	return;
}

int x_audio_init(struct x_audio_priv_t *p, struct rbuf_t *rbuf, unsigned long long vstep, volatile unsigned long long * video_time, int speed, int bits, int channels)
{
	int i;
	
	DPRINT("Initialising SDL audio %d,%d,%d:\n", speed, bits, channels);

	if(SDL_Init(SDL_INIT_AUDIO) != 0) {
		EPRINT("Error initing SDL video!\n");
		return -1;
	}

	p->spec.freq = speed;
	p->spec.format = (bits==8)?AUDIO_U8:AUDIO_S16;
	p->spec.samples = ASAMPLES;
	p->spec.channels = channels;
	p->spec.callback = sdl_audio_callback;
	p->spec.userdata = p;
	if(SDL_OpenAudio(&p->spec, NULL) != 0) {
		EPRINT("Could not get audio param!\n");
		return -1; 
	}
	
	i = speed;
	if(bits == 16) i *= 2;
	if(channels == 2) i *= 2;
	p->sps = i;p->window = 100000000ULL;
	
	p->rbuf = rbuf;
	p->rd = NULL;
	p->used = 0;
	p->silence = 0;
	p->discard = 0;
	p->video_time = video_time;
	DPRINT("Audio: buf size = %d\n", p->spec.size);
	
	memset(&p->sblock[0], p->spec.silence, SILENCE_BLOCK);
	
	SDL_PauseAudio(0);
	
	return 0;
}

void x_audio_stop(struct x_audio_priv_t *p)
{
	SDL_PauseAudio(1);
}
