#include <string.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>

#include "decode_divx.h"

int divx_init(struct divx_priv_t *d, char *fcc, int bpp, int w, int h)
{
	DEC_MEM_REQS mem;
	void *dlh;
	int codec_ver = 412;
	int dec_ver;

	memset(d, 0, sizeof(struct divx_priv_t));

	DPRINT("Getting video codec for %s\n", fcc);
	if(strncmp(fcc, "divx", 4) == 0) {
		dlh = dlopen("libdivxdecore.so", RTLD_NOW);
	} else if(strncmp(fcc, "DIVX", 4) == 0) {
		dlh = dlopen("libdivxdecore.so", RTLD_NOW);
	} else if(strncmp(fcc, "divx", 4) == 0) {
		dlh = dlopen("libdivxdecore.so", RTLD_NOW);
	}  else if(strncmp(fcc, "DX50", 4) == 0) {
		dlh = dlopen("libdivxdecore.so", RTLD_NOW);
		codec_ver = 500;
	}  else if(strncmp(fcc, "dx50", 4) == 0) {
		dlh = dlopen("libdivxdecore.so", RTLD_NOW);
		codec_ver = 500;
#if 1 /* xvid decoder not robust against interrupted streams */
	} else if(strncmp(fcc, "XVID", 4) == 0) {
		dlh = dlopen("libdivxdecore.so", RTLD_NOW);
	} else if(strncmp(fcc, "xvid", 4) == 0) {
		dlh = dlopen("libdivxdecore.so", RTLD_NOW);
#else
	} else if(strncmp(fcc, "XVID", 4) == 0) {
		dlh = dlopen("libxvidcore.so", RTLD_NOW);
	} else if(strncmp(fcc, "xvid", 4) == 0) {
		dlh = dlopen("libxvidcore.so", RTLD_NOW);
#endif
	} else {
		EPRINT("No handler for movie type %s.\n", fcc);
		goto error;
	}
	if(!dlh) {
		EPRINT("Could not open codec library \"%s\". Please make sure it is correctly installed (%s).\n", fcc, dlerror());
		goto error;
	}
	d->decore = dlsym(dlh, "decore");
	if(!d->decore) {
		EPRINT("Could not find decore symbol in library (%s).\n", dlerror());
		goto error;
	}

#if (DECORE_VERSION >= 20020322)
	dec_ver = 5000;
#else
	dec_ver = 4000;
	if(codec_ver == 500) {
		EPRINT("This DivX library does not support DivX5 - please install a newer codec and recompile.\n");
		goto error;
	}
#endif
	d->handle = (long)time(NULL);
	d->param.x_dim = w;
	d->param.y_dim = h;
	if(bpp==0) {
		DPRINT("Using overlay output\n");
		d->param.output_format = DEC_USER;
		d->buf = NULL;
	} else {
		DPRINT("Using RGB output %i\n", bpp);
		switch(bpp) {
			case 8:
				d->param.output_format = DEC_YV12;
				d->bpp = 1;
				break;
			
			case 15:
				d->param.output_format = DEC_RGB555_INV;
				d->bpp = 2;
				break;
			
			case 16:
				d->param.output_format = DEC_RGB565_INV;
				d->bpp = 2;
				break;
			
			case 24:
				d->param.output_format = DEC_RGB24_INV;
				d->bpp = 3;
				break;
			
			case 32:
				d->param.output_format = DEC_RGB32_INV;
				d->bpp = 4;
				break;
			default:
				EPRINT("divx invalid pixel depth %i\n", bpp);
				goto error;
		}
		d->buf = malloc(w*h*d->bpp);
		d->w = w;
	}
#if (DECORE_VERSION >= 200200322)
	d->param.codec_version = codec_ver;
	d->param.build_number = 0;
#endif
	d->param.time_incr = 15; /* check! */
	
	if(d->decore(d->handle, DEC_OPT_MEMORY_REQS, &d->param, &mem) != DEC_OK) {
		EPRINT("divx get mem req.\n");
		goto error;
	}
	d->param.buffers.mp4_edged_ref_buffers = malloc(mem.mp4_edged_ref_buffers_size);
	if(!d->param.buffers.mp4_edged_ref_buffers)
		goto error1;
	memset(d->param.buffers.mp4_edged_ref_buffers, 0, mem.mp4_edged_ref_buffers_size);
	d->param.buffers.mp4_edged_for_buffers = malloc(mem.mp4_edged_for_buffers_size);
	if(!d->param.buffers.mp4_edged_for_buffers)
		goto error1;
	memset(d->param.buffers.mp4_edged_for_buffers, 0, mem.mp4_edged_for_buffers_size);
	d->param.buffers.mp4_edged_back_buffers = malloc(mem.mp4_edged_back_buffers_size);
	if(!d->param.buffers.mp4_edged_back_buffers)
		goto error1;
	memset(d->param.buffers.mp4_edged_back_buffers, 0, mem.mp4_edged_back_buffers_size);
	d->param.buffers.mp4_display_buffers = malloc(mem.mp4_display_buffers_size);
	if(!d->param.buffers.mp4_display_buffers)
		goto error1;
	memset(d->param.buffers.mp4_display_buffers, 0, mem.mp4_display_buffers_size);
	d->param.buffers.mp4_state = malloc(mem.mp4_state_size);
	if(!d->param.buffers.mp4_state)
		goto error1;
	memset(d->param.buffers.mp4_state, 0, mem.mp4_state_size);
	d->param.buffers.mp4_tables = malloc(mem.mp4_tables_size);
	if(!d->param.buffers.mp4_tables)
		goto error1;
	memset(d->param.buffers.mp4_tables, 0, mem.mp4_tables_size);
	d->param.buffers.mp4_stream = malloc(mem.mp4_stream_size);
	if(!d->param.buffers.mp4_stream)
		goto error1;
	memset(d->param.buffers.mp4_stream, 0, mem.mp4_stream_size);
	d->param.buffers.mp4_reference = malloc(mem.mp4_reference_size);
	if(!d->param.buffers.mp4_reference)
		goto error1;
	memset(d->param.buffers.mp4_reference, 0, mem.mp4_reference_size);

	if(d->decore(d->handle, DEC_OPT_INIT, &d->param, NULL) != DEC_OK) {
		EPRINT("divx get mem req.\n");
		goto error1;
	}
	
	return 0;
error1:
	if(d->param.buffers.mp4_edged_ref_buffers)
		free(d->param.buffers.mp4_edged_ref_buffers);
	if(d->param.buffers.mp4_edged_for_buffers)
		free(d->param.buffers.mp4_edged_for_buffers);
	if(d->param.buffers.mp4_edged_back_buffers)
		free(d->param.buffers.mp4_edged_back_buffers);
	if(d->param.buffers.mp4_display_buffers)
		free(d->param.buffers.mp4_display_buffers);
	if(d->param.buffers.mp4_state)
		free(d->param.buffers.mp4_state);
	if(d->param.buffers.mp4_tables)
		free(d->param.buffers.mp4_tables);
	if(d->param.buffers.mp4_stream)
		free(d->param.buffers.mp4_stream);
	if(d->param.buffers.mp4_reference)
		free(d->param.buffers.mp4_reference);
error:
	return -1;

}

void divx_close(struct divx_priv_t *d)
{
	d->decore(d->handle, DEC_OPT_RELEASE, NULL, NULL);
	free(d->param.buffers.mp4_edged_ref_buffers);
	free(d->param.buffers.mp4_edged_for_buffers);
	free(d->param.buffers.mp4_edged_back_buffers);
	free(d->param.buffers.mp4_display_buffers);
	free(d->param.buffers.mp4_state);
	free(d->param.buffers.mp4_tables);
	free(d->param.buffers.mp4_stream);
	free(d->param.buffers.mp4_reference);
}

int divx_decode(struct divx_priv_t *d, void *bs, int len, int render, DEC_PICTURE *pic)
{
	if(!d->buf) {
		d->frame.bmp = pic;
		d->frame.bitstream = bs;
		d->frame.length = len;
		d->frame.render_flag = render;
	} else {
		d->frame.bmp = d->buf;
		d->frame.bitstream = bs;
		d->frame.length = len;
		d->frame.render_flag = render;
		d->frame.stride = d->w;
	}

	if(d->decore(d->handle, DEC_OPT_FRAME, &d->frame, NULL) != DEC_OK) {
		EPRINT("divx decode.\n");
		return -1;
	}
	
	if(d->buf) {
		pic->y = d->buf;
		pic->stride_y = d->w * d->bpp;
	}
	
	return 0;
}
