#include <string.h>

#include "mcplay.h"
#include "display_sdl.h"

int sdl_video_init(struct sdl_video_priv_t *p, int w, int h, int wd, int hd)
{
	DPRINT("Initialising SDL video:\n");
	if(SDL_Init(SDL_INIT_VIDEO) != 0) {
		EPRINT("Error initing SDL video!\n");
		return -1;
	}
	if((p->screen = SDL_SetVideoMode(wd, hd, 0, SDL_HWSURFACE|SDL_ANYFORMAT|SDL_RESIZABLE)) == NULL) {
		EPRINT("Error getting display surface!\n");
		return -1;
	}
	p->dest.x = p->dest.y = 0;
	p->dest.w = wd;
	p->dest.h = hd;
	p->w = w;
	p->h = h;
	p->init = 1;
	
	if((p->overlay=SDL_CreateYUVOverlay(w, h, SDL_YV12_OVERLAY, p->screen)) == NULL) {
		EPRINT("Error creating overlay surface - falling back to RGB!\n");
		if(p->screen->format->BitsPerPixel == 8) {
			SDL_Color colors[256];
			int i;
			EPRINT("Using 8bpp surface - going grayscale\n");
			for(i=0;i<256;i++) {
				colors[i].r=i;
				colors[i].g=i;
				colors[i].b=i;
				SDL_SetPalette(p->screen, SDL_LOGPAL|SDL_PHYSPAL, colors, 0, 256);
			}
		}
		return p->screen->format->BitsPerPixel;
	} else {
		DPRINT("Overlay: planes = %d; hw = %d\n", p->overlay->planes, p->overlay->hw_overlay);
		return 0;
	}
	return 0;
}

void sdl_video_resize(struct sdl_video_priv_t *p, int wd, int hd)
{
	DPRINT2("RESIZE FOUND (%dx%d)\n", wd, hd);
	
	if(!p->overlay) {
		EPRINT("Cannot resize RGB surfaces yet...\n");
		return;
	}
	
	p->init = 0;
	if((p->screen = SDL_SetVideoMode(wd, hd, 0, SDL_HWSURFACE|SDL_ANYFORMAT|SDL_RESIZABLE)) == NULL) {
		EPRINT("Error getting display surface!\n");
		return;
	}
	p->dest.w = wd;
	p->dest.h = hd;
	SDL_FreeYUVOverlay(p->overlay);
	if((p->overlay=SDL_CreateYUVOverlay(p->w, p->h, SDL_YV12_OVERLAY, p->screen)) == NULL) {
		EPRINT("Error creating overlay surface!\n");
		return;
	}
	DPRINT("Overlay: planes = %d; hw = %d\n", p->overlay->planes, p->overlay->hw_overlay);
	p->init = 1;
	return;
}

void sdl_video_display(struct sdl_video_priv_t *p, unsigned char * y, unsigned char * u, unsigned char * v, int sy, int suv)
{
	int i;
	
	if(!p->init) {
		DPRINT2("sdl_video_display on uninit display!\n");
		return;
	}

	if(p->overlay) {
		if(SDL_LockYUVOverlay(p->overlay) != 0)
			EPRINT("Could not lock overlay, trying to render anyway!\n");

		for(i=0; i<p->h; i++) {
			memcpy(p->overlay->pixels[0] + p->overlay->pitches[0] * i,
				y + sy * i, p->w);
		}
		for(i=0; i<p->h/2; i++) {
			memcpy(p->overlay->pixels[1] + p->overlay->pitches[1] * i,
				u + suv * i, p->w/2);
		}
		for(i=0; i<p->h/2; i++) {
			memcpy(p->overlay->pixels[2] + p->overlay->pitches[2] * i,
				v + suv * i, p->w/2);
		}
		
		SDL_UnlockYUVOverlay(p->overlay);

		if(SDL_DisplayYUVOverlay(p->overlay, &p->dest) != 0)
			EPRINT("Error rendering overlay, continuing anyway.\n");
	} else {
		DPRINT2("RENDER\n");
		if(SDL_LockSurface(p->screen) != 0)
			EPRINT("Could not lock surface, trying to render anyway!\n");

		for(i=0; i<p->h; i++) {
			memcpy(p->screen->pixels + p->screen->pitch * i,
				y + sy * i, p->w * p->screen->format->BytesPerPixel);
		}
		SDL_UnlockSurface(p->screen);
		SDL_UpdateRect(p->screen, 0, 0, 0, 0);
	}
}

void sdl_video_quit(struct sdl_video_priv_t *p)
{
	SDL_Quit();
}
