#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <netdb.h>

#include "mcplay.h"
#include "input_mc.h"

int mcsi_init(struct mcsi_t *m, char *url)
{
	struct hostent *he;
	struct ip_mreq mreq;
	char *ip;
	int port;
	int i;
	
	memset(m, 0, sizeof(struct mcsi_t));
	
	DPRINT("Opening url \"%s\":\n", url);

	if(strcmp(url, "-") == 0) {
		DPRINT("Streaming from stdin!\n");
		m->fd = -1;
		return 0;
	}

	if(strncmp(url + 3, "://", 3) != 0)
		goto url_usage;
	ip = url + 6;
	i = strcspn(ip, ":");
	if(i == strlen(ip))
		goto url_usage;
	ip[i] = 0;
	if(!(he = gethostbyname(ip))) {
		EPRINT("Could not resolve ip/hostname \"%s\".\n", ip);
		goto error;
	}
	port = atoi(ip + i + 1);
	DPRINT2("IP: %s, PORT: %d\n", ip, port);

	if(strncmp(url, "udp", 3) == 0) {
		fprintf(stderr, "MULTICAST...\n");
		if((m->fd = socket(AF_INET,SOCK_DGRAM,0)) < 0) {
			perror("Multicast SOCKET");
			goto error;
		}
	
		m->addr.sin_family=AF_INET;
		m->addr.sin_addr.s_addr=htonl(INADDR_ANY);
		m->addr.sin_port=htons(port);
		if(bind(m->fd,( struct sockaddr *)&m->addr, sizeof(m->addr)) < 0) {
			perror("Multicast BIND");
			goto error1;
		}
		
		memcpy(&(mreq.imr_multiaddr), he->h_addr_list[0], he->h_length);
		mreq.imr_interface.s_addr=htonl(INADDR_ANY);
		if(setsockopt(m->fd, IPPROTO_IP, IP_ADD_MEMBERSHIP, &mreq, sizeof(mreq)) < 0) {
			perror("Multicast SETSOCKOPT");
			goto error1;
		}
		m->type = 0;
		return 0;
	}
	if(strncmp(url, "tcp", 3) == 0) {
		struct protoent *new_protocol;
		
		fprintf(stderr, "UNICAST...\n");
		
		new_protocol = getprotobyname("tcp");
#ifndef CYGWIN
		endprotoent();
#endif
		if (!new_protocol) {
			perror("get tcp protocol");
			goto error;
		}
		
		if((m->fd = socket(PF_INET, SOCK_STREAM, new_protocol->p_proto)) < 0) {
			perror("Unicast SOCKET");
			goto error;
		}
	
		m->addr.sin_family=he->h_addrtype;
		memcpy(&(m->addr.sin_addr), he->h_addr_list[0], he->h_length);
		m->addr.sin_port=htons(port);

		i = connect(m->fd, (struct sockaddr *) &m->addr, sizeof(m->addr));
		if(i < 0) {
			perror("Multicast CONNECT");
			goto error1;
		}
		m->type = 1;
		return 0;
	}
	goto url_usage;
error1:
	close(m->fd);
error:
	return -1;
url_usage:
	EPRINT("Invalid url! The url must be of the form:\n");
	EPRINT("  tcp://<ip or hostname>:<port>      ,or\n");
	EPRINT("  udp://<ip or hostname>:<port>      ,or\n");
	EPRINT("  -    (to stream from stdin)\n");
	return -1;
}

int mcsi_read_raw(struct mcsi_t *m, unsigned char *buf, int len)
{
	if(m->fd < 0) {
		return read(STDIN_FILENO,
			buf,
			len);
	} else if (m->type) {
		return read(m->fd, 
			buf,
			len);
	} else {
		return recvfrom(m->fd, 
			buf,
			len,
			0,
			NULL,
			0);
	}
} 

int mcsi_read(struct mcsi_t *m, int head, unsigned char *buf, int len)
{
	int i = 0;
	int s = 0;
	if(head) {
		for(;;) {
			i = m->blength - m->bpointer;
			if(i < sizeof(uint64_t)) {
				memcpy(m->buf, m->buf + m->bpointer, i);
				m->bpointer = 0;
				m->blength = i;
				i = mcsi_read_raw(m, m->buf + m->blength, MCSI_MAXPACK);
				if(i <= 0) {
					perror("Multicast RECVS");
					return -1;
				}
				m->blength += i;
				continue;
			}
			if(*((uint64_t *)(m->buf + m->bpointer)) != MCS_SYNC) {
				s = 1;
				m->bpointer++;
				continue;
			}
			break;
		}
	}
	if(s)
		DPRINT2("RESYNC!\n");
	while(len > 0) {
		if(m->bpointer == m->blength) {
			DPRINT2("F");
			i = mcsi_read_raw(m, m->buf, MCSI_MAXPACK);
			if(i <= 0) {
				perror("Multicast RECV");
				return -1;
			}
			m->bpointer = 0;
			m->blength = i;
		}
		i = m->blength - m->bpointer;
		if(i > len)
			i = len;
		memcpy(buf, m->buf + m->bpointer, i);
		m->bpointer += i;
		buf += i;
		len -= i;
	}
	return 0;
}

void mcsi_done(struct mcsi_t *m)
{
	if(m->fd >= 0)
		close(m->fd);
	return;
}
