#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <SDL.h>
#include <SDL_thread.h>
#include <assert.h>
#ifdef CYGWIN
 #include <cygint.h>
#else
 #include <stdint.h>
#endif
#include <string.h>

#include "mcplay.h"
#include "decode_divx.h"
#include "display_sdl.h"
#include "rbuf.h"
#include "sig_sdl.h"
#ifdef AUDIO_OSS
 #include "audio_oss.h"
#else
 #include "audio_sdl.h"
#endif
#include "input_mc.h"
#include "aconv_mad.h"

struct sdl_video_priv_t sdl_video_priv;
struct divx_priv_t divx;
struct rbuf_t rvid;
int vdepth;
volatile unsigned long long video_time = MAXU64;
void *sig_sem = NULL;

int video_thread(void * foo)
{
	DEC_PICTURE pic;
	struct rbuf_data_t *rd;
	int i;
	int video_mode = 0;

	while(1) {
		i = rbuf_depth(&rvid);
		if(video_mode == 0) {
			if(i < (vdepth/2)) {
				DPRINT("Buffering ... \n"); 
				usleep(100000UL);
				continue;
			} else {
				video_mode = 1;
			}
		}
		if(i <= (vdepth/10)) {
			video_mode = 0;
			continue;
		} else if((video_mode > 1) && (i <= (vdepth/2))) {
			DPRINT("Vid OK %d:%d\n", video_mode, vdepth);
			video_mode = 1;
		}
		if(i >= ((vdepth*3)/4)) {
			EPRINT("Vid too full %d:%d\n", i, (vdepth*3)/4);
			video_mode++;
		}
		if(video_mode == 1) {
			SDL_SemWait(sig_sem);
			while(SDL_SemTryWait(sig_sem)==0);
		}
		rd = rbuf_pop_tail(&rvid);
		if(!rd)
			continue;
		SDL_LockAudio();
		video_time = rd->time;
		SDL_UnlockAudio();
		if(rd->full) {
			if(video_mode > 2) {
				divx_decode(&divx, rd->data, rd->full, 0, &pic);
			} else {
				divx_decode(&divx, rd->data, rd->full, 1, &pic);
				sdl_video_display(&sdl_video_priv, pic.y, pic.v, pic.u, pic.stride_y, pic.stride_uv);
			}
		}
		rbuf_push_tail(&rvid);
	}
}

int main(int argc, char **argv)
{
	struct x_audio_priv_t x_audio_priv;
	struct aconv_mad_priv_t aconv_mad;
	int aconv_mad_init = 0;
	struct rbuf_t raud;
	struct rbuf_t rcaud;

	struct mcsi_t mcsi;
	struct mcs_header mheader;
	struct mcs_system msystem;
	char *junk = NULL;
	int jsize = 0;

	struct rbuf_data_t *rd;
	SDL_Thread *vthread = NULL;
	int w, h;

	volatile int main_run = 1;
	int init = 0;
	int audio_init = 0;
	
	if(argc != 2) {
		EPRINT("Usage: %s <url>\n", argv[0]);
		return -1;
	}

	assert(mcsi_init(&mcsi, argv[1]) == 0);
	while((mcsi_read(&mcsi, 1, (void *)&mheader, sizeof(mheader)) == 0) && (main_run)) {
		SDL_Event ev;
		while(SDL_PollEvent(&ev)) {
			if(ev.type == SDL_VIDEORESIZE) {
				sdl_video_resize(&sdl_video_priv, ev.resize.w, ev.resize.h);
			} else if(ev.type == SDL_QUIT) {
				main_run = 0;
				DPRINT("QUIT!!\n");
			}
		}
		DPRINT2("RXFRAME 0x%016llX %d %d\n", mheader.sync, mheader.type, mheader.size);
		if(!init) {
			if(mheader.type != MCS_FRAME_SYSTEM) {
				goto junkbuf;
			} else { /* SYSTEM */
				int f;
				DPRINT("First init:\n");
				assert(mheader.size == sizeof(msystem));
				mcsi_read(&mcsi, 0, (void *)&msystem, mheader.size);
				w = msystem.video_width;
				h = msystem.video_height;
				vdepth = 2000000000ULL / msystem.video_time;
				assert((f = sdl_video_init(&sdl_video_priv, w, h, w, h)) >= 0);
				assert(rbuf_new(&rvid, 2000000000ULL / msystem.video_time, msystem.video_size) == 0);
				assert(divx_init(&divx, (char *)&msystem.video_format, f, w, h) == 0);
				assert((vthread = SDL_CreateThread(video_thread, NULL)) != NULL);
				assert((sig_sem = sdl_sig_init(msystem.video_time, &main_run)) != NULL);
				
#ifndef NOAUDIO
				if(msystem.audio_rate) {
					unsigned long long temp;
					temp = msystem.audio_rate * msystem.audio_channels;
					if(msystem.audio_bits == 16)
						temp *= 2;
					temp *= msystem.audio_time;
					temp /= 1000000000ULL;
					if(msystem.audio_format == 0x55) { /* MP3 */
						assert(rbuf_new(&rcaud, 4000000000ULL / msystem.audio_time, msystem.audio_size) == 0);
						assert(rbuf_new(&raud, 4000000000ULL / msystem.audio_time, temp) == 0);
						assert(aconv_mad_start(&aconv_mad,
							&rcaud,
							&raud,
							temp) == 0);
						aconv_mad_init = 1;
						x_audio_init(&x_audio_priv, &raud, msystem.video_time, &video_time, msystem.audio_rate,
							msystem.audio_bits, msystem.audio_channels);
					} else if(msystem.audio_format == 0x01) { /* WAV */
						assert(rbuf_new(&rcaud, 4000000000ULL / msystem.audio_time, msystem.audio_size) == 0);
						x_audio_init(&x_audio_priv, &rcaud, msystem.video_time, &video_time, msystem.audio_rate,
							msystem.audio_bits, msystem.audio_channels);
					}
					audio_init = 1;
				}
#endif
				init = 1;
			}
		} else { /* INIT */
			if(mheader.type == MCS_FRAME_SYSTEM) {
				DPRINT2("SYSTEM\n");
				goto junkbuf;
			} else if(mheader.type == MCS_FRAME_VIDEO) {
				DPRINT2("VID packet rx\n");
				rd = rbuf_pop_head(&rvid);
				if(!rd) {
					DPRINT("No free buffers on head! (video)\n");
					goto junkbuf;
				} else {
					assert(mheader.size < rvid.max);
					mcsi_read(&mcsi, 0, (void *)rd->data, mheader.size);
					rd->time = mheader.time;
					rbuf_push_head(&rvid, mheader.size);
					DPRINT2("VIDQ depth %d\n", rbuf_depth(&rvid));
				}
			} else if(mheader.type == MCS_FRAME_AUDIO) {
				DPRINT2("AUD packet rx %d\n", mheader.size);
				if(audio_init) {
					rd = rbuf_pop_head(&rcaud);
					if(!rd) {
						DPRINT("No free buffers on head! (audio)\n");
						goto junkbuf;
					} else {
						assert(mheader.size < rcaud.max);
						mcsi_read(&mcsi, 0, (void *)rd->data, mheader.size);
						rd->time = mheader.time;
						rbuf_push_head(&rcaud, mheader.size);
						DPRINT2("AUDQ depth %d\n", rbuf_depth(&rcaud));
					}
				}
			}
		}
		continue;
junkbuf:
		DPRINT2("DROPFRAME 0x%016llX %d %d\n", mheader.sync, mheader.type, mheader.size);
		if(mheader.size == 0)
			continue;
		if(mheader.size > jsize) {
			free(junk);
			junk = malloc(mheader.size);
			if(!junk) {
				perror("malloc junk");
				main_run = 0;
				continue;
			}
			jsize = mheader.size;
		}
		mcsi_read(&mcsi, 0, junk, mheader.size);
	}
	if(aconv_mad_init)
		aconv_mad_stop(&aconv_mad);
	if(audio_init)
		x_audio_stop(&x_audio_priv);
	if(vthread)
		SDL_KillThread(vthread);
	sdl_sig_stop();
	mcsi_done(&mcsi);
	sdl_video_quit(&sdl_video_priv);
	if(junk)
		free(junk);
	return 0;
}
