#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <arpa/inet.h>
#ifdef CYGWIN
 #include <cygint.h>
#else
 #include <stdint.h>
#endif
#include <netdb.h>
#include <errno.h>

#include "mcplay.h"
#include "input_mc.h"

int main(int argc, char **argv)
{
	struct mcsi_t m;
	struct hostent *he;
	struct sockaddr_in addr;
	char *ip;
	int port;
	char buf[MCSI_MAXPACK];
	int i;
	int fdo;
	
	if(argc != 3) {
		EPRINT("Usage: %s <url> <url>\n", argv[0]);
		return -1;
	}
	
	if(mcsi_init(&m, argv[1])) {
		EPRINT("Could not open input url!\n");
		return -1;
	}

	DPRINT("Opening multicast output channel %s:\n", argv[argc-1]);
	if(strncmp(argv[2], "udp://", 6) != 0) {
		EPRINT("Malformed url!\n");
		goto error;
	}
	ip = argv[argc-1] + 6;
	i = strcspn(ip, ":");
	if(i == strlen(ip)) {
		EPRINT("Malformed UDP address \"<ip|hostname>:<port>\" or \"-\"\n");
		goto error;
	}
	ip[i] = 0;
	if(!(he = gethostbyname(ip))) {
		EPRINT("Could not resolve ip/hostname \"%s\".\n", ip);
		goto error;
	}
	port = atoi(ip + i + 1);
	DPRINT("IP: %s, PORT: %d\n", ip, port);
	
	if((fdo = socket(AF_INET,SOCK_DGRAM,0)) < 0) {
		perror("Multicast SOCKET");
		goto error;
	}

	addr.sin_family=he->h_addrtype;
	memcpy(&(addr.sin_addr), he->h_addr_list[0], he->h_length);
	addr.sin_port=htons(port);
	
	while(1) {
		i = mcsi_read_raw(&m, buf, MCSI_MAXPACK);
		if(i == 0) {
			break;
		} else if(i < 0) {
			if((errno == EAGAIN) || (errno == EINTR)) {
				continue;
			} else {
				break;
			}
		} else {
			sendto(fdo, buf, i, 0, (struct sockaddr *)&addr, sizeof(addr));
		}
	}
	fprintf(stderr, "Server closed connection!\n");
	mcsi_done(&m);
	close(fdo);
	return 0;
error:
	return -1;
}
