#include <unistd.h>
#include <string.h>
#include <signal.h>

#include "mcplay.h"
#include "aconv_mpg123.h"

int audio_thread_c(void * foo)
{
	volatile struct aconv_mpg123_priv_t *a =
		(volatile struct aconv_mpg123_priv_t *)foo;
	struct rbuf_data_t *rd = NULL;
	char * p;
	int len, i;
	
	DPRINT2("CWRITE %lu\n", (unsigned long)a);

	while(1) {
		rd = rbuf_pop_tail(a->rcaud);
		if(!rd) {
			usleep(100000);
			continue;
		}
		a->atime_c = rd->time;
		p = rd->data;
		len = rd->full;
		while(len > 0) {
			i = write(a->c_wr, p, len);
			if(i < 0) {
				perror("cwrite");
				return 0;
			}
			len -= i;
			p += i;
		}
		rbuf_push_tail(a->rcaud);
	}
	return 0;
}

int audio_thread_u(void * foo)
{
	volatile struct aconv_mpg123_priv_t *a =
		(volatile struct aconv_mpg123_priv_t *)foo;
	struct rbuf_data_t *rd = NULL;
	char * p;
	int len, i;
	
	DPRINT2("UREAD %lu\n", (unsigned long)a);

	while(1) {
		rd = rbuf_pop_head(a->raud);
		if(!rd) {
			usleep(100000);
			continue;
		}
		rd->time = a->atime_c;
		p = rd->data;
		len = a->targ_size;
		while(len > 0) {
			i = read(a->u_rd, p, len);
			if(i < 0) {
				perror("uread");
				return 0;
			}
			len -= i;
			p += i;
		}
		rbuf_push_head(a->raud, a->targ_size);
	}
	return 0;
}

int aconv_mpg123_start(struct aconv_mpg123_priv_t *a, struct rbuf_t *rcaud, struct rbuf_t *raud, int targ_size)
{
	int filedes[2];
	
	memset(a, 0, sizeof(struct aconv_mpg123_priv_t));
	a->targ_size = targ_size;
	a->rcaud = rcaud;
	a->raud = raud;
	if(pipe(filedes) < 0) {
		perror("Creating compressed pipe");
		goto error;
	}
	a->c_rd = filedes[0];
	a->c_wr = filedes[1];
	if(pipe(filedes) < 0) {
		perror("Creating uncompressed pipe");
		goto error;
	}
	a->u_rd = filedes[0];
	a->u_wr = filedes[1];

	a->mid = fork();
	if(a->mid < 0) {
		perror("fork");
		goto error;
	}
	if(a->mid == 0) { /* child */
		char *argv[6];
		extern char **environ;
		close(STDIN_FILENO);
		dup2(a->c_rd, STDIN_FILENO);
		close(a->c_rd);
		close(STDOUT_FILENO);
		dup2(a->u_wr, STDOUT_FILENO);
		close(a->u_wr);
		argv[0] = "mpg123";
		argv[1] = "-s";
		argv[2] = "-b";
		argv[3] = "1";
		argv[4] = "-";
		argv[5] = 0;
		execve("/bin/mpg123", argv, environ);
		execve("/usr/bin/mpg123", argv, environ);
		execve("/usr/local/bin/mpg123", argv, environ);
		perror("execve");
	}

	a->athread_c = SDL_CreateThread(audio_thread_c, a); 
	a->athread_u = SDL_CreateThread(audio_thread_u, a);
	return 0;
error:
	if(a->c_rd > 0)
		close(a->c_rd);
	if(a->c_wr > 0)
		close(a->c_wr);
	return -1;
}

void aconv_mpg123_stop(struct aconv_mpg123_priv_t *a) {
	if(a->athread_c)
		SDL_KillThread(a->athread_c);
	if(a->athread_u)
		SDL_KillThread(a->athread_u);
	if(a->mid > 0)
		kill(a->mid, SIGKILL);
}
