#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <linux/soundcard.h>
#include <SDL_thread.h>

#include "mcplay.h"
#include "rbuf.h"
#include "audio_oss.h"

#define ASAMPLES 1024

int oss_audio_thread(void *foo)
{
	struct x_audio_priv_t *p = (struct x_audio_priv_t *)foo;
	int n = 0;
	int i;
	unsigned long long vtime, atime, ttime;
	
	DPRINT("Audio thread starting!\n");
	while(*(p->video_time) == MAXU64);
	DPRINT("Audio thread is go!\n");
	while(p->run) {
		/* Make sure we have data available */
		if(!p->rd) {
			p->rd = rbuf_pop_tail(p->rbuf);
			if(!p->rd) {
				DPRINT("Damn, out of audio data... %d\n", rbuf_depth(p->rbuf));
				usleep(25000);
				continue;
			}
			/* got empty buffer - try again*/
			if(p->rd->full == 0) {
				rbuf_push_tail(p->rbuf);
				p->rd = NULL;
				continue;
			}
			p->used = 0;
			DPRINT2("Video Time - %llu ; Audio Time - %llu: %llu, %d\n", *(p->video_time), p->rd->time, p->window, p->delta);
			vtime = *(p->video_time);
			atime = p->rd->time;
			
			if(ioctl(p->fd, SNDCTL_DSP_GETODELAY, &i) < 0) {
				perror("Getting output delay");
				i = 0;
			}
			/* time played audio will be behind */
			ttime = i;
			ttime *= 1000000ULL;
			ttime /= (unsigned long long)p->sps;
			ttime *= 1000ULL;
			/* This is the time we must play now to match vid */
			vtime += ttime;	
		
			if((vtime > atime) && ((vtime - atime) > p->window)) {
				ttime = vtime - atime;
				ttime /= 1000ULL;
				ttime *= (unsigned long long)p->sps;
				ttime /= 1000000ULL;
				ttime >>= 1;
				ttime <<= 1;
				p->discard = ttime;
				p->silence = 0;
				DPRINT(" *** discard *** %f %f %f %f %lld %d %d\n", vtime/1e6, atime/1e6, (vtime-atime)/1e6, p->window/1e6, ttime, rbuf_depth(p->rbuf), p->rbuf->count);
			} else if((atime > vtime) && ((atime - vtime) > p->window)) {
				ttime = atime - vtime;
				ttime /= 1000ULL;
				ttime *= (unsigned long long)p->sps;
				ttime /= 1000000ULL;
				ttime >>= 1;
				ttime <<= 1;
				p->silence = ttime;
				p->discard = 0;
				DPRINT(" *** silence *** %f %f %f %f %lld %d %d\n", vtime/1e6, atime/1e6, (atime-vtime)/1e6, p->window/1e6, ttime, rbuf_depth(p->rbuf), p->rbuf->count);
			} else {
				p->silence = 0;
				p->discard = 0;
			}
		}
		while(p->silence) {
			n = (p->bsize > p->silence) ? p->silence : p->bsize;
			write(p->fd, &p->sblock[0], n);
			p->silence -= n;
		}
		i = p->rd->full - p->used;
		if(p->discard) {
			if(p->discard > i) {
				p->discard -= i;
				rbuf_push_tail(p->rbuf);
				p->rd = NULL;
			} else {
				p->used += p->discard;
				i -= p->discard;
				p->discard = 0;
			}
			continue;
		}

		n = (p->bsize > i) ? i : p->bsize;
		write(p->fd, p->rd->data + p->used, n);
		p->used += n;
		if (p->used >= p->rd->full) {
			rbuf_push_tail(p->rbuf);
			p->rd = NULL;
		}
	}
	p->run = 1;
	return 0;
}

int x_audio_init(struct x_audio_priv_t *p, struct rbuf_t *rbuf, unsigned long long vstep, volatile unsigned long long * video_time, int speed, int bits, int channels)
{
	int i;
	
	DPRINT("Initialising OSS audio %d,%d,%d:\n", speed, bits, channels);
	
	p->fd = open("/dev/dsp", O_WRONLY);
	if(p->fd < 0) {
		perror("Could not open dsp");
		return -1;
	}
	i = 0x00800008U;
	if(ioctl(p->fd, SNDCTL_DSP_SETFRAGMENT, &i) < 0) {
		perror("Setting fragment size");
	}
	i = (bits == 8) ? AFMT_U8 : AFMT_S16_LE;
	if(ioctl(p->fd, SNDCTL_DSP_SAMPLESIZE, &i) < 0) {
		perror("Setting samplesize");
		return -1;
	}
	i = channels - 1;
	if(ioctl(p->fd, SNDCTL_DSP_STEREO, &i) < 0) {
		perror("Setting mono/stereo");
		return -1;
	}
	i = speed;
	if(ioctl(p->fd, SNDCTL_DSP_SPEED, &i) < 0) {
		perror("Setting speed");
		return -1;
	}
	if(ioctl(p->fd, SNDCTL_DSP_GETBLKSIZE, &i) < 0) {
		perror("Getting block size");
		i = 256;
	}
	DPRINT("Got oss block size: %i\n", i);
	p->bsize = i;
	
	i = speed;
	if(bits == 16) i *= 2;
	if(channels == 2) i *= 2;
	p->sps = i;			/* samples per second */
	p->window = 100000000ULL;	/* window is 1/10 second */

	p->rbuf = rbuf;
	p->rd = NULL;
	p->used = 0;
	p->silence = 0;
	p->discard = 0;
	p->video_time = video_time;
	
	if(bits == 16) {
		memset(&p->sblock[0], 0, SILENCE_BLOCK);
	} else {
		memset(&p->sblock[0], 128, SILENCE_BLOCK);
	}
	
	p->run = 1;
	if(SDL_CreateThread(oss_audio_thread, p) == NULL) {
		EPRINT("Could not create timer thread!\n");
		return -1;
	}
	
	return 0;
}

void x_audio_stop(struct x_audio_priv_t *p)
{
	p->run = 0;
	DPRINT("Stopping audio\n");
	while(!p->run);
	DPRINT("Stopped audio\n");
}
