#include <string.h>

#include "mcplay.h"
#include "rbuf.h"
#include "audio_sdl.h"

#define ASAMPLES 1024

void sdl_audio_callback(void *userdata, Uint8 *stream, int len)
{
	struct x_audio_priv_t *p = (struct x_audio_priv_t *)userdata;
	int n = 0;
	int i;
	unsigned long long vtime, atime;
	
	DPRINT2("AUDIO CALLBACK!\n");
	if(*(p->video_time) != MAXU64) {
		while (n < len) {
			if(!p->rd) {
				p->rd = rbuf_pop_tail(p->rbuf);
				if(!p->rd) {
					DPRINT("Damn, out of audio data... %d\n", rbuf_depth(p->rbuf));
					break;
				}
				if(p->rd->full == 0) {
					rbuf_push_tail(p->rbuf);
					p->rd = NULL;
					continue;
				}
				p->used = 0;
				DPRINT2("Video Time - %llu ; Audio Time - %llu: %llu, %d\n", *(p->video_time), p->rd->time, p->window, p->delta);
				vtime = *(p->video_time);
				/* - = audio plays earlier */
//				atime = p->rd->time - p->ofs;
//				atime = p->rd->time + p->ofs;
				atime = p->rd->time - 500000000ULL; /* OSS  = 0.5s buffer */
//				atime = p->rd->time - 100000000ULL;
				
				if((vtime > atime) && ((vtime - atime) > p->window)) {
					DPRINT(" *** discard *** %f %f %f %f %d %d %d\n", vtime/1e6, atime/1e6, (vtime-atime)/1e6, p->window/1e6, p->delta, rbuf_depth(p->rbuf), p->rbuf->count);
					p->discard = p->delta;
					p->silence = 0;
				} else if((atime > vtime) && ((atime - vtime) > p->window)) {
					DPRINT(" *** silence *** %f %f %f %f %d %d %d\n", vtime/1e6, atime/1e6, (atime-vtime)/1e6, p->window/1e6, p->delta, rbuf_depth(p->rbuf), p->rbuf->count);
					p->silence = p->delta;
					p->discard = 0;
				} else {
					p->silence = 0;
					p->discard = 0;
				}
			}
			i = p->rd->full - p->used;
			if(p->silence) {
				if(p->silence > (len - n)) {
					p->silence -= len - n;
					break; /* fill silence */
				} else{
					memset(stream + n, p->spec.silence, p->silence);
					n += p->silence;
					p->silence = 0;
				}
			}
			if(p->discard) {
				if(p->discard > i) {
					p->discard -= i;
					rbuf_push_tail(p->rbuf);
					p->rd = NULL;
					continue;
				} else {
					p->used += p->discard;
					i -= p->discard;
					p->discard = 0;
				}
			}
			i = (i>(len-n))?(len-n):i;
			memcpy(stream + n, p->rd->data + p->used, i);
			n += i;
			p->used += i;
			if (p->used >= p->rd->full) {
				rbuf_push_tail(p->rbuf);
				p->rd = NULL;
			}
		}
	}
	if(n < len) {
		DPRINT2("Filling silence\n");
		memset(stream + n, p->spec.silence, len-n);
	}
	return;
}

int x_audio_init(struct x_audio_priv_t *p, struct rbuf_t *rbuf, unsigned long long vstep, volatile unsigned long long * video_time, int speed, int bits, int channels)
{
	unsigned long long temp;
	
	DPRINT("Initialising SDL audio %d,%d,%d:\n", speed, bits, channels);

	if(SDL_Init(SDL_INIT_AUDIO) != 0) {
		EPRINT("Error initing SDL video!\n");
		return -1;
	}

	p->spec.freq = speed;
	p->spec.format = (bits==8)?AUDIO_U8:AUDIO_S16;
	p->spec.samples = ASAMPLES;
	p->spec.channels = channels;
	p->spec.callback = sdl_audio_callback;
	p->spec.userdata = p;
	if(SDL_OpenAudio(&p->spec, NULL) != 0) {
		EPRINT("Could not get audio param!\n");
		return -1; 
	}
	/* Bytes per sample */
	p->Bps = channels << ((bits==8)?0:1);
	/* bytes per second */
	p->sps = p->Bps * speed;
	/* nanosec stored in buffer */
	p->ofs = (unsigned long long)p->spec.size * 1000000000ULL;
	p->ofs /= p->sps;
	p->rbuf = rbuf;
	p->rd = NULL;
	p->used = 0;
	p->silence = 0;
	p->discard = 0;
	p->video_time = video_time;
	p->window = 100000000ULL; /* Synch window is 0.1s */
//	p->window = vstep * 4;
	temp = p->sps * p->window;
	temp /= 4; /* adjust by max 1/4 of window at a time */
	temp /= 1000000000ULL;
	temp &= 0xfffffffffffffffeULL;
	p->delta = temp;
	DPRINT("Audio: buf size = %d\n", p->spec.size);
	SDL_PauseAudio(0);
	
	return 0;
}

void x_audio_stop(struct x_audio_priv_t *p)
{
	SDL_PauseAudio(1);
}
