#include <signal.h>
#include <sys/time.h>
#include <SDL.h>

#include "mcplay.h"
#include "sig_sdl.h"

SDL_sem *sdl_sig_sem = NULL;
volatile int *sdl_sig_run = NULL;

void sdl_sig_signal_int(int foo)
{
	DPRINT("\nCTRL-C...\n");
	*sdl_sig_run = 0;
}

void sdl_sig_signal_alarm(int foo)
{
//	DPRINT("\nSignal...\n");
	SDL_SemPost(sdl_sig_sem);
}

void *sdl_sig_init(unsigned long long time, volatile int *run)
{
	struct sigaction action;
	struct itimerval it;
	
	sdl_sig_run = run;
	action.sa_handler = sdl_sig_signal_int;
	sigfillset(&action.sa_mask);
	sigdelset(&action.sa_mask, SIGINT);
	action.sa_flags = SA_RESTART|SA_NOMASK;
	if (sigaction(SIGINT, &action, NULL) != 0)
		EPRINT("Could not capture SIGINT handler.\n");
	
	sdl_sig_sem = SDL_CreateSemaphore(0);
	if(!sdl_sig_sem) {
		EPRINT("Could not create signal semaphore!\n");
		return NULL;
	}

	action.sa_handler = sdl_sig_signal_alarm;
	sigfillset(&action.sa_mask);
	sigdelset(&action.sa_mask, SIGALRM);
	action.sa_flags = SA_RESTART|SA_NOMASK;
	if (sigaction(SIGALRM, &action, NULL) != 0) {
		EPRINT("Could not capture alarm handler.\n");
		SDL_DestroySemaphore(sdl_sig_sem);
		return NULL;
	}
	
	time /= 1000ULL;
	it.it_interval.tv_sec = time / 1000000ULL;
	it.it_interval.tv_usec = time % 1000000ULL;
	it.it_value.tv_sec = it.it_interval.tv_sec;
	it.it_value.tv_usec = it.it_interval.tv_usec;
	EPRINT("Timer: %ld:%ld\n", it.it_value.tv_sec, it.it_value.tv_usec);
	if(setitimer(ITIMER_REAL, &it, NULL) != 0) {
		EPRINT("Could not set timer!\n");
		return NULL;
	}
	return sdl_sig_sem;
}

void sdl_sig_stop(void)
{
	struct sigaction action;
	struct itimerval it;
	
	it.it_interval.tv_sec = 0;
	it.it_interval.tv_usec = 0;
	it.it_value.tv_sec = 0;
	it.it_value.tv_usec = 0;
	if(setitimer(ITIMER_REAL, &it, NULL) != 0) {
		EPRINT("Could not clear timer!\n");
	}

	action.sa_handler = SIG_DFL;
	sigfillset(&action.sa_mask);
	sigdelset(&action.sa_mask, SIGALRM);
	action.sa_flags = SA_RESTART|SA_NOMASK;
	if (sigaction(SIGALRM, &action, NULL) != 0) {
		EPRINT("Could not restore alarm handler.\n");
	}
	
	SDL_DestroySemaphore(sdl_sig_sem);
}
