#include <string.h>
#include <stdlib.h>
#include "mcplay.h"
#include "rbuf.h"

int rbuf_new(struct rbuf_t * rbuf, int n, int size)
{
	int i;
	
	rbuf->buf = (struct rbuf_data_t *)malloc(n * sizeof(struct rbuf_data_t));
	if(!rbuf->buf) {
		EPRINT("Error allocating new ring buffer!\n");
		return -1;
	}
	memset(rbuf->buf, 0, (n * sizeof(struct rbuf_data_t)));
	for(i=0; i<n; i++) {
		rbuf->buf[i].data = malloc(size);
		if(!rbuf->buf[i].data) {
			EPRINT("Error allocating buffer data!\n");
			while(--i >= 0)
				free(rbuf->buf[i].data);
			free(rbuf);
			return -1;
		}
	}
	rbuf->count = n;
	rbuf->head = 0;
	rbuf->tail = 0;
	rbuf->max = size;
	return 0;
}

void rbuf_free(struct rbuf_t *rbuf)
{
	int i = rbuf->count;
	while(--i >= 0)
		free(rbuf->buf[0].data);
	free(rbuf->buf);
}

int rbuf_depth(struct rbuf_t *rbuf)
{
	int i = rbuf->head;
	if(i < rbuf->tail)
		i += rbuf->count;
	i -= rbuf->tail;
	if((i == 0) && (rbuf->buf[rbuf->tail].full))
		i = rbuf->count;
	return i;
}

struct rbuf_data_t *rbuf_pop_tail(struct rbuf_t *rbuf)
{
	if(rbuf->buf[rbuf->tail].valid == 0) {
		DPRINT2("No full buffers to pop on tail.\n");
		return NULL;
	}
	return &(rbuf->buf[rbuf->tail]);
}

void rbuf_push_tail(struct rbuf_t *rbuf)
{
	int i;
	
	i = rbuf->tail;
	rbuf->buf[i].valid = 0;
	if(++i >= rbuf->count)
		i = 0;
	rbuf->tail = i;
}

struct rbuf_data_t *rbuf_pop_head(struct rbuf_t *rbuf)
{
	if(rbuf->buf[rbuf->head].valid != 0) {
		DPRINT2("No empty buffers to pop on head.\n");
		return NULL;
	}
	return &(rbuf->buf[rbuf->head]);
}

void rbuf_push_head(struct rbuf_t *rbuf, int n)
{
	int i;
	
	i = rbuf->head;
	rbuf->buf[i].full = n;
	rbuf->buf[i].valid = 1;
	if(++i >= rbuf->count)
		i = 0;
	rbuf->head = i;
}
