#include <signal.h>
#include <sys/time.h>
#include <SDL.h>
#include <SDL_thread.h>
#include <unistd.h>

#include "mcplay.h"
#include "sig_sdl.h"

SDL_sem *sdl_sig_sem = NULL;
volatile int *sdl_sig_run = NULL;
volatile int sdl_timer_run = 1;
volatile unsigned long long sdl_sig_tpf = 0;

void sdl_sig_signal_int(int foo)
{
	DPRINT("\nCTRL-C...\n");
	*sdl_sig_run = 0;
}

unsigned long long ntime(void)
{
	unsigned long long t;
	struct timeval tv;
	
	gettimeofday(&tv, NULL);
	t = tv.tv_sec;
	t *= 1000000ULL;
	t += tv.tv_usec;
	t *= 1000ULL;
	return t;
}

int sdl_timer_thread(void *foo)
{
	unsigned long long ftime, t;
	
	ftime = ntime();
	while(sdl_timer_run) {
		t = ntime();
		ftime += sdl_sig_tpf;
		if(ftime > t) {
			t = ftime - t;
			t /= 1000ULL;
			usleep((unsigned long)t);
		}
		SDL_SemPost(sdl_sig_sem);
	}
	sdl_timer_run = 1;
	return 0;
}

void *sdl_sig_init(unsigned long long time, volatile int *run)
{
	struct sigaction action;
	
	sdl_sig_run = run;
	sdl_sig_tpf = time;
	
	action.sa_handler = sdl_sig_signal_int;
	sigfillset(&action.sa_mask);
	sigdelset(&action.sa_mask, SIGINT);
	action.sa_flags = SA_RESTART|SA_NOMASK;
	if (sigaction(SIGINT, &action, NULL) != 0)
		EPRINT("Could not capture SIGINT handler.\n");
	
	sdl_sig_sem = SDL_CreateSemaphore(0);
	if(!sdl_sig_sem) {
		EPRINT("Could not create signal semaphore!\n");
		return NULL;
	}

	if(SDL_CreateThread(sdl_timer_thread, NULL) == NULL) {
		EPRINT("Could not create timer thread!\n");
		return NULL;
	}
	return sdl_sig_sem;
}

void sdl_sig_stop(void)
{
	struct sigaction action;
	
	action.sa_handler = SIG_DFL;
	sigfillset(&action.sa_mask);
	action.sa_flags = SA_RESTART|SA_NOMASK;
	if (sigaction(SIGALRM, &action, NULL) != 0) {
		EPRINT("Could not restore alarm handler.\n");
	}

	sdl_timer_run = 0;
	while(!sdl_timer_run);
	SDL_DestroySemaphore(sdl_sig_sem);
}
