void x_audio_next_n(struct x_audio_priv_t *p, int req, void **targ, int *len)
{
	unsigned long long atime, vtime, ttime;
	int i, n;
#ifdef AUDIO_OSS
	static int first = 1;
#endif
	
	while(1) {
		/* If we are playing data or silence, fall through */
		if((!p->rd) && (!p->silence)) { 
			p->rd = rbuf_pop_tail(p->rbuf);
			if(!p->rd) {
				DPRINT2("Damn, out of audio data... %d\n", rbuf_depth(p->rbuf));
				usleep(25000);
				continue;
			}
			/* got empty buffer - try again*/
			if(p->rd->full == 0) {
				rbuf_push_tail(p->rbuf);
				p->rd = NULL;
				continue;
			}
			p->used = 0;
			DPRINT2("Video Time - %llu ; Audio Time - %llu: %llu, %d\n", *(p->video_time), p->rd->time, p->window, p->delta);
			vtime = *(p->video_time);
			atime = p->rd->time;
			
#ifdef AUDIO_OSS
			/* OSS can determine offset accurately */
			if(!first) {
				if(ioctl(p->fd, SNDCTL_DSP_GETODELAY, &i) < 0) {
					perror("Getting output delay");
					i = 0;
				}
			} else {
				i = 0;
				first = 0;
			}
			/* time played audio will be behind */
			ttime = i;
			ttime *= 1000000ULL;
			ttime /= (unsigned long long)p->sps;
			ttime *= 1000ULL;
			/* This is the time we must play now to match vid */
			vtime += ttime;	
#else
			/* SDL we have to guess */
#ifdef CYGWIN
			vtime += 100000000ULL;
#else 
			vtime += 500000000ULL;
#endif
#endif
			
			if((vtime > atime) && ((vtime - atime) > p->window)) {
				ttime = vtime - atime;
				ttime /= 1000ULL;
				ttime *= (unsigned long long)p->sps;
				ttime /= 1000000ULL;
				ttime >>= 1;
				ttime <<= 1;
				p->discard = ttime;
				p->silence = 0;
				DPRINT(" *** discard *** %f %f %f %f %lld %d %d\n", vtime/1e6, atime/1e6, (vtime-atime)/1e6, p->window/1e6, ttime, rbuf_depth(p->rbuf), p->rbuf->count);
			} else if((atime > vtime) && ((atime - vtime) > p->window)) {
				ttime = atime - vtime;
				ttime /= 1000ULL;
				ttime *= (unsigned long long)p->sps;
				ttime /= 1000000ULL;
				ttime >>= 1;
				ttime <<= 1;
				p->silence = ttime;
				p->discard = 0;
				DPRINT(" *** silence *** %f %f %f %f %lld %d %d\n", vtime/1e6, atime/1e6, (atime-vtime)/1e6, p->window/1e6, ttime, rbuf_depth(p->rbuf), p->rbuf->count);
			} else {
				p->silence = 0;
				p->discard = 0;
			}
		}
		if(p->silence) {
			n = (req > p->silence) ? p->silence : req;
			n = (n > SILENCE_BLOCK) ? SILENCE_BLOCK : n;
			*targ = &p->sblock[0];
			*len = n;
			p->silence -= n;
			return;
		}
		i = p->rd->full - p->used;
		if(p->discard) {
			if(p->discard > i) {
				p->discard -= i;
				rbuf_push_tail(p->rbuf);
				p->rd = NULL;
			} else {
				p->used += p->discard;
				i -= p->discard;
				p->discard = 0;
			}
			continue;
		}

		n = (req > i) ? i : req;
		*targ = p->rd->data + p->used;
		*len = n;
		p->used += n;
		if (p->used >= p->rd->full) {
			rbuf_push_tail(p->rbuf);
			p->rd = NULL;
		}
		return;
	}
}
