/* 
   timer.c (C) Justin Schoeman 2000 (justin@suntiger.ee.up.ac.za)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
*/
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <signal.h>
#include <time.h>
#include <string.h>
#include <sys/time.h>

int child_pid;

time_t cur_time, start_time, end_time;

void usage(void)
{
 fprintf(stderr, "Usage: timer <start time> <end time> <command>\n");
 fprintf(stderr, "  start time: time to start command\n");
 fprintf(stderr, "    end time: time to end command\n");
 fprintf(stderr, "     command: command to be executed\n");
 fprintf(stderr, "              --- time can be [+]SS:MM:HH:DD:MM:YY\n");
 fprintf(stderr, "              --- the + sign inndicates a realtive time.\n");
 exit(-1);
}

time_t tconv(char *tstring, time_t intime)
{
 struct tm tms, c_tms;
 int i;
 time_t tt;
 char *ts2;
 
 c_tms=*localtime(&intime);
 if(tstring[0]=='+') /* skip the leading + */
  ts2=&(tstring[1]);
 else
  ts2=&(tstring[0]);
 i=sscanf(ts2, "%d:%d:%d:%d:%d:%d", &(tms.tm_sec), &(tms.tm_min), &(tms.tm_hour), &(tms.tm_mday), &(tms.tm_mon), &(tms.tm_year));
 if(tstring[0]=='+')
 {
  switch(i)
  {
   case 6: c_tms.tm_year+=tms.tm_year;
   case 5: c_tms.tm_mon+=tms.tm_mon;
   case 4: c_tms.tm_mday+=tms.tm_mday;
   case 3: c_tms.tm_hour+=tms.tm_hour;
   case 2: c_tms.tm_min+=tms.tm_min;
   case 1: c_tms.tm_sec+=tms.tm_sec; break;
   case 0: return 0;
   default: return 0;
  }
 } else
 {
  switch(i)
  {
   case 6: c_tms.tm_year=tms.tm_year + 100; /* Y2K */
   case 5: c_tms.tm_mon=tms.tm_mon - 1; /* 0 based month */
   case 4: c_tms.tm_mday=tms.tm_mday;
   case 3: c_tms.tm_hour=tms.tm_hour;
   case 2: c_tms.tm_min=tms.tm_min;
   case 1: c_tms.tm_sec=tms.tm_sec; break;
   case 0: return 0;
   default: return 0;
  }
 }
 if((tt=mktime(&c_tms))==-1)return 0;
 return tt;
}

int main(int argc, char ** argv)
{
 char s[1024];
 if(argc<4)usage();
 cur_time=time(NULL);
 if((start_time=tconv(argv[1], cur_time))==0)usage();
 if((end_time=tconv(argv[2], start_time))==0)usage();
 fprintf(stderr, "Starting command %s at %s", argv[3], ctime(&start_time));
 fprintf(stderr, "Command will terminate at %s", ctime(&end_time));

 while((cur_time=time(NULL))<start_time)
 {
  strcpy(s, ctime(&cur_time));
  s[strlen(s)-1]=0;
  fprintf(stderr, "Current time is: %s.           \r", s);
  sleep(1);
 }

 switch (child_pid=fork())
 {
  case -1: fprintf(stderr, "Could not fork!\n"); exit(-1); break;
  case 0:
  	fprintf(stderr, "\nStarting command...\n");
  	execvp(argv[3], &(argv[3]));
  	perror("\nCould not execute command!");
  	break;
  default:
	while((cur_time=time(NULL))<end_time)
	{
	 strcpy(s, ctime(&cur_time));
	 s[strlen(s)-1]=0;
	 fprintf(stderr, "Current time is: %s.           \r", s);
	 sleep(1);
	}	
	fprintf(stderr, "\nTerminating command...\n");
  	kill(child_pid, SIGINT);
  	break;
 }
 return 0;
}
