/* 
   v4l1_core.c (C) Justin Schoeman 2001 (justin@suntiger.ee.up.ac.za)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
*/

#include <stdlib.h>
#include <stdio.h> 
#include <fcntl.h> 
#include <string.h>
#include <sys/ioctl.h>
#include <sys/mman.h> 
#include <errno.h>    
#include <unistd.h>   
#include <linux/videodev.h>
#include <pthread.h>

#include "nvrec.h"

#if (defined __LINUX_VIDEODEV2_H) || (defined V4L2_MAJOR_VERSION)

void v4lgen_core_usage()
{
	v4l1_core.cap_core_usage();
	v4l2_core.cap_core_usage();
}

struct cap_core_t * v4lgen_core_init(char ** args, unsigned int **pfmt)
{
	struct cap_core_t * cct = NULL;
	struct v4l2_capability v4l2cap;
	struct video_capability v4l1cap;
	char * fdev = NULL;
	int fd = -1;
	NV_ARGS;
	
	NV_SCAN_ARGS {
		if(strcmp(NV_THIS_ARG, "-v") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -v.\n");
				goto usage;
			}
			fdev = NV_THIS_ARG;
		}
	}
	IPRINT("Device %s\n", fdev);

	fd = open(fdev, O_RDWR);
	if(fd < 0) {
		PERROR("Cannot open capture device");
		goto error;
	}
#ifndef V4L2_MAJOR_VERSION
	if (ioctl(fd, VIDIOC_QUERYCAP, &v4l2cap) == 0) {
		if((v4l2cap.capabilities & V4L2_CAP_VIDEO_CAPTURE) &&
		   (v4l2cap.capabilities & V4L2_CAP_STREAMING)) {
			close(fd);
			fd = -1;
		   	cct = v4l2_core.cap_core_init(args, pfmt);
			
		   	if(cct) {
				IPRINT("Got capable v4l2 device: '%s' (driver:'%s').\n", v4l2cap.driver,v4l2cap.card);
				return cct;
			}
		}
	}
#else
	if (ioctl(fd, VIDIOC_QUERYCAP, &v4l2cap) == 0) {
		if((v4l2cap.type == V4L2_TYPE_CAPTURE) &&
		   (v4l2cap.flags & V4L2_FLAG_STREAMING)) {
			close(fd);
			fd = -1;
			cct = v4l2_core.cap_core_init(args, pfmt);
			if(cct) {
				IPRINT("Got capable v4l2 device: %s.\n", v4l2cap.name);
				return cct;
			}
		}
	}
#endif
	if(fd < 0)
		fd = open(fdev, O_RDWR);
	if(fd < 0) {
		PERROR("Cannot open capture device");
		goto error;
	}
	if (ioctl(fd, VIDIOCGCAP, &v4l1cap) == 0) {
		if(v4l1cap.type & VID_TYPE_CAPTURE) {
			close(fd);
			fd = -1;
		   	cct = v4l1_core.cap_core_init(args, pfmt);
		   	if(cct) {
				IPRINT("Got capable v4l1 device: %s.\n", v4l1cap.name);
				return cct;
			}
		}
	}
	EPRINT("Could not get any capable capture devices.\n");
error:
	if(fd >= 0)
		close(fd);
	return NULL;
usage:
	EPRINT("Error parsing options.  Please run without any command line options for\n");
	EPRINT("detailed usage information.\n");
	return NULL;
}

#else

struct cap_core_t * v4lgen_core_init(char ** args, unsigned int **pfmt)
{
	return v4l1_core.cap_core_init(args, pfmt);
}

void v4lgen_core_usage()
{
	return v4l1_core.cap_core_usage();
}

#endif

struct cap_core_t v4lgen_core = {
	cap_core_init: v4lgen_core_init,
	cap_core_close: NULL,
	cap_core_next: NULL,
	cap_core_status: NULL,
	cap_core_ftime: NULL,
	cap_core_overrides: NULL,
	cap_core_usage: v4lgen_core_usage,
	priv: NULL
};
